/*
 * =====================================================================================
 *
 *       Filename:  dump4svmlearn.cpp
 *
 *    Description:  Convert dump_rhog output to binary svm_learn format.
 *
 *        Version:  1.0
 *        Created:  Tuesday 12 June 2007 01:37:07  IST IST
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Navneet Dalal
 *        Company:  Lear / INRIA
 *
 * =====================================================================================
 */

#include <list>
#include <vector>
#include <string>
#include <fstream>
#include <iostream>

#include <lear/io/ioext.h>
#include <lear/blitz/blitzio.h>
#include <lear/io/fileheader.h>
#include <lear/io/biistream.h>
#include <lear/io/biostream.h>
#include <lear/util/customoption.h>
#include <lear/util/util.h>

#include <lear/cmdline.h>

#include "rawdescio.h"
#include "formatstream.h"

using std::cout; using std::cerr; using std::endl;
// ==========================================================================
// -------- Command typedefs        -----------------------------------------
// ==========================================================================
typedef std::vector<std::string>    FileListVector;

typedef int                         TargetType;

typedef blitz::Array<int,1>         Array1DInt;


enum FileFormat { BiSVMLight, SVMLight, BiMatlab, Matlab};

// ==========================================================================
// -------- Command line parameters -----------------------------------------
// ==========================================================================
static FileListVector posfile, negfile;
static std::string outfile;

static ptrdiff_t maxposvec, maxnegvec;
static int verbose;
static FileFormat formatType;

/// length of feature vector
static ptrdiff_t featureLength;

/// local feature vector cache
static Array1DType feature;

static void writeData(FormatOutStream<RealType,TargetType>* to, 
                      std::string const &filename,  
                      ptrdiff_t& maxvec, 
                      const TargetType target)
{
  RawDescIn desc(filename, verbose);

  if (featureLength <= 0) {
    featureLength = desc.featureLength();
    // allocate space for computing norm and std
    feature.resize(featureLength);
    
  } else if ((ptrdiff_t)desc.featureLength() != featureLength) {
    ThrowException("Unequal feature length in file " << filename
                   <<". Expected feature length " << featureLength
                   <<", found feature length " << desc.featureLength());
  }
  ptrdiff_t size = (ptrdiff_t)desc.featureCount();
  if (maxvec > 0){
    size = std::min(maxvec, (ptrdiff_t)desc.featureCount());
    maxvec -=size;
  }
  
  for (int m= 0; m< size && desc; ++m) {
    desc.next(feature); 
    to->write(feature, target);
    if (verbose > 5){
      cout << "Read descriptor " << std::setw(5) << m << endl;
    }
  }
  if (verbose > 2) {
    cout << "Processed descriptors " << std::setw(7) << size ;
    if (verbose > 3) {
      cout << " from file " << filename;
    }
    cout << endl;
  }
}


static void compute() {
  FormatOutStream<RealType,TargetType>* to = 0;
  switch (formatType) {
  case BiSVMLight:
    to= new BiSVMLightOutStream<RealType, TargetType> (outfile);
    break;
  case SVMLight:
    to= new SVMLightOutStream<RealType, TargetType> (outfile);
    break;
  case Matlab:
    to= new MatlabOutStream<RealType, TargetType>    (outfile);
    break;
  case BiMatlab:
    to= new BiMatlabOutStream<RealType, TargetType>  (outfile);
    break;
  default :
    ThrowException("Unspecified format specified");
  }


  if (verbose > 5) {
    cout << "Created formated stream" << endl;
  }

  TargetType target = 1;
  for (FileListVector::const_iterator f=posfile.begin();
       f!= posfile.end();++f) {
    writeData(to,*f,maxposvec,target);
    if (maxposvec==0) break;
  }
  target = -1;
  for (FileListVector::const_iterator f=negfile.begin();
       f!= negfile.end();++f) {
    writeData(to,*f,maxposvec,target);
    if (maxnegvec==0) break;
  }
  delete to;
  if (verbose > 5) {
    cout << "All done " << endl;
  }
}



int main(int argc, char** argv) {
  lear::Cmdline cmdline;
  using namespace lear;

  lear::CustomOption fileformat("out file format options ",BiSVMLight);
  fileformat
    //        .add("BiSVMLight. Binary custom svmlight format.",BiSVMLight)
    .add("BiSVMLight.",BiSVMLight)
    .add("SVMLight. Default svmlight format",SVMLight)
    .add("BiMatlab. Format: <int: ignore this> <int:#features> <int: feat_dim> <float: #features*feat_dim>",BiMatlab)
    .add("Matlab. Format: <First line: #features, feat_dim> Next #features line with feat_dim numbers in each line",Matlab)
    ;

  { 
    cmdline.commandName("dump2svmlearn");
    cmdline.version("0.0.1", "Author: Navneet Dalal "
                    "(mailto: Navneet.Dalal@inrialpes.fr)");
    cmdline.brief( "Dump descriptor to svm_learn format");

    cmdline.description( "Dump descriptor to svm_learn format");

    using namespace lear;
    cmdline.usageIssues(
                        "  'posfile'        positive descriptor file, directory, or a list file.\n"
                        "  'negfile'        negative descriptor file, directory, or a list file.\n"
                        );
    cmdline.appendUsageIssues(fileformat.usage());

    cmdline.addOption()
      ("verbose,v",option<int>(&verbose)
       ->defaultValue(0)->minValue(0)->maxValue(9),
       "verbose level")

      ("maxpos,P",option<ptrdiff_t>(&maxposvec)
       ->defaultValue(-1)->minValue(-1),
       "maximum positive vectors, -1 implies no limit")
      ("maxneg,N",option<ptrdiff_t>(&maxnegvec)
       ->defaultValue(-1)->minValue(-1),
       "maximum negative vectors, -1 implies no limit")

      ("posfile,p",option< FileListVector >(&posfile),
       "positive descriptor input file")
      ("negfile,n",option< FileListVector >(&negfile),
       "negative descriptor input file")

      // select data format
      ("format,f",option<std::string>(&(fileformat.option))
       ->defaultValue(fileformat.defaultOption()),
       "specify out file format")
      ;

    cmdline.addArgument()
      ("outfile",option<std::string>(&outfile),"out file")
      ;
  } 

  int status = cmdline.parse(argc, argv);
  if (status != cmdline.ok) 
    return status;

  formatType = static_cast<FileFormat> (fileformat.check());
  try {
    compute();
  } catch(lear::Exception& e) {
    cerr << "Caught " << e.what() << endl;
    return 1;
  } catch(std::exception& e) {
    cerr << "Caught " << (e.what() ? e.what() : "UNKNOWN ERROR") << endl;
    return 1;
  } catch(...) {
    cerr << "Caught unknown exception" << endl;
    return 1;
  }
  return 0;
}


