/*
 * ============================================================================
 *
 *       Filename:  dump_rhog.cpp
 *
 *    Description:  Provides functionality to dump dense RHOGDense
 *    descriptors.
 *
 *        Version:  1.0
 *        Created:  Monday 11 June 2007 12:34:11  IST IST
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Navneet Dalal
 *        Company:  Lear / INRIA
 *
 * ============================================================================
 */


#include <memory>
#include "windetectmain.h"
#include <lear/image/inimageiter.h>
#include <lear/exception.h>

using namespace lear;

int main(int argc, char** argv) {
  using namespace std;
  lear::Cmdline cmdline;
  bool presetexamples, suppressnonjittered;


  WinDetectDumpMain windetectmain;
  RHOGDenseMain rhogdensemain;

  WinDetectDump windetect;
  {  // cmdline
    cmdline.commandName("dump_rhog");
    cmdline.version("0.0.1", "Author: Navneet Dalal "
                    "(mailto: Navneet.Dalal@inrialpes.fr)");
    cmdline.brief("Compute RHOGDense features over images");
      
    cmdline.description(
"Compute RHOG fetures using image gradient strength over image windows.  When "
"running in normal mode, the whole algorithm is:"
"\n"
"1) divide image window into uniformly spaced dense points with point spacing "
"given by 'descstride'"
"\n"
"2) for each point, compute 3-dimensional position-orientation gradient "
"histogram"
"\n"
"3) finally create a big feature vector containing descriptor vector from each "
"point in the window"
"\n"
"When running in presetexamples mode, step 1 is replaced by loading a set of "
"preselected image windows.  In this mode, infile is a file describing this "
"set of windows.  Typically, one produces this file by running classify_rhog "
"with the --falsetxt option."
"\n"
);
      
    cmdline.addOption()
      ("presetexamples", bool_option(&presetexamples)->defaultValue(false),
       "If true, infile is an examples file that specifies which files and "
       "windows to use.  infile is interpreted as a set of positive or "
       "negative examples based on the value of the --label option.") 
      ("suppressnonjittered", bool_option(&suppressnonjittered)
       ->defaultValue(true),
       "If --suppressnonjittered=true and --presetexamples=true and --label=P "
       "then suppress descriptors at non-jittered locations from the output. "
       "If --presetexamples=false or --label=N, this option has no effect. "
       "This is useful when one wants to produce negative examples in the "
       "vicinity of all positive examples.  If one wants to produce positive "
       "examples from a positive preset examples infile, this option must be "
       "disabled.  Note that Dalal found that the existing jitter "
       "implementation doesn't work well.  In order for it to work well, we "
       "believe that you'd want to take jitter positions within a "
       "winstride-sized (or smaller) box and make them *positive* examples "
       "and then in a jitter region that excludes the winstride-sized (or "
       "larger) box, produce negative examples.")
      ;
    // set command line options. The statement below will fill up usage, and 
    // code for reading various command line options
    windetectmain.setCommonMainParam(cmdline, &windetect) ;
    windetectmain.setDumpParam(cmdline, &windetect) ;
    rhogdensemain.setRHOGDenseParam(cmdline) ;
  }
    
  int status = cmdline.parse(argc, argv);
  if (status != cmdline.ok)
    return status;
    
  if (!windetectmain.imageext.empty() && windetectmain.imageext[0] != '.')
    windetectmain.imageext = '.' + windetectmain.imageext;
    
  // now set options read from command line 
  windetectmain.fill(&windetect);
    
  std::vector<RHOGDenseParam const *> desc = rhogdensemain.fill();
    
  windetect.init(desc); 
  for (unsigned i= 0; i< desc.size(); ++i) 
    delete desc[i];
    
  try {
    if (windetectmain.outfile.size() == 0) {
      ThrowException("An output file is required right now.");
    }

    if (presetexamples) {
      if (windetectmain.testlocs.size()) {
        ThrowException("Sorry, writing the test locations is "
                       "not supported in presetexamples mode.");
      }
      windetect.writePresetExamples(windetectmain.infile, 
                                    windetectmain.outfile,
                                    suppressnonjittered);
    } else {
      auto_ptr<InputImageIterator> 
        infiles(createInputImageIterator(windetectmain.infile, 
                                         windetectmain.imageext));
      windetect.writeWinDesc(*infiles, windetectmain.outfile, 
                             windetectmain.testlocs);
    }
  } catch(lear::Exception& e) {
    cerr << "Caught "<< e.what() << endl;
    return 1;
  } catch(std::exception& e) {
    cerr << "Caught "<< (e.what() ? e.what() : "UNKNOWN ERROR") << endl;
    return 1;
  } catch(...) {
    cerr << "Caught unknown exception" << endl;
    return 1;
  }
  return 0;
}


