/*
 * =====================================================================================
 *
 *       Filename:  rawdescio.cpp
 *
 *    Description:  Provides implementation to rawdescio.h
 *
 *        Version:  1.0
 *        Created:  Tuesday 12 June 2007 02:11:47  IST IST
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Navneet Dalal
 *        Company:  Lear / INRIA
 *
 * =====================================================================================
 */

#include <iostream>
#include <iomanip>

#include <lear/types.h>
#include <lear/io/ioext.h>
#include <lear/blitz/tvmio.h>
#include <lear/blitz/blitzio.h>

#include "rawdescio.h"

const lear::FileHeader  RawDescIn::header("RawDesc",100);
using namespace std;
using namespace lear;

RawDescIn::RawDescIn(string const &filename_, const int verbose):
  filename_(filename_),from(filename_.c_str()),
  descNum_(0), current(0)
{
  if (!from) {
    ThrowException("Unable to open input file "+filename_);
  }
  from >> fheader; // Reader file header [RawDesc|100] 
  if (verbose > 7) 
    cout<< "Tag " << fheader.tag() 
        << " Version " << fheader.version() << endl;
  if (!fheader.identify(header.tag())) {
    ThrowException("File " + filename_+" does not look like a valid "
                   "descriptor file");
  }
  if (fheader.version() < 100) { // check for version number
    ThrowException("File " + filename_+" has unsupported version number ");
  }

  if (verbose > 5) {
    cout << "Processing file " << filename_ << endl;
  }
  FileHeader win_header;
  from >> win_header;// read window header [WinDesc|100]

  from >> descNum_; // how many descriptors for the feature vector

  descLength_.resize(descNum_);
  descExtent_.resize(descNum_);

  // extract the necessary metadata for each descriptor
  for (int i=0; i<descNum_; ++i) {
    FileHeader desc_header;
    from >> desc_header;// read descriptor header [RHOGDen|96]
    int32 len; // metadata size (including padding)
    from >> len; 
    BiIStream::pos_type const justAfterDescLen = from.tellg();

    from >> descLength_(i); // descriptor size/length
    from >> descExtent_(i); // size of descriptor window in pixels

    from.seekg(justAfterDescLen + (BiIStream::pos_type)len); // skip parts of the descriptor metadata we don't care about
  }

  // Read the cell grid locations for each descriptor
  int gridNum_;
  from >> gridNum_;// read number of Grid Vector 
  if (gridNum_ != descNum_) { // must be same as descriptor numbers
    ThrowException("The feature vector is to be composed of " << descNum_ <<
                   " descriptors, but " << gridNum_ << 
                   " cell grids are specified in " << filename_ << ".");
  }
  for (int i=0; i<gridNum_; ++i) {
    GridType g;
    from >> g;
    grid_.push_back(g);
  }

  from >> extent_;
  if (!from) {
    ThrowException("Unable to read window size from file "+filename_);
  }

  from >> length_;
  if (!from) {
    ThrowException("Unable to read feature length from file "+filename_);
  }

  from >> size_;
  if (!from) {
    ThrowException("Unable to read descriptor number from file "+filename_);
  }
  if (verbose > 2) {
    cout 
      << "Feature number    " << setw(5) << size_ <<'\n'
      << "\tFeature dimension " << setw(5) << length_ << '\n'
      << "\tDescriptor extent " << setw(3) << extent_[0] << 
      'x' << setw(3) << extent_[1] << endl;
  }

  startpos = from.tellg();
}

Array1DType& RawDescIn::next(Array1DType& f) {
  from >> f; 
  if (length_ != f.size()) {
    ThrowException("Feature length (" << f.size() << ") is not equal "
                   "to expected length (" << length_ << ") in  file \""
                   << filename_ << "\".");
  }
  if (!from) {
    ThrowException("Premature I/O error in file " + filename_);
  }
  ++current;
  readExtra();
  return f;
}


void RawDescIn::readExtra() {
  validxyinfo_=false;
  if (fheader.version() >= 110) {
    from >> fwinlbound_ >> fwinextent_ >> fwinscale_;
    validxyinfo_=true;
  } 
  validfilename_=false;
  if (fheader.version() >= 120) {
    from >> fwinfilename_;
    validfilename_=true;
  } 
}
void RawDescIn::next_split(
                           Array2DType& rf, Array1DType const &f, const int index) const
{
  using namespace blitz;
  int s = 0;
  for (int i= 0; i< index; ++i) 
    s += descLength_(i);

  Array1DType tf ( f(Range(s,s+grid(index).size()*descLength(index)-1)) );

  // create a new storage array 
  std::copy(tf.begin(),tf.end(),rf.begin());
}


Array2DType RawDescIn::next_split( Array1DType const &f, const int index) const
{
  // create a new storage array 
  Array2DType sf(grid(index).size(), descLength(index));
  next_split(sf,f,index);
  return sf;
}

