#include <stdio.h>
#include <iostream>
#include <string>

#include <lear/cmdline.h>
#include <lear/image/imageio.h>
#include <lear/util/fileutil.h>
#include <lear/cvision/classify.h>

#include <pdcuda/rhogdense.h>

#include "rawdescio.h"

typedef blitz::Array<blitz::TinyVector<float,3>, 2> RGBImage;
typedef blitz::Array<float, 2>                      GrayImage;

using namespace std;
using namespace lear;

void renderBlockDescriptorsNonMerged(string const &fname, float *data, 
                                     int numCellX, int numCellY,
                                     int orientBin, 
                                     int numBlockX, int numBlockY, 
                                     bool semiCirc,
                                     int visScale, 
                                     int cellBorder, int blockBorder,
                                     float minVal, float maxVal)
{
  static uint8 const CellGridColor[]  = {0,0,255};
  static uint8 const BlockGridColor[] = {0,255,255};
  static uint8 const HistColor[]      = {255,255,255};

  int const visCellWidth   = cellBorder + visScale;
  int const visCellHeight  = cellBorder + visScale;
  int const visBlockWidth  = blockBorder + visScale*numCellX + cellBorder*(numCellX-1);
  int const visBlockHeight = blockBorder + visScale*numCellY + cellBorder*(numCellY-1);

  RGBImage img(visBlockWidth*numBlockX + blockBorder,
               visBlockHeight*numBlockY + blockBorder);
  memset(&img(0,0), 0, sizeof(img(0,0))*img.size());

  // Draw cell gridlines
  for (int bx=0; bx<numBlockX; bx++) {
    for (int cx=1; cx<numCellX; cx++) {
      for (int border=0; border<cellBorder; border++) {
        for (int y=0; y<img.extent()[1]; y++) {
          for (int c=0; c<3; c++) {
            img(bx*visBlockWidth + cx*visCellWidth - border + blockBorder - 1, y)[c] = CellGridColor[c];
          }
        }
      }
    }
  }
  for (int by=0; by<numBlockY; by++) {
    for (int cy=1; cy<numCellY; cy++) {
      for (int border=0; border<cellBorder; border++) {
        for (int x=0; x<img.extent()[0]; x++) {
          for (int c=0; c<3; c++) {
            img(x, by*visBlockHeight + cy*visCellHeight - border + blockBorder - 1)[c] = CellGridColor[c];
          }
        }
      }
    }
  }

  // Draw block gridlines
  for (int bx=0; bx<=numBlockX; bx++) {
    for (int border=0; border<blockBorder; border++) {
      for (int y=0; y<img.extent()[1]; y++) {
        for (int c=0; c<3; c++) {
          img(bx*visBlockWidth+border, y)[c] = BlockGridColor[c];
        }
      }
    }
  }
  for (int by=0; by<=numBlockY; by++) {
    for (int border=0; border<blockBorder; border++) {
      for (int x=0; x<img.extent()[0]; x++) {
        for (int c=0; c<3; c++) {
          img(x, by*visBlockHeight+border)[c] = BlockGridColor[c];
        }
      }
    }
  }

  // Iterate over all blocks
  for (int by=0; by<numBlockY; by++) {
    for (int bx=0; bx<numBlockX; bx++) {

      // Iterate over all cells      
      for (int cy=0; cy<numCellY; cy++) {
        for (int cx=0; cx<numCellX; cx++) {

          // Iterate over the visualization area for a block
          for (int vy=0; vy<visScale; vy++) {
            for (int vx=0; vx<visScale; vx++) {
              // Note: I'm being a little sloppy here about quantizing
              // bins and there may be having half-bin misalignments.
              // TODO: !semiCirc is broken
              int o = int(atan2(vy+0.5f-visScale/2.0f, 
                                vx+0.5f-visScale/2.0f)
                          * 180/M_PI + 180);
              if (semiCirc) o %= 180;
              o = (o * orientBin / 180);
              
              // (cy,cx,bin,y,x): last dim is most contiguous 
              RealType const binVal = data[((((bx)*numBlockY + 
                                              by)*numCellX + 
                                             cx)*numCellY + 
                                            cy)*orientBin + 
                                           o];

              float const colorScaling = 
                MIN(1, MAX(0, (binVal - minVal) / (maxVal - minVal)));
              for (int c=0; c<3; c++) {
                img(bx*visBlockWidth + cx*visCellWidth + blockBorder + vx,
                    by*visBlockHeight + cy*visCellHeight + blockBorder + vy)[c] =
                  int(HistColor[c] * colorScaling);
              }
            }
          }
        }
      }
    }
  }

  // Write image
  lear::ImageIO::write(fname, img);
}

#define CHECK_STAT(NAME,ST,EXPECTED) \
    if (ST != EXPECTED) { \
      ThrowException("Unexpected " << NAME << ": " <<  \
                     ST << " (expected " << EXPECTED << ")"); \
    }


int main(int argc, char **argv) {
  int    verbose;
  int    visScale, cellBorder, blockBorder;
  string rhogFile;
  int    featureIndex;
  string outImgFile;
  float  minVal, maxVal;

  lear::Cmdline cmdline;
  cmdline.commandName("render_rhog");
  cmdline.version("0.0.2", "Author: Gerald Dalley "
                  "(mailto: dalleyg@ieee.org)");
  cmdline.brief("Renders an RHOG-based feature vector for a detector window");
  cmdline.description("Reads a .RHOG file and renders the requested detector "
                      "window feature vectors as an image for visualization "
                      "purposes.");

  // FUTURE WORK:
  // Allow renderings directly from input images (this will require
  // adding support for nearly all of the dump_rhog capabilities here).

  cmdline.addOption()
    ("verbose,v",option<int>(&verbose)
     ->defaultValue(2)->minValue(0)->maxValue(9),
     "verbose level")
    ("scale,s",option<int>(&visScale)
     ->defaultValue(31)->minValue(1),
     "rendered size of an RHOG block descriptor (in pixels)")
    ("cellborder",option<int>(&cellBorder)
     ->defaultValue(2)->minValue(0),
     "border around each rendered RHOG cell descriptor (in pixels)") 
    ("blockborder",option<int>(&blockBorder)
     ->defaultValue(6)->minValue(0),
     "border around each rendered RHOG block descriptor (in pixels)")
    ("index,i",option<int>(&featureIndex)
     ->minValue(0)->defaultValue(0),
     "index of the feature vector to render in the input file "
     "(--index=0 means the first detector window feature vector found in "
     "the input .RHOG file)")
    ("min",option<float>(&minVal)
     ->defaultValue(0),
     "bin values at or below this value are drawn as black.")
    ("max",option<float>(&maxVal)
     ->defaultValue(1),
     "bin values at or above this value are drawn as white.")
    ;
  cmdline.addArgument()
    ("rhog", option<std::string>(&rhogFile), "input (.RHOG) file name")
    ("out", option<std::string>(&outImgFile), "output image file name")
    ;
  int status = cmdline.parse(argc, argv);
  if (status != cmdline.ok) return status;

  //////////////////////////////////////////////////////////////////////////


  try {
    Array1DType feature; 
    // Read the feature
    if (strcasecmp(lear::extension(rhogFile).c_str(), "alt") == 0) {
      CHECK_STAT("index", featureIndex, 0);
      lear::LinearClassify cls(rhogFile);
      feature.resize(cls.length());
      feature = cls.classifier();
    } else if (strcasecmp(lear::extension(rhogFile).c_str(), "RHOG") == 0) {
      RawDescIn in(rhogFile, verbose);
      if (in.featureCount() <= featureIndex) {
        ThrowException("Feature vector " << featureIndex << " was requested, "
                       "but only " << in.featureCount() << " features are "
                       "found in " << rhogFile);
      }
      feature.resize(in.featureLength());
      for (int i=0; i<=featureIndex; i++) {
        // FUTURE: just skip past the features we don't want
        in.next(feature);
      }
                     
      // FUTURE: allow the user to specify the window geometry details
      // and just validate the total feature length here.  For now, we
      // restrict this application to working with the most typical
      // setup.
      CHECK_STAT("descriptor window width",  in.extent()[0], pdcuda::WinSizeX);
      CHECK_STAT("descriptor window height", in.extent()[1], pdcuda::WinSizeY);
      CHECK_STAT("number of descriptor types", in.descNumber(), 1);
      CHECK_STAT("feature vector length", 
                 in.featureLength(), pdcuda::WinDescLen);
    } else {
      ThrowException("Unrecognized input file format");
    }

    // Render it
    renderBlockDescriptorsNonMerged(outImgFile, feature.data(),
                                    pdcuda::NumCellX, pdcuda::NumCellY,
                                    pdcuda::OrientBin, 
                                    pdcuda::NumBlockX, pdcuda::NumBlockY, 
                                    pdcuda::SemiCirc,
                                    visScale, cellBorder, blockBorder,
                                    minVal, maxVal);

  } catch(lear::Exception& e) {
    cerr << "Caught " << e.what() << endl;
    return 1;
  } catch(std::exception& e) {
    cerr << "Caught " << (e.what() ? e.what() : "UNKNOWN ERROR") << endl;
    return 1;
  } catch(...) {
    cerr << "Caught unknown exception" << endl;
    return 1;
  }
  return 0;
}

/*
  # look at results for a hard negatives test
  reset ; make -j16 && \
  (((time nice -n +10 app/dump_rhog --cuda=0 --samples 0 --infile /scratch/dalleyg/dalalTriggs/trunk/experiments/inria_20080211/w64x128c8x8_all/hard_neg.txt --outfile cpuhard.RHOG --poscases 2416 --negcases 12180 --memorylimit 1700 --dumphard 1 --hardscore 0 --verbose=9 --imageext=png --window=64,128 --winstride=8,8 --topleft=16,16 --cellsize=8,8 --numcell=2,2 --orientbin=9 --descstride=8,8 --proc=rgb_sqrt --norm=l2hys --fullcirc=0 --wtscale=2 --gscale=0 --maxvalue=0.2 --epsilon=1 ) 2>&1) > cpuhard.log) && \
  (((time nice -n +10 app/dump_rhog --cuda=1 --samples 0 --infile /scratch/dalleyg/dalalTriggs/trunk/experiments/inria_20080211/w64x128c8x8_all/hard_neg.txt --outfile gpuhard.RHOG --poscases 2416 --negcases 12180 --memorylimit 1700 --dumphard 1 --hardscore 0 --verbose=9 --imageext=png --window=64,128 --winstride=8,8 --topleft=16,16 --cellsize=8,8 --numcell=2,2 --orientbin=9 --descstride=8,8 --proc=rgb_sqrt --norm=l2hys --fullcirc=0 --wtscale=2 --gscale=0 --maxvalue=0.2 --epsilon=1 ) 2>&1) > gpuhard.log) && app/compare_rhog cpuhard.RHOG gpuhard.RHOG && \
  reset; export IDX=0 && make -j16 && app/render_rhog cpuhard.RHOG cpuhard.png -i $IDX && app/render_rhog gpuhard.RHOG gpuhard.png -i $IDX && montage -geometry 441x945 -border 1 -bordercolor blue -background blue cpuhard.png gpuhard.png t.png && display t.png

  # look at results for positive training samples
  reset ; make -j16 && \
  (((time nice -n +10 app/dump_rhog --cuda=0 --samples=1 /scratch/dalleyg/dalalTriggs/trunk/sourcedata/INRIAPerson/train/inria/pos_20080211_w64x128c8x8_all.lst cpu.RHOG --verbose=9 --imageext=png --window=64,128 --winstride=8,8 --topleft=16,16 --cellsize=8,8 --numcell=2,2 --orientbin=9 --descstride=8,8 --proc=rgb_sqrt --norm=l2hys --fullcirc=0 --wtscale=2 --gscale=0 --maxvalue=0.2 --epsilon=1 ) 2>&1) > cpu.log) && \
  (((time nice -n +10 app/dump_rhog --cuda=1 --samples=1 /scratch/dalleyg/dalalTriggs/trunk/sourcedata/INRIAPerson/train/inria/pos_20080211_w64x128c8x8_all.lst gpu.RHOG --verbose=9 --imageext=png --window=64,128 --winstride=8,8 --topleft=16,16 --cellsize=8,8 --numcell=2,2 --orientbin=9 --descstride=8,8 --proc=rgb_sqrt --norm=l2hys --fullcirc=0 --wtscale=2 --gscale=0 --maxvalue=0.2 --epsilon=1 ) 2>&1) > gpu.log) && \
  reset; export IDX=0 && make -j16 && app/render_rhog cpu.RHOG cpu.png -i $IDX && app/render_rhog gpu.RHOG gpu.png -i $IDX && montage -geometry 441x945 -border 1 -bordercolor blue -background blue cpu.png gpu.png t.png && display t.png
  
 */
