/**
 * @file
 * @brief Cmdline parsing class.
 *
 * Cmdline is Navneet Dalal's interface to boost::program_options
 * package.  It hides various program_options commands and presents
 * user a even much simpler interface for parsing commands.
 *
 * @todo
 * 1. Check if file exists
 * 2. Put confilicting and depends upon checks
 */


#ifndef _LEAR_CMDLINE_H_
#define _LEAR_CMDLINE_H_


#include <string>
#include <iostream>
#include <boost/program_options.hpp>

#include <lear/util/util.h>
#include <lear/option.h>


namespace lear {

  /// make parameter as namespace lear alias.
  // using boost::program_options::parameter;

  // make a alias "po" in lear for boost::program_options
  namespace po = boost::program_options;

  class Cmdline;
  /**
   * Provide a wrapper and holder to command line arguments. This way, we
   * can specify command line options and positional parameters in one command
   */
  class ArgumentHolder {
  public:
    ArgumentHolder(char const *title, int warpWidth) :
      desc(title, warpWidth)
    {};
    ArgumentHolder& operator()( 
                               char const *name, 
                               po::value_semantic const *s, 
                               char const *description,
                               const int count=1)
    {
      desc.add_options()(name,s,description);
      position.add(name,count);
      args.push_back(name);
      return *this;
    }


  private:
    friend class Cmdline;

    /// Options_description for parameters
    po::options_description  desc;

    /// Store parameter's positions
    po::positional_options_description  position;

    std::vector< std::string> args;
  };



  /**
   * Cmdline provide user a command line parser. It supports optional, 
   * positional, required, conflicting and dependent options. Also range 
   * checking for basic types is provided.
   *
   * Internally, Cmdline uses boost::program_options library.
   */

  class Cmdline {
  public:
    /// status enum. Return various error types.
    enum status { ok=0, info=1, failed, invalid_usage, duplicate_usage, 
                  conflicting_usage}; 

    Cmdline(std::string const configfile_ = "",
            bool const helpOnNoArgs_ = true);

    status parse(int argc, char** argv);

    po::options_description_easy_init addOption() {
      return descConfigfile.add_options();
    }
    po::options_description_easy_init addCmdline() {
      return descCmdline.add_options();
    }
    po::options_description_easy_init addHidden() {
      return descHidden.add_options();
    }
    ArgumentHolder& addArgument() {
      return argument;
    }

        
    std::string version() const {
      return version_ + "\n" + versionDetails_;
    }

    void version(std::string const &v, 
                 std::string const &d=std::string() ) 
    {
      version_ = v;
      versionDetails_ = d;
    }

    std::string brief() const {
      return brief_ ;
    }

    void brief(std::string const &b ) {
      brief_ = b;
      warptext(brief_,WARP_WIDTH);
    }
    void appendBrief(std::string const &b ) {
      brief_ += b;
      warptext(brief_,WARP_WIDTH);
    }

    std::string description() const {
      return description_ ;
    }

    void description(std::string const &d ) {
      description_ = d;
      warptext(description_,WARP_WIDTH);
    }

    std::string usageString() const {
      return usageString_ ;
    }

    std::string commandName() const {
      return commandName_ ;
    }

    void commandName(std::string const &c ) {
      commandName_ = c;
    }

    std::string usageIssues() const {
      return usageIssues_ ;
    }

    void usageIssues(std::string const &u ) {
      usageIssues_ = u;
      warptext(usageIssues_,WARP_WIDTH);
    }
    void appendUsageIssues(std::string const &u ) {
      usageIssues_ += u;
      warptext(usageIssues_,WARP_WIDTH);
    }
        

    void printUsage(std::ostream& o = std::cout) const ;
    void printOptions(std::ostream& o = std::cout) const ;
    void printVersion(std::ostream& o = std::cout) const;
    void printError(std::string const &mesg, std::ostream& o = std::cerr) const;
    void output(std::ostream& os, po::options_description const &d) const; 

    //          /// option 'a' conflicts with option 'b'
    //          void conflicts(std::string const &a, std::string const &b) {
    //              conflict_.push_back(make_pair(a,b));
    //          }
    //          /// option 'which' depens 'on' option
    //          void depends(std::string const &which, std::string const &on) {
    //              depend_.push_back(make_pair(which,on));
    //          }
  private:
    static const unsigned WARP_WIDTH;
    po::options_description desc, descCmdline, descConfigfile;
    po::options_description  descHidden, descVisual;

    ArgumentHolder argument;

    /// Variable map for options
    po::variables_map vm;

    /// Show usage message if no arguments are provided on command line
    bool helpOnNoArgs;

    std::string configfile;

    std::string version_, versionDetails_;
    std::string usageString_, usageIssues_;
    std::string brief_, description_;
    std::string commandName_;
        
    //          std::vector<std::pair<std::string, std::string> > conflict_;
    //          std::vector<std::pair<std::string, std::string> > depend_;

    void constructUsageString(std::string const &name);

    /* Auxilliary functions for checking input for validity. */

    /* Function used to check that 'opt1' and 'opt2' are not specified
       at the same time. */
    void conflicting_options(char const *opt1, char const *opt2) const;

    /* Function used to check that of 'for_what' is specified, then
       'required_option' is specified too. */
    void option_dependency(
                           char const *for_what, 
                           char const *required_option) const;
  };


  /// << operator overload for os stream. Print complete usage information
  std::ostream& operator<<(std::ostream& os, Cmdline const &desc) ;

} // namespace lear
#endif // _LEAR_CMDLINE_H_


