#ifndef _LEAR_DESC_NORMALIZER_H_
#define _LEAR_DESC_NORMALIZER_H_

#include <cmath>

#include <blitz/array.h>
#include <lear/io/biostream.h>

namespace lear {

  template<class RealType_>
  class DescNormalizer {
  public:
    enum {Method = 0};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    DescNormalizer(){}
    virtual ~DescNormalizer(){}

    template<int N>
    void operator()(blitz::Array<RealType,N>& vec) const {
      using namespace blitz;
      Array1DType t(vec.data(),shape(vec.size()),neverDeleteData);
      doit(t);
    }
    virtual char const *toString() const {
      return "'NONE'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }
    virtual void print(std::ostream& o) const {
      o <<    "Normalizer " << toString();
    }
    virtual void print(lear::BiOStream& o) const {
      o << toMethod();
    }

  protected:
    // Does nothing
    virtual void doit(Array1DType&) const {}
  };

  template<class RealType_>
  class L1Normalizer : public DescNormalizer<RealType_> {
    typedef DescNormalizer<RealType_>           Parent;
  public:
    enum {Method = 1};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    L1Normalizer(RealType const epsilon): 
      DescNormalizer<RealType>(), epsilon_(epsilon)
    {}

    RealType epsilon() const { return epsilon_; }

    virtual char const *toString() const {
      return "'L1 norm'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }

    virtual void print(std::ostream& o) const {
      using namespace std;
      o <<    "Normalizer " << toString() <<  
        ", eps " << setw(2) << setprecision(1) << epsilon_;
    }
    virtual void print(lear::BiOStream& o) const {
      o << toMethod() << epsilon_;
    }
  protected:
    virtual void doit(Array1DType& vec) const {
      RealType const norm = (RealType)(blitz::sum(vec) + epsilon_*vec.size());
      vec /= norm;
    }
    /// add small epsilon to make normalization process stable
    RealType epsilon_;
  };

  template<class RealType_>
  class L2Normalizer : public L1Normalizer<RealType_>{
    typedef L1Normalizer<RealType_>             Parent;
  public:
    enum {Method = 2};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    L2Normalizer(RealType const epsilon): L1Normalizer<RealType>(epsilon) {}

    virtual char const *toString() const {
      return "'L2 norm'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }
    virtual void print(std::ostream& o) const {
      L1Normalizer<RealType>::print(o);
    }
    virtual void print(lear::BiOStream& o) const {
      L1Normalizer<RealType>::print(o);
    }
  protected:
    virtual void doit(Array1DType& vec) const {
      RealType const norm = (RealType)(std::sqrt(blitz::sum(vec*vec)) + 
                                       Parent::epsilon_*vec.size());
      vec /= norm;
    }
  };

  template <class T> 
  inline T square(T x) { return x*x; }

  template<class RealType_>
  class L2HysNormalizer : public L2Normalizer<RealType_>
  {
  public:
    enum {Method = 3};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    L2HysNormalizer(RealType const epsilon_, RealType const maxval_)
      : L2Normalizer<RealType>(epsilon_),maxval_(maxval_), epsilon2_(0.01f) {}

    RealType maxValue() const { return maxval_; }
    RealType epsilon2() const { return epsilon2_; }

    virtual char const *toString() const {
      return "'L2 + hys'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }
    virtual void print(std::ostream& o) const {
      L2Normalizer<RealType_>::print(o);
      using namespace std;
      o << ", hysth " << setw(4) << setprecision(1) << maxval_;
    }
    virtual void print(lear::BiOStream& o) const {
      L2Normalizer<RealType_>::print(o);
      o << maxval_;
    }

  protected:
    virtual void doit(Array1DType& vec) const {
      /* Original 5-loop code
         L2Normalizer<RealType>::doit(vec);
         for (int i= 0; i< vec.size(); ++i) 
         if (vec(i) > maxval_)
         vec(i) = maxval_;
         //L2Normalizer<RealType>::doit(vec);
         RealType const norm = (RealType)(std::sqrt(sum(vec*vec)) + epsilon2_);
         vec /= norm;
      */
     
      /// New 3-loop code (local 3x speedup)
      // find l2 norm
      int const N = vec.size();
      RealType norm = 0;
      for (int i=0; i<N; ++i) {
        norm += square(vec(i));
      }
      norm = (RealType)std::sqrt(norm) + L1Normalizer<RealType>::epsilon_*N;

      // apply l2 norm, clip values, find hysnorm
      RealType hysnorm = 0;
      for (int i=0; i<N; ++i) {
        RealType const val = vec(i) / norm;
        RealType const clipped = (val>maxval_) ? maxval_ : val;
        vec(i) = clipped;
        hysnorm += square(clipped);
      }
      hysnorm = (RealType)(std::sqrt(hysnorm) + epsilon2_);

      // apply hysnorm
      vec /= hysnorm;
    }
    RealType maxval_, epsilon2_;
  };

  template<class RealType_>
  class L1SqrtNormalizer : public L1Normalizer<RealType_>{
  public:
    enum {Method = 4};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    L1SqrtNormalizer(RealType const epsilon): 
      L1Normalizer<RealType>(epsilon) {}

    virtual char const *toString() const {
      return "'L1 + sqrt'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }

  protected:
    virtual void doit(Array1DType& vec) const {
      L1Normalizer<RealType>::doit(vec);
      vec = blitz::sqrt(vec);
    }
  };

  ///// Second set of normalizers.. These one check if norm < epsilon, if yes normalize them to set to 1.
  template<class RealType_>
  class L1TradNormalizer : public L1Normalizer<RealType_> {
    typedef L1Normalizer<RealType_>             Parent;
  public:
    enum {Method = 5};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    L1TradNormalizer(RealType const epsilon): 
      Parent(epsilon)
    {}

    virtual char const *toString() const {
      return "'L1Trad norm'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }

  protected:
    virtual void doit(Array1DType& vec) const {
      RealType norm = blitz::sum(vec);
      if (norm < Parent::epsilon_*vec.size()) norm = 1;
      vec /= norm;
    }
  };

  template<class RealType_>
  class L2TradNormalizer : public L2Normalizer<RealType_>{
    typedef L2Normalizer<RealType_>             Parent;
  public:
    enum {Method = 6};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    L2TradNormalizer(RealType const epsilon): Parent(epsilon) {}

    virtual char const *toString() const {
      return "'L2Trad norm'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }
  protected:
    virtual void doit(Array1DType& vec) const {
      RealType norm = std::sqrt(blitz::sum(vec*vec));
      if (norm < Parent::epsilon_*vec.size()) norm = 1;
      vec /= norm;
    }
  };

  template<class RealType_>
  class L2TradHysNormalizer : public L2HysNormalizer<RealType_>{
    typedef L2HysNormalizer<RealType_>          Parent;
  public:
    enum {Method = 7};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    L2TradHysNormalizer(RealType const epsilon, RealType const maxval_)
      : Parent(epsilon,maxval_) {}

    virtual char const *toString() const {
      return "'L2Trad + hys'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }
  protected:
    virtual void doit(Array1DType& vec) const {
      RealType norm = std::sqrt(blitz::sum(vec*vec));
      if (norm < Parent::epsilon_*vec.size()) norm = 1;
      vec /= norm;
      for (int i= 0; i< vec.size(); ++i) 
        if (vec(i) > Parent::maxval_)
          vec(i) = Parent::maxval_;
      norm = std::sqrt(sum(vec*vec));
      if (norm < Parent::epsilon_*vec.size()) norm = 1;
      vec /= norm;
    }
  };

  template<class RealType_>
  class L1TradSqrtNormalizer : public L1TradNormalizer<RealType_>{
    typedef L1TradNormalizer<RealType_>          Parent;
  public:
    enum {Method = 8};
    typedef RealType_                           RealType;
    typedef blitz::Array<RealType,1>            Array1DType;

    L1TradSqrtNormalizer(RealType const epsilon): Parent(epsilon) {}

    virtual char const *toString() const {
      return "'L1Trad+sqrt'";
    }
    virtual unsigned toMethod() const {
      return Method;
    }

  protected:
    virtual void doit(Array1DType& vec) const {
      L1TradNormalizer<RealType>::doit(vec);
      vec = blitz::sqrt(vec);
    }
  };

  template<class RealType_> std::ostream& 
  operator<<( std::ostream& o, DescNormalizer<RealType_> const &d)
  { d.print(o); return o; }
  template<class RealType_> lear::BiOStream& 
  operator<<( lear::BiOStream& o, DescNormalizer<RealType_> const &d)
  { d.print(o); return o; }

}

#endif // _LEAR_NORMALIZER_H_

