#ifndef _LEAR_PERSPECTIVE_SCALE_PYRAMID_H_
#define _LEAR_PERSPECTIVE_SCALE_PYRAMID_H_

#include <iostream>
#include <map>
#include <set>
#include <blitz/tinyvec.h>
#include <blitz/tinymat.h>
#include <lear/image/rescale.h>
#include <lear/cvision/imageslider.h>
#include <lear/cvision/perspective.h>

namespace lear {
  /**
  */
  class PerspectiveScalePyramid {
  public:
    typedef float                             RealType;
    typedef blitz::TinyVector<int,2>          IndexType;

    // Dimension indices for various matrices and vectors
    static int const X = PerspectiveProjector::X;
    static int const Y = PerspectiveProjector::Y;
    static int const Z = PerspectiveProjector::Z;
    static int const W = PerspectiveProjector::W; 
    static int const S = PerspectiveProjector::S; 

    typedef std::set<int> TopRows;
    // Mapping from detector window scale to image rows where that scale 
    // should be used.  [finalScale]{scaledRow}
    typedef std::map<RealType, TopRows> ScaleToTopRows;
    
  public:
    PerspectiveScalePyramid(
      IndexType         const imageExtent, 
      IndexType         const winsize, 
      int               const borderPixels,     // winsize[Y] = borderPixels*2 + person height in pixels
      IndexType         const winstride, 
      RealType          const pedestrianHeight, // canonical pedestrian height (world coordinates)
      blitz::TinyVector<RealType,3>   const rodrigues,        // world-to-camera Rodrigues rotation vector
      blitz::TinyVector<RealType,3>   const translation,      // world-to-camera translation
      blitz::TinyVector<RealType,2>   const focalLength,      // focal length of the camera (pixels)
      RealType          const scalePrecision, 
      RealType          const scaleRatio_ = 1.2,
      RealType                endScale_ = 0,
      RealType                startScale_ = 1)
      : imageExtent(imageExtent), winsize(winsize)
    { 
      using namespace blitz;
      for (int i=0; i<focalLength.numElements; i++) {
        if (focalLength[i] <= 0) {
          ThrowException("Focal length elements must be strictly positive.  "
                         "In the future we might add support for negative "
                         "focal lengths to handle image flipping");
        }
      }

      if ((scaleRatio_ <= 1) && (endScale_ != startScale_)) {
        ThrowException("Scale ratio must be greater than 1");
      }

      // check if endScale will give proper window size
      if (endScale_ > 0) {
        IndexType e (imageExtent/endScale_);
        if (sum(e<winsize)) {
          endScale_ = 0;
          std::cout << 
            " Scaling using end scale will create too small image. Changing\n"
            " to appropiate end scale" << std::endl; 
        }
      }

      // if no endscale, compute endscale from image size
      if (endScale_ == 0) {
        TinyVector<RealType,2> e = imageExtent;
        endScale_ = min(e/winsize);
      } else if (endScale_ < 0) {
        // negative endscale. provides further zoom out on image
        TinyVector<RealType,2> e = imageExtent;
        endScale_ = min(e/winsize*(-endScale_));
      }
 
      // number of steps in pyramid
      int const numBaseScales = (endScale_ == startScale_) ? 1 :
        int(std::log(endScale_/startScale_)/std::log(scaleRatio_)) + 1;
      if (numBaseScales <= 0) return;

      PerspectiveProjector p(imageExtent, winsize, borderPixels, 
        pedestrianHeight, rodrigues, translation, focalLength);
      
      for (int scaleNum=0; scaleNum<numBaseScales; scaleNum++) {
        RealType const baseScale = 
          RealType(startScale_ * pow(scaleRatio_, scaleNum));

        // I tried using an ImageSlider<1> and got really odd blitz template
        // errors, so for now we'll use a degenerate 2D slider.
        TinyVector<RealType,2> const extendedVStrip
          (winsize[X], (imageExtent[Y]+winsize[1])/baseScale);
        typedef lear::ImageSlider<2> SliderType;
        SliderType vslider(extendedVStrip, winsize, winstride);
        for (SliderType::iterator siter = vslider.begin(); 
          siter != vslider.end(); ++siter) 
        {
          RealType const topRowScaled = (*siter)[Y];

          // do all the perspective math in the unscaled image space
          RealType const topRow = topRowScaled * baseScale;

          RealType const scale = p.scaleFromTopRow(topRow, baseScale);
          if (scale <= 0) continue; // don't look for people in the sky
          RealType const quantScale =
            (scalePrecision > 0) ?
            (int(scale / scalePrecision + 0.5) * scalePrecision) : scale;
          int const topRowRescaled = int(topRow/quantScale);

          int const bottomRowUnscaled = int((winsize[Y] + topRowRescaled) * quantScale);
          if (bottomRowUnscaled > imageExtent[Y]) continue;

          scaleData[quantScale].insert(topRowRescaled);
        }
      }
    }

    size_t   numDistinctScales() const { return scaleData.size();          }
    RealType smallestScale()     const { return scaleData.begin()->first;  }
    RealType largestScale()      const { return scaleData.rbegin()->first; }
    
    ScaleToTopRows const &scaleToTopRows() const { return scaleData; }
    
  protected: 
    ScaleToTopRows          scaleData;
    IndexType         const imageExtent;
    IndexType         const winsize;
    
    friend class CommonIter;

    template <typename Parent_, typename MapIterator, typename SetIterator>
    class CommonIter
    {
    protected:
      typedef Parent_        Parent;
      typedef MapIterator    MapIter;
      typedef SetIterator    SetIter;
      typedef CommonIter<Parent, MapIterator, SetIterator> Base;

    public:
      CommonIter(Parent &parent, MapIterator mapBegin) 
        : parent(parent), mapIter(mapBegin) 
      {
        while (mapIter != parent.scaleData.end()) {
          if (mapIter->second.begin() == mapIter->second.end()) {
            ++mapIter;
          } else {
            setIter = mapIter->second.begin();
            break;
          }
        }
      }

      CommonIter &operator++() {
        if (mapIter != parent.scaleData.end()) {
          assert(setIter != mapIter->second.end());
          ++setIter;
          if (setIter == mapIter->second.end()) {
            ++mapIter;
            while (mapIter != parent.scaleData.end()) {
              if (mapIter->second.begin() == mapIter->second.end()) {
                ++mapIter;
              } else {
                setIter = mapIter->second.begin();
                break;
              } 
            }
          }
        }
        return *this;
      }

      // TODO: bidirectional iterator
  
      template <typename P, typename MI, typename SI>
      bool operator!=(CommonIter<P,MI,SI> const &rhs) const {
        return !(*this == rhs);
      }

      template <typename P, typename MI, typename SI>
      bool operator==(CommonIter<P,MI,SI> const &rhs) const {
        if (this->mapIter == this->parent.scaleData.end()) {
          return (rhs.mapIter == rhs.parent.scaleData.end());
        } else if (rhs.mapIter == rhs.parent.scaleData.end()) {
          return false;
        } else {
          return (this->mapIter == rhs.mapIter) && 
            (this->setIter == rhs.setIter);
        }
      }

      RealType  scale()   const { return mapIter->first; }
      int       topRow()  const { return *setIter; }
      IndexType topleft() const { return IndexType(0, topRow()); }
      IndexType extent()  const {
        return IndexType(int(parent.imageExtent[X]*scale()), parent.winsize[Y]);
      }

    protected:
      Parent      &parent;
      MapIterator  mapIter;
      SetIterator  setIter;
    };

  public:
    class iterator : public CommonIter<PerspectiveScalePyramid, ScaleToTopRows::iterator, std::set<int>::iterator> {
    public:
      iterator &operator++()    { return (iterator &)(++(*(Base*)this)); }
      iterator &operator++(int) { return ++(*this); }
    protected:
      friend class PerspectiveScalePyramid;
      iterator(Parent &parent, MapIter mapIter): CommonIter<Parent,MapIter,SetIter>(parent, mapIter) {}
    };

    class const_iterator : public CommonIter<PerspectiveScalePyramid const, ScaleToTopRows::const_iterator, std::set<int>::const_iterator> {
    public:
      const_iterator &operator++()    { return (const_iterator &)(++(*(Base*)this)); }
      const_iterator &operator++(int) { return ++(*this); }
    protected:
      friend class PerspectiveScalePyramid;
      const_iterator(Parent &parent, MapIter mapIter): CommonIter<Parent,MapIter,SetIter>(parent, mapIter) {}
    };

    iterator        begin()        { return       iterator(*this, scaleData.begin());  }
    iterator        end()          { return       iterator(*this, scaleData.end());    }
    const_iterator  begin()  const { return const_iterator(*this, scaleData.begin());  }
    const_iterator  end()    const { return const_iterator(*this, scaleData.end());    }
  };

}

#endif // _LEAR_PERSPECTIVE_SCALE_PYRAMID_H_


