#ifndef _LEAR_RHOG_DENSE_H_
#define _LEAR_RHOG_DENSE_H_

#include <iostream>
#include <algorithm>

#include <blitz/array.h>
#include <blitz/tinyvec.h>

#include <lear/io/biostream.h>

#include <lear/cvision/phistogram.h>
#include <lear/cvision/iprocessor.h>
#include <lear/cvision/dnormalizer.h>

namespace lear {

  /**
   * Compute Rectangular Histogram of Oriented Gradients over image gradient.
   *
   * Different image normalizer and image preprocessor can be supplied.
   */
  class RHOGDense {
  public:
    typedef IProcessor::RealType                    RealType; 

  protected:
    typedef lear::PrecisionHistogram<RealType,RealType,3>    
    HistogramType;

  public:

    enum {N=2};
    typedef blitz::TinyVector<int,N>                IndexType;
    typedef IProcessor::InfoType                    Preprocessor;

    /// feature vector dimension 
    enum {Rank = HistogramType::Rank};

    /// feature vector element type
    typedef HistogramType::HistElementType          ElemType;

    /// feature array type
    typedef HistogramType::HistType                 FeatType;

    /// feature vector array index type
    typedef HistogramType::IndexType                FeatIndexType;

    /**
     * Initializes RHOGDense descriptor.
     *
     * It owns preprocessor and normalizer pointers and ensures that they are
     * freed.
     */
    RHOGDense(IndexType const cellsize_,
              IndexType const numcell_,
              IndexType const stride_,
              const int orientbin_,
              RealType const wtscale_,
              bool const semicirc_,
              const IProcessor *p,
              const DescNormalizer<RealType> *n);
    ~RHOGDense()
    {
      delete processor_;
      delete normalizer_;
    }

    template<class PixelType>
    Preprocessor preprocess(blitz::Array<PixelType,N> const &image) const 
    { return (*processor_)(image); }

    /// Only changes hist_ and tmag_ variables. Rest all remains constant.
    FeatType& operator() (IndexType const point, Preprocessor const &p) const ;

    int       size()        const { return descsize_; }
    int       length()      const { return descsize_; }
    int       orientbin()   const { return orientbin_; }
    IndexType extent()      const { return extent_; }
    IndexType stride()      const { return stride_; }
    IndexType numcell()     const { return numcell_; }
    bool      semicirc()    const { return semicirc_; }
    RealType  weightScale() const { return wtscale_; }

    IProcessor               const *processor()  const { return processor_; }
    DescNormalizer<RealType> const *normalizer() const { return normalizer_; }

    void print(lear::BiOStream& o) const ;

    void print(std::ostream& o) const ;

  protected:
    /// cell size i.e. position shift tolerance size (in pixels)
    IndexType cellsize_; 

    /// number of cells or numcell_ bin in the descriptor
    IndexType numcell_;

    /// optimal stride_
    IndexType stride_;

    /// number of orientation bins
    int orientbin_; 
    
    /// scale for weighting function span
    RealType wtscale_;
        
    /// if true, orientation bin range is 0-180 degree range, else 0-360
    bool semicirc_;

    /// image pre-processor to use
    IProcessor              const *processor_;

    /// histogram normalizer to use
    DescNormalizer<RealType> const *normalizer_;

    /// descriptor size (#cells * #orientbins)
    int descsize_;

    /// filter bound 
    IndexType extent_, ubound_; 
    
    /// weight_ing array
    blitz::Array<RealType,N> weight_;

    /// Histogram extent: max values for (x,y,orient) in a block
    ///   h_extent = (blockwidth, blockheight, (semicirc_+1)*180.)
    blitz::TinyVector<RealType,3>   h_extent;
    /// Size of bins in (x,y,orient) for the histogram.
    ///   h_bandwidth = (cellsize x, cellsize y, (semicirc_+1)*180./orientbin_)
    blitz::TinyVector<RealType,3>   h_bandwidth;

    /// temp array to save filter response
    mutable Preprocessor::MagAType tmag_;

    mutable HistogramType hist_;
    
  };

  BiOStream& operator<<(BiOStream& o, RHOGDense const &d);

}

std::ostream& operator<<(std::ostream& o, lear::RHOGDense const &d);


#endif // _LEAR_RHOG_DENSE_H_
