#ifndef _LEAR_WIN_DESCRIPTOR_H_
#define _LEAR_WIN_DESCRIPTOR_H_

#include <list>
#include <memory>

#include <blitz/tinyvec.h>
#include <lear/exception.h>

#include <lear/io/biostream.h>
#include <lear/cvision/densegrid.h>
#include <lear/cvision/cachedesc.h>
#include <lear/cvision/classify.h>

#include <lear/cvision/rhogdense.h>
#include <lear/cvision/iprocessor.h>

// Needs to be modified so that make install works with the header files
#include "../../lib/cfg.h"

#ifdef HAVE_CUDA
#include <pdcuda/imresize.h>
#include <pdcuda/arrays.h>
#endif

namespace lear {

  //---------------------------------------------------------------------------

  /** WinDescriptor 
  *  A collection of RHOG descriptors for a single image window + metadata about
  *  the window.
  */
  class WinDescriptorBase {  
  public:
    typedef IProcessor::RealType                RealType; 

    typedef IProcessor::GrayType                GrayType;
    typedef IProcessor::RGBType                 RGBType;

    typedef IProcessor::GrayImage               GrayImage;
    typedef IProcessor::RGBImage                RGBImage;

    typedef RGBImage                            ImageType;

    typedef RHOGDense                           DescType;
    typedef std::list<DescType*>                DescCont; 
 
    typedef DescType::ElemType                  ElemType;
    typedef blitz::TinyVector<int,2>            IndexType;

    /// Combined big feature type for all points on the grid
    typedef blitz::Array<ElemType, 1>           FeatType;
    typedef blitz::Array<IndexType,1>           IndexArray;

  public:
    WinDescriptorBase(IndexType const t_extent_);
    virtual ~WinDescriptorBase(); 

    /// Latch in the image to be examined.  Call this first.
    virtual void     setimage(GrayImage const &image) = 0;
    virtual void     setimage(RGBImage  const &image) = 0;

    /// Return a rescaled version of the input image.  This is useful
    /// if the descriptor has a way of doing fast resizes and/or if
    /// the user wants a resized image that uses the same algorithm as
    /// preprocess uses.
    virtual void     getimage(RGBImage &image, IndexType const &sz) = 0;

    /// Do any necessary preprocessing, including cropping and
    /// resampling.  After each call to setimage, preprocess must be
    /// called before calling compute the first time.
    virtual void     preprocess(IndexType topleft, 
                                IndexType extent,
                                IndexType newsize) = 0;
    void             preprocess(IndexType newsize) { preprocess(IndexType(0,0), image_extent(), newsize); }
    void             preprocess(int w, int h)      { preprocess(IndexType(w,h)); }

    /// Compute a window descriptor feature at the given position in
    /// the (possibly) cropped and resampled image.  setimage and
    /// preprocess must be called before compute every time the input
    /// image changes or the desired cropping and/or resampling
    /// changes.
    virtual FeatType compute(IndexType gridTopLeft) const = 0; // FUTURE: use an expression template to avoid a memcpy

    virtual size_t   length() const = 0; // total length of the window's descriptor
    IndexType extent()       const { return extent_; } // size of the descriptor's window, in pixels
    IndexType image_extent() const { return image_extent_; } // size of the input image, before scaling, in pixels

    virtual void print(std::ostream&    o) const = 0;
    virtual void print(lear::BiOStream& o) const = 0;

  protected:
    /// original input image size
    IndexType image_extent_;

    /// window size
    IndexType const extent_;

    lear::FileHeader header() const;
  };
   
  //---------------------------------------------------------------------------

  class WinDescriptor : public WinDescriptorBase 
  {
  public:
    typedef DenseGrid<2>::GridType              GridType;
    typedef std::list<GridType>                 GridCont; 

    typedef std::list<DescType::Preprocessor>   Preprocessor;

    /// Singular feature type, i.e. feature type at each point
    typedef DescType::FeatType         SingFeatType;

  public:
    WinDescriptor(IndexType t_extent_, DescCont const &t_desc_,
                  GridCont  const &t_grid_, int cachesize = 16);

    virtual void     setimage(GrayImage const &image);
    virtual void     setimage(RGBImage  const &image);
    virtual void     getimage(RGBImage &image, IndexType const &sz);
    virtual void     preprocess(IndexType topleft,  // top-left corner for cropping
                                IndexType extent,   // cropped image size
                                IndexType newsize); // size after cropping, then resampling
    virtual FeatType compute(IndexType gridTopLeft) const;
    virtual size_t   length() const { return length_; } // TODO: hunt down usages and cache return val

    virtual void print(std::ostream&    o) const;
    virtual void print(lear::BiOStream& o) const;

  private:
    typedef CacheDesc<ElemType,2>               CacheType;
    typedef std::list< CacheType >              CacheCont;

    /// descriptor array object.  These are actual descriptors
    /// encapsulated with code to compute them.
    DescCont const desc_;

    /// compute descriptor over points in this grid
    GridCont const grid_;

    Preprocessor preprocessor;

    /// output description length
    int initlength_, length_, numItem_;

    IndexArray          indexrange_;

    mutable CacheCont   cache_;

    RGBImage  currRgb;
    GrayImage currGray;
    bool      rgbIsCurr;

  private:
    std::string title() const { return "Win Descriptor ::       "; }

    template <class PixelType>
    Preprocessor& template_preprocess(blitz::Array<PixelType,2> const &image, 
                                      IndexType topleft, IndexType extent,
                                      IndexType newsize);
    struct DescOp {
      typedef DescType::Preprocessor Processor;
      DescType const *d;
      Processor const &p;

      DescOp(DescType const *d, Processor const &p) : d(d),p(p){}
      WinDescriptorBase::ElemType* operator()(IndexType loc) const {
        return (*d)(loc,p).data();
      }
      size_t size() const { return d->size(); }
    };
  };
   
  //---------------------------------------------------------------------------

#ifdef HAVE_CUDA
  class WinDescriptorCuda : public WinDescriptorBase {
  public:
    typedef  blitz::Array<RealType,5> BlockDescriptors;

  public:
    WinDescriptorCuda(IndexType                           t_extent_, 
                      WinDescriptorBase::DescCont  const &t_desc_,
                      IndexType                           fineStride_,
                      bool                                quantizePos,
                      bool                                useLanczos=true);

    virtual void     setimage(GrayImage const &image);
    virtual void     setimage(RGBImage  const &image);
    virtual void     getimage(RGBImage &image, IndexType const &sz);
    virtual void     preprocess(IndexType topleft, 
                                IndexType extent,
                                IndexType newsize);
    virtual FeatType compute(IndexType gridTopLeft) const;

    virtual size_t   length() const;

    virtual void print(std::ostream&    o) const;
    virtual void print(lear::BiOStream& o) const;

    // Classify all windows.  Has the same preconditions as compute.
    // NOTE: this array is ROW MAJOR indexed: access elements as
    // (y,x), not (x,y) as nearly everything else in the lear
    // namespace.
    blitz::Array<RealType,2> 
    classifyAllRowMajor(LinearClassify const &c) const;

    IndexType fineStride() const { return fineStride_; }

  private:
    IndexType                     const fineStride_;
    bool                                quantizePos;
    RealType                            epsilon;
    RealType                            maxvalue;
    RealType                            epsilon2;
    pdcuda::ScopedDeviceArray<float>    dBD;

    BlockDescriptors            mutable hBD; // block descriptors 
    bool                        mutable hBDdirty;
    int                                 hBDw, hBDh;

    static int                    const lanczosWindowRadius = 2;
    bool                                useLanczos;
    std::auto_ptr<pdcuda::ImageResizer> resizer;

  private:
    std::string title() const { return "Win Descriptor (CUDA) ::"; }
  };
#endif

}

//-----------------------------------------------------------------------------

std::ostream& operator<<( std::ostream& o, lear::WinDescriptorBase const &w) ;
lear::BiOStream& operator<<( lear::BiOStream& o, lear::WinDescriptorBase const &w) ;

#endif // _LEAR_WIN_DESCRIPTOR_H_

