/** 
* @file
* @brief Defines various painting operations
*/


#ifndef _LEAR_PIMAGE_H_
#define _LEAR_PIMAGE_H_

/* standard headers */
#include <lear/windowsystem.h>
#include <string>

namespace lear {

  class Painter;
  /**
  * @brief PImage class. Mainly for use by Painter
  */
  class PImage {
  public:
    PImage() : image()
#ifdef WIN32
      ,imageIsValid(false) 
#endif
    {}

    PImage(std::string const &imagename) : image()
#ifdef WIN32
      ,imageIsValid(false)
#endif
    {
      load(imagename);
    }

    PImage(int w, int h, int depth=32);

    void load(std::string const &imagename);

    int width() const {
#ifndef WIN32
      if (image) {
        imlib_context_set_image(image);
        return imlib_image_get_width(); 
      }else 
        return -1;
#else
      return imageIsValid ? image.GetWidth() : -1;
#endif
    }

    int height() const {
#ifndef WIN32
      if (image) {
        imlib_context_set_image(image);
        return imlib_image_get_height();
      }else 
        return -1;
#else
      return imageIsValid ? abs(image.GetHeight()) : -1;
#endif
    }

    PImage crop( int x, int y, int width, int height){
#ifndef WIN32
      if (image) {
        imlib_context_set_image(image);
        return PImage( imlib_create_cropped_image(x, y, width, height));
      }else 
        return PImage();
#else
      if (!imageIsValid) return PImage();
      CImage tmp;
      if (!tmp.Create(width, -height, 32)) return PImage();
      if (!image.Draw(tmp.GetDC(), 0,0,width,height, x,y,width,height)) return PImage();
      return PImage(tmp);
#endif
    }

    DATA32* data(){
#ifndef WIN32
      if (image) {
        imlib_context_set_image(image);
        return imlib_image_get_data();
      } else {
        return 0;
      }
#else
      if (!imageIsValid) return NULL;
      ThrowException("We need to convert to a top-down row major ARGB "
                     "format (with the most significant byte being the "
                     "alpha channel).");
      return (DATA32*)image.GetBits();
#endif
    }

    // data must be the pointer returned by DATA32* data() function
    bool data(DATA32* data) {
#ifndef WIN32
      if (image) {
        imlib_context_set_image(image);
        imlib_image_put_back_data(data);
      }
      return data;
#else
      if (!imageIsValid || !data) return false;
      ThrowException("We need to convert from a top-down row major ARGB "
                     "format (with the most significant byte being the "
                     "alpha channel).");
      memcpy(image.GetBits(), data, width()*height()*sizeof(DATA32));
      return true;
#endif
    }

    int depth() const { return 32; }
    int rows() const { return width(); }
    int cols() const { return height(); }
#ifndef WIN32
    bool valid() const {return static_cast<bool>(image);}
#else
    bool valid() const {return imageIsValid;}
#endif

  private:
#ifndef WIN32
    PImage(Imlib_Image image):image(image){}
    Imlib_Image image;
#else
    PImage(ATL::CImage &image) { 
      this->image.Attach(image);
      imageIsValid = true;
    }
    ATL::CImage image;
    bool imageIsValid;
#endif
    friend class Painter;
  };

} // namespace lear

#endif // #ifdef _LEAR_PIMAGE_H_

