#include <lear/cvision/classify.h>
#include <lear/exception.h>
#include <stdio.h>

namespace lear {

  typedef lear::IProcessor::RealType RealType;

  LinearClassify::LinearClassify(std::string &modelfile, const int verbose) 
    : length_ (0), linearwt_(0),  linearbias_(0)
  {
    if (verbose > 2) std::cout << "Reading model file: " << modelfile;

    FILE *modelfl;
    if ((modelfl = fopen (modelfile.c_str(), "rb")) == NULL)
      { ThrowException("Unable to open the modelfile"); }

    char version_buffer[10];
    if (!fread (&version_buffer,sizeof(char),10,modelfl))
      { ThrowException("Unable to read version"); }

    if(strcmp(version_buffer,"V6.01")) {
      ThrowException("Version of model-file does not match version of svm_classify!"); 
    }
    /* read version number */
    int version = 0;
    if (!fread (&version,sizeof(int),1,modelfl)) {
      ThrowException("Unable to read version number"); 
    }
    if (version < 200) {
      ThrowException("Does not support model file compiled for light version");
    }

    int kernel_type;
    fread(&(kernel_type),sizeof(int),1,modelfl);

    {// ignore these
      int poly_degree;
      fread(&(poly_degree),sizeof(int),1,modelfl);

      double rbf_gamma;
      fread(&(rbf_gamma),sizeof(double),1,modelfl);

      double  coef_lin;
      fread(&(coef_lin),sizeof(double),1,modelfl); 
      double coef_const;
      fread(&(coef_const),sizeof(double),1,modelfl);

      int l;
      fread(&l,sizeof(int),1,modelfl);
      char* custom = new char[l];
      fread(custom,sizeof(char),l,modelfl);
      delete[] custom;
    }

    int totwords;
    fread(&(totwords),sizeof(int),1,modelfl);

    {// ignore these
      int totdoc;
      fread(&(totdoc),sizeof(int),1,modelfl);

      int sv_num;
      fread(&(sv_num), sizeof(int),1,modelfl);
    }

    fread(&linearbias_, sizeof(double),1,modelfl);

    if(kernel_type == 0) { /* linear kernel */
      /* save linear wts also */
      linearwt_ = new double[totwords+1];
      length_ = totwords;
      fread(linearwt_, sizeof(double),totwords+1,modelfl);
    } else {
      ThrowException("Only supports linear SVM model files");
    }
    fclose(modelfl);

    if (verbose > 2) 
      std::cout << " Done" << std::endl;
  }

  LinearClassify::LinearClassify(LinearClassify const &o) :
    length_(o.length_), linearbias_(o.linearbias_)
  {
    linearwt_ = new double[length_];
    std::copy(o.linearwt_, o.linearwt_+o.length_, linearwt_);
  }

  LinearClassify& LinearClassify::operator=(LinearClassify const &o) 
  {
    if (&o != this) {
      if (linearwt_) 
        delete[] linearwt_;

      length_=o.length_; 
      linearbias_=o.linearbias_;

      linearwt_ = new double[length_];
      std::copy(o.linearwt_, o.linearwt_+o.length_, linearwt_);
    } 
    return *this;
  }

  double LinearClassify::operator()(double const *desc) const 
  {
    double sum = 0;
    for (int i= 0; i< length_; ++i) 
      sum += linearwt_[i]*desc[i]; 
    return sum - linearbias_;
  }

  float LinearClassify::operator()(float const *desc) const 
  {
    double sum = 0;
    for (int i= 0; i< length_; ++i) {
      sum += linearwt_[i]*desc[i]; 
    }
    return sum - linearbias_;
  }

  blitz::Array<RealType,1> LinearClassify::classifier() const
  {
    blitz::Array<RealType,1> c(length_);
    for (int i=0; i<length_; i++) {
      c(i) = linearwt_[i];
    }
    return c;
  }

  extern const double         PERSON_WEIGHT_VEC[];
  extern const int            PERSON_WEIGHT_VEC_LENGTH;
  LinearClassify::LinearClassify() 
  {
    linearbias_ = 6.6657914910925990525925044494215;
    length_ = PERSON_WEIGHT_VEC_LENGTH;
    linearwt_ = new double[length_];
    std::copy(PERSON_WEIGHT_VEC, PERSON_WEIGHT_VEC+PERSON_WEIGHT_VEC_LENGTH, linearwt_);
  }
#include "persondetectorwt.tcc"

};
