/* standard headers */
#include <iostream>
#include <iomanip>
#include <sstream>
#include <vector>
#include <lear/image/painter.h>
#include <lear/util/fileutil.h>
#include <lear/exception.h>

#include "cfg.h"
#ifdef HAVE_OPENMP
#  include <boost/thread/thread.hpp>
#  include <boost/thread/mutex.hpp>
boost::mutex imlibMutex; // imlib is not threadsafe
#  define SYNCHRONIZED \
  boost::mutex::scoped_lock imlibLock(imlibMutex)
#else
#  define SYNCHRONIZED
#endif

#ifdef WIN32
#pragma warning (disable : 4800) // BOOL->bool conversion
#endif

using namespace lear;

#define T(test) if (!(test)) throw Exception("FAILED: " #test)

Painter::Painter(lear::IProcessor::RGBImage const &img) : 
  fill(false)
{
#ifndef WIN32
  int w = img.extent()[0];
  int h = img.extent()[1];
  {
    SYNCHRONIZED;
    image = imlib_create_image(w, h);
    if (!image){
      ThrowException("Unable to create " << std::setw(4) << w << " X " 
                     << std::setw(4) << h << " pixmap.");
    }
    imlib_context_set_image(image);
    DATA32 *data = imlib_image_get_data();
    for (int y=0; y<h; y++) {
      for (int x=0; x<w; x++) {
        blitz::TinyVector<float,3> const &in = img(x,y);
#define CVT(x) ((x)<0)?0:((x)>255?255:int(x))
        *data++ = ((CVT(in[0]) << 16) | // red
                   (CVT(in[1]) <<  8) | // green
                   (CVT(in[2])));       // blue
#undef CVT
      }
    }
    imlib_image_put_back_data(data - w*h);
  }
#else
# error "NOT YET IMPLEMENTED"
#endif
  setfont();
  setColor(0,0,0);
  noBrush();
}

Painter::Painter(std::string const &imagename/*, ColorMode mode*/) :
  fill(false)
{
  load(imagename);
  setfont();
  setColor(0,0,0);
  noBrush();
}

Painter::Painter(int w, int h, int depth/*, Optimization optimization*/):
fill(false)
{
#ifndef WIN32
  { 
    SYNCHRONIZED;
    image = imlib_create_image(w,h);
  }
  if (!image){
    ThrowException("Unable to create " << std::setw(4) << w << " X " 
                   << std::setw(4) << h << " pixmap at depth " 
                   << std::setw(3) << depth);
  }
#else
  if (!image.Create(w,-h,depth*8)) {
    ThrowException("Unable to create " << std::setw(4) << w << " X " 
                   << std::setw(4) << h << " pixmap at depth "
                   << std::setw(3) << depth );
  }
  if (!dc.Attach(image.GetDC())) {
    image.ReleaseDC();
    ThrowException("Unable to attach to created " << std::setw(4) << w << " X " 
                   << std::setw(4) << h << " pixmap at depth " 
                   << std::setw(3) << depth);
  }
#endif
  setfont();
  setColor(0,0,0);
  noBrush();
}

bool Painter::load(std::string const &imagename/*, ColorMode mode*/) {
#ifndef WIN32
  SYNCHRONIZED;
  // an image handle 
  image = imlib_load_image(imagename.c_str());;
  // if the load was successful
  if (!image){
    ThrowException("Unable to load image " + imagename);
  }
  return true;
#else
  CString n(imagename.c_str());
  HRESULT hr = image.Load(n);
  if (!SUCCEEDED(hr)) {
    ThrowException("Unable to load image " + imagename);
  }
  dc.Detach(); // TODO: needed?
  if (!dc.Attach(image.GetDC())) {
    image.ReleaseDC();
    ThrowException("Unable to attach to loaded image " + imagename);
  }
  return true;
#endif
}

Painter::~Painter() {
#ifndef WIN32
  if (image) {
    SYNCHRONIZED;
    imlib_context_set_image(image);
    imlib_free_image();
  }
  if (font) {
    SYNCHRONIZED;
    imlib_context_set_font(font);
    imlib_free_font();
  }
#else
  image.ReleaseDC();
  dc.Detach();
  // Nothing to do: MFC classes clean up after themselves
#endif
}

#ifdef WIN32
static bool setFontFace(CFont &font, LOGFONT &lf, std::string const &name) {
  memset(lf.lfFaceName, 0, LF_FACESIZE*sizeof(TCHAR));
  size_t const n = LF_FACESIZE < name.size() ? LF_FACESIZE : name.size();
  for (size_t i=0; i<n; i++) {
    lf.lfFaceName[i] = name[i];
  }
  return (bool)font.CreateFontIndirect(&lf);
}
#endif

#define TRYFONT(fontname) \
  if (!font) { \
    /*fprintf(stderr, "Trying %s\n", fontname);*/ \
    fontstr = fontname; \
    font = imlib_load_font(fontstr.c_str()); \
  }

void Painter::setfont(std::string defaultfont) {
#ifndef WIN32
  SYNCHRONIZED;
  imlib_set_font_cache_size(512 * 1024);
  /* add the ./ttfonts dir to our font path - you'll want a
  * notepad.ttf in that dir for the text to display */
  imlib_add_path_to_font_path("/usr/X11R6/lib/X11/fonts/TTF");
  imlib_add_path_to_font_path("/usr/share/fonts/truetype/freefont");
  imlib_add_path_to_font_path("/usr/share/fonts/truetype/msttcorefonts");
  imlib_add_path_to_font_path(".ttf");
  
  std::string fontstr=defaultfont+"/10";
  font = imlib_load_font(fontstr.c_str());

  if (!font) {
    int numfonts;
    char ** fontlist = imlib_list_fonts(&numfonts);
    if (numfonts) {
      TRYFONT("aerial/10"); TRYFONT("arial/10");
      TRYFONT("verdana/10");
      TRYFONT("times/10");
      TRYFONT("monotype/10");
      TRYFONT("morpheus/10");
      TRYFONT("notepad/10");
      TRYFONT("cinema/10");
      TRYFONT("grunge/10");
      TRYFONT(fontlist[0]);
    }
  }
  if (!font) {
    std::cerr << "Could not find aerial, times, verdana, monotype, morpheus, notepad, cinema, grunge fonts."
      " No fonts loaded" << std::endl;
  }
#else
  if (((HFONT)font != NULL)) {
    T(dc.SelectStockObject(DEVICE_DEFAULT_FONT) != NULL);
    font.DeleteObject();
  }

  bool found = false;
  for (int quality=ANTIALIASED_QUALITY; quality>=DEFAULT_QUALITY; quality--) {
    LOGFONT fontDef = { 10, 0, 0,0, FW_DONTCARE, 0,0,0, ANSI_CHARSET, 
      OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, 
      quality, DEFAULT_PITCH, L"" };
    if (setFontFace(font, fontDef, defaultfont) ||
      setFontFace(font, fontDef, "Lucida Console") ||
      setFontFace(font, fontDef, "Arial") ||
      setFontFace(font, fontDef, "Verdana") ||
      setFontFace(font, fontDef, "Times New Roman") ||
      setFontFace(font, fontDef, "Times")) 
    {
      found = true;
      break;
    }
  }

  if (!found) {
    std::cerr << "Could not find " << defaultfont << 
      ", Lucida Console, Arial, Verdana, Times New Roman, or Times fonts."
      " No fonts loaded" << std::endl;
  }

  // Ignoring the return value for now.  NULL can mean the font is already 
  // selected, AFAIK: so it's not meaningful to check the return value.
  T(dc.SelectObject(&font) != NULL);
#endif
}

bool Painter::save(std::string const &imagename) const {
#ifndef WIN32
  if (!image) return false;
  SYNCHRONIZED;
  imlib_context_set_image(image);
  std::string ext = lear::extension(imagename);
  if (ext.length()) {
    imlib_image_set_format(ext.c_str());
  } else {
    imlib_image_set_format("png");
  }

  Imlib_Load_Error reterr;
  imlib_save_image_with_error_return(imagename.c_str(), &reterr);

  if (reterr != IMLIB_LOAD_ERROR_NONE) {
    ThrowException(std::string("Unable to save file: ") + imagename);
  } 
  return true;
#else
  CString n(imagename.c_str());
  HRESULT hr = image.Save(n);
  if (!SUCCEEDED(hr)) {
    ThrowException(std::string("Unable to save file: ") + imagename);
  }
  return true;
#endif
}


void Painter::drawPoint(const int xt, const int yt, const int size, 
                        const PointStyle style) 
{
#ifndef WIN32
  if (!image)  return;
  SYNCHRONIZED;
  imlib_context_set_image(image);

  const int x = static_cast<int> (xt);
  const int y = static_cast<int> (yt);

  switch(style) {
  case Plus:
    imlib_image_draw_line(x-size, y, x+size, y,0);
    imlib_image_draw_line(x, y-size, x, y+size,0);
    break;
  case Cross:
    imlib_image_draw_line(x-size, y -size, x+size, y+size,0);
    imlib_image_draw_line(x-size, y +size, x+size, y-size,0);
    break;
  case Dot:
    imlib_image_draw_line(x, y, x, y,0);
    break;
  }
#else
  // TODO: do I need to do something with a 'CBrush brush(RGB(0,0,0));'?
  switch (style) {
  case Plus:
    drawLine(xt-size,yt, xt+size,yt);
    drawLine(xt,yt-size, xt,yt+size);
    break;
  case Cross:
    drawLine(xt-size,yt-size, xt+size,yt+size);
    drawLine(xt-size,yt+size, xt+size,yt-size);
    break;
  case Dot:
    image.SetPixel(xt, yt, RGB(0,0,0));
    break;
  }
#endif
}

void Painter::drawLine(const int x1, const int y1, const int x2, const int y2) 
{
#ifndef WIN32
  if (!image) return;
  SYNCHRONIZED;
  imlib_context_set_image(image);
  imlib_image_draw_line(
    static_cast<int>(x1),
    static_cast<int>(y1),
    static_cast<int>(x2),
    static_cast<int>(y2),0);
#else
  dc.MoveTo(x1,y1);
  dc.LineTo(x2,y2);
#endif
}

//void Painter::fillRect(const int x, const int y, const int w, const int h) {
//    if (!image)  return;
//    imlib_context_set_image(image);
//
//        imlib_image_fill_rectangle(
//            static_cast<int>(x), static_cast<int>(y),
//            static_cast<int>(w), static_cast<int>(h));
//}
//
void Painter::drawRect(const int x, const int y, const int w, const int h) 
{
#ifndef WIN32
  if (!image)  return;
  SYNCHRONIZED;
  imlib_context_set_image(image);

  if (fill) {
    imlib_image_fill_rectangle(
      static_cast<int>(x), static_cast<int>(y),
      static_cast<int>(w), static_cast<int>(h));
  } else{
    imlib_image_draw_rectangle(
      static_cast<int>(x), static_cast<int>(y),
      static_cast<int>(w), static_cast<int>(h));
  }
#else
  if (fill) {
    RECT const rect = { x, y, x+w, y+h };
    dc.FillRect(&rect, &brush);
  } 
  dc.MoveTo(x,   y);
  dc.LineTo(x+w, y);
  dc.LineTo(x+w, y+h);
  dc.LineTo(x,   y+h);
  dc.LineTo(x,   y);
#endif
}

void Painter::setBrush(const int r, const int g, const int b) 
{ 
#ifndef WIN32
  setColor(r,g,b);
#else
  setColor(r,g,b); 
  COLORREF const color = RGB(r,g,b);
  LOGBRUSH lb;
  if (((HBRUSH)brush == NULL) || (brush.GetLogBrush(&lb) == 0)) {
    T(brush.CreateSolidBrush(color));
  } else if (lb.lbColor != color) {
    T(dc.SelectStockObject(NULL_BRUSH) != NULL);
    brush.DeleteObject();
    T(brush.CreateSolidBrush(color) != NULL);
  }
  T(dc.SelectObject(&brush) != NULL);
  T(dc.SetBkColor(color) != 0x80000000);
  dc.SetBkMode(OPAQUE);
#endif
  fill = true;
}

void Painter::noBrush() {
  fill = false;
#ifdef WIN32
  T(dc.SelectStockObject(NULL_BRUSH) != NULL);
  dc.SetBkMode(TRANSPARENT);
#endif
}

void Painter::drawEllipse(const int x, const int y, 
                          const int l1, const int l2, const double angle)
{
  if (angle != 0) ThrowException("rotation not implemented");
#ifndef WIN32
  if (!image)  return;
  SYNCHRONIZED;
  imlib_context_set_image(image);

  if (fill) {
    imlib_image_fill_ellipse(
      static_cast<int>(x), static_cast<int>(y),
      static_cast<int>(l1), static_cast<int>(l2));
  } else {
    imlib_image_draw_ellipse(
      static_cast<int>(x), static_cast<int>(y),
      static_cast<int>(l1), static_cast<int>(l2));
  }
#else
  T(dc.Ellipse(x-l1, y-l2, x+l1, y+l2));
#endif
}


void Painter::drawImage(PImage const &image_, 
                        int dx, int dy, int dw, int dh, 
                        int sx, int sy, int sw, int sh) 
{
#ifndef WIN32
  if (!image)  return;
  SYNCHRONIZED;
  imlib_context_set_image(image_.image);

  if (sw ==-1) sw = imlib_image_get_width(); 
  if (sh ==-1) sh = imlib_image_get_height(); 

  imlib_context_set_image(image);
  if (dw ==-1) dw = sw;
  if (dh ==-1) dh = sh;

  imlib_blend_image_onto_image(image_.image,0,
    static_cast<int>(sx),
    static_cast<int>(sy),
    static_cast<int>(sw),
    static_cast<int>(sh),
    static_cast<int>(dx),
    static_cast<int>(dy),
    static_cast<int>(dw),
    static_cast<int>(dh));
#else
  if (sw==-1) sw = image_.width();
  if (sh==-1) sh = image_.height();
  if (dw==-1) dw = image.GetWidth();
  if (dh==-1) dh = abs(image.GetHeight());

  T(StretchBlt(image_.image.GetDC(), dx,dy,dw,dy, 
               this->image.GetDC(), sx,sy,sw,sh, SRCCOPY));

  image_.image.ReleaseDC();
  this->image.ReleaseDC();
#endif
}

bool Painter::drawText(std::string const &text, const int x, const int y)
{
#ifndef WIN32
  if (!image)  return false;
  SYNCHRONIZED;
  imlib_context_set_image(image);

  if (font) {
    imlib_context_set_font(font);
    //int text_w, text_h;
    //imlib_get_text_size(text.c_str(), &text_w, &text_h); 
    imlib_text_draw(
      static_cast<int>(x),
      static_cast<int>(y), text.c_str());
    return true;
  }
  return false;
#else
  CString txt(text.c_str());
  return dc.TextOutW(x,y, txt, (int)text.size());
#endif
}

bool Painter::getTextSize(std::string const &text, int& width, int &height)
{
#ifndef WIN32
  if (font) {
    SYNCHRONIZED;
    imlib_context_set_font(font);
    imlib_get_text_size(text.c_str(), &width, &height); 
    return true;
  }
  return false;
#else
  CString txt(text.c_str());
  CSize ext = dc.GetTextExtent(txt);
  width  = ext.cx;
  height = ext.cy;
  return true;
#endif
}

void Painter::setColor(const int r, const int g, const int b) {
#ifndef WIN32
  if (!image)  return;
  SYNCHRONIZED;
  imlib_context_set_image(image);

  int a = 255;
  imlib_context_set_color(
    static_cast<int>(r), static_cast<int>(g),
    static_cast<int>(b), static_cast<int>(a));
#else
  COLORREF const color = RGB(r,g,b);
  LOGPEN lp;
  if (((HPEN)pen == NULL) || (pen.GetLogPen(&lp) == 0)) {
    T(pen.CreatePen(PS_SOLID, 1, color));
  } else if (lp.lopnColor != color) {
    T(dc.SelectStockObject(NULL_PEN) != NULL);
    pen.DeleteObject();
    T(pen.CreatePen(PS_SOLID, 1, color));
  }
  T(dc.SelectObject(&pen) != NULL);
  dc.SetTextColor(RGB(r,g,b));
#endif
}

void Painter::fillBackground() {
#ifndef WIN32
  if (!image)  return;
  SYNCHRONIZED;
  imlib_context_set_image(image);
  // TODO: I think we need to switch to fill mode here
  drawRect(0,0, imlib_image_get_width(), imlib_image_get_height()); 
#else
  int const ctx = dc.SaveDC();
  T(ctx != 0);
  dc.FillSolidRect(0,0, image.GetWidth(),image.GetHeight(), dc.GetBkColor());
  T(dc.RestoreDC(ctx));
#endif
}
