#include <pdcuda/imresize.h>

#include <stdio.h>
#include <iostream>
#define _USE_MATH_DEFINES
#include <math.h>

#include <cutil/cutil_inline.h>
#include <cutil/error_checker.h>

#include <pdcuda/imtransfer.h>
#include <pdcuda/math.cuh>
#include <pdcuda/threads.h>

using namespace std;
using namespace pdcuda;

// Access must be guarded with t2_mut.  If we want to allow multiple
// streams and multiple devices, we'll need to have one texture and
// one lock per stream.
static texture<float4, 2, cudaReadModeElementType> t2;
static Mutex                                       t2_mut;

// Access must be guarded with t3_mut.  If we want to allow multiple
// streams and multiple devices, we'll need to have one texture and
// one lock per stream.
static texture<float, 3, cudaReadModeElementType> t3;
static Mutex                                      t3_mut;

__device__ inline static
float lanczos(float x, int a) {
  if (x==0) {
    return 1;
  } else if (x > a || x < -a) {
    return 0;
  } else {
    return a * sinf(M_PI * x) * sinf(M_PI * x / a) / (M_PI * M_PI * x * x);
  }
}

///////////////////////////////////////////////////////////////////////////////
// ImageResizerLanczos2D

__host__
pdcuda::
ImageResizerLanczos2D::ImageResizerLanczos2D(float const *data, 
                                             int w, int h,
                                             int windowRadius,
                                             bool dataOnHost, 
                                             bool dataColumnMajor)
  : ImageResizer(w,h), windowRadius(windowRadius), data(NULL)
{
  this->data = rgbToCudaArray2d(data, w, h, dataOnHost, dataColumnMajor);
}

__host__
pdcuda::
ImageResizerLanczos2D::~ImageResizerLanczos2D() 
{
  if (data) {
    cutilSafeCall(cudaFreeArray(data));
    data = NULL;
  }
}

// Profiling data (up to 2x upsampling for a 1920x1080 image with 8x8
// blocks on a GTX 280 / Core i7 965):
//
//   52,308.77 us GPU time
//       50%   occupancy
//       44    bytes shared memory per block
//       31    registers per thread (limiting factor)
//        8    warps / multiprocessor (limiting factor)
//  
//   branches:           3,931,200
//   divergent branches:    86,400 (2.2%)
//   bank conflicts:           N/A
//
// To increase the occupancy, we have a few possibilities (by either
// modifying the code to use fewer registers and/or changing
// blockDim).  I think it will be very hard to reduce the register
// count below 19.  Using --maxrregcount to automatically reduce the
// register count results in worse performance, so I think we'd have
// to recode this in PTX assembly directly to get a benefit.
//
//   threads/block reg/thread occupancy
//   ------------- ---------- --------- 
//        32           31        25%
//
//        64           31        50%  (current)
//        64            0        50%
//
//        96           31        47%
//        96           26        56%
//        96           21        75%
//        96            0        75%
//
//       128           31        50%
//       128           24        63%
//       128           20        75% 
//       128           16       100%
//
// We've tested some slightly modified versions of the code to
// discover what is taking all the time.  This is for a series of
// resizing using upscaling ratios (new:orig) of 1:1, 1:0.87, 1:0.76,
// 1:0.66, 1:0.57, and 1:0.5 with a 1920x1080 source image (tested
// with testcudapipeline:r708):
//
//   28ms: writing results to gmem (coalesced simulation), no gmem reads
//   37ms: writing results to gmem (1 packed pixel / thread), no gmem reads
//   37ms: ...and computing indexing variables
//   80ms: ...and running a spatial box filter with full gmem reads
//   82ms: ...and normalizing the filtered values
//  197ms: ...switching to a Lanczos filter
//
// If compiled with --fast-math, here are the results that differ
//
//   37ms: ...and computing indexing variables
//   71ms: ...and running a spatial box filter with full gmem reads
//   72ms: ...and normalizing the filtered values
//   94ms: ...switching to a Lanczos filter
//
__global__ static
void ResizeLanczos2D_Kernel(float *dOut, 
                            int x0, int y0, int wc, int hc, int wo, int ho, 
                            int r)
{
  int   const xo = blockDim.x * blockIdx.x + threadIdx.x;
  int   const yo = blockDim.y * blockIdx.y + threadIdx.y;

  if (xo >= wo || yo >= ho) return;

  float const xi = (xo + 0.5f) * wc / wo;
  float const yi = (yo + 0.5f) * hc / ho;

  int   const xiStart = (wc>wo) ? ceilf(xi - r*wc/wo) : ceilf(xi - r);
  int   const yiStart = (hc>ho) ? ceilf(yi - r*hc/ho) : ceilf(yi - r);

  int   const nx = (wc>wo) ? floorf(2*r*wc/wo) : 2*r-1;
  int   const ny = (hc>ho) ? floorf(2*r*hc/ho) : 2*r-1;

  float const xlStart = (wc>wo) ? ((xiStart+0.5f-xi)*wo/wc):(xiStart+0.5f-xi);
  float const ylStart = (hc>ho) ? ((yiStart+0.5f-yi)*ho/hc):(yiStart+0.5f-yi);

  float const dxl = (wc>wo) ? ((float)wo)/wc : 1;
  float const dyl = (hc>ho) ? ((float)ho)/hc : 1;

  float  norm = 0;
  float4 val  = {0,0,0,0};
  {
    int   yii = yiStart;
    float yli = ylStart;
    for (int j=0; j<=ny; j++, yii++, yli+=dyl) {
      int         xii = xiStart;
      float       xli = xlStart;
      float const ly  = lanczos(yli, r);

      for (int i=0; i<=nx; i++, xii++, xli+=dxl) {
        float const lx = lanczos(xli, r);
        float const l = lx * ly;
        norm += l;
        val  += (l * tex2D(t2, x0+xii, y0+yii));
      }
    }
  }
  norm = 1 / norm;
  val.x = max(0.f, min(val.x * norm, 255.f));
  val.y = max(0.f, min(val.y * norm, 255.f));
  val.z = max(0.f, min(val.z * norm, 255.f));
  //val.w = max(0.f, min(val.w * norm, 255.f)); // not needed

  dOut[yo*wo*3 + xo*3    ] = val.x;
  dOut[yo*wo*3 + xo*3 + 1] = val.y;
  dOut[yo*wo*3 + xo*3 + 2] = val.z;
}

__host__
float *pdcuda::ImageResizerLanczos2D::resize(int w, int h) const
{
  return croppedResize(0,0, this->w,this->h, w,h);
}

__host__
float *pdcuda::ImageResizerLanczos2D::croppedResize(int x0, int y0,
                                                    int wc, int hc,
                                                    int w,  int h) const
{
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

  ScopedLock ll(t2_mut);

  t2.addressMode[0] = cudaAddressModeClamp;
  t2.addressMode[1] = cudaAddressModeClamp;
  t2.filterMode = cudaFilterModePoint;
  t2.normalized = false;    // access with integer texture coordinates

  cutilSafeCall(cudaBindTextureToArray(t2, data, channelDesc));
  
  float *dResized = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dResized, w*h*3*sizeof(float)));

  dim3 const block(16,8,1);
  dim3 const grid((int)ceil(w/float(block.x)), (int)ceil(h/float(block.y)));
  cudaGetLastError(); // reset error status
  ResizeLanczos2D_Kernel<<<grid,block>>>(dResized, x0,y0, wc,hc, w,h, 
                                         this->windowRadius);
  cutilSafeCall(cudaGetLastError());

  cutilSafeCall(cudaUnbindTexture(t2));

  return dResized;
}

///////////////////////////////////////////////////////////////////////////////
// ImageResizerLanczos3D

__host__
pdcuda::
ImageResizerLanczos3D::ImageResizerLanczos3D(float const *data, 
                                             int w, int h,
                                             int windowRadius,
                                             bool dataOnHost, 
                                             bool dataColumnMajor)
  : ImageResizer(w,h), windowRadius(windowRadius), data(NULL)
{
  this->data = rgbToCudaArray3d(data, w, h, dataOnHost, dataColumnMajor);
}

__host__
pdcuda::
ImageResizerLanczos3D::~ImageResizerLanczos3D() 
{
  if (data) {
    cutilSafeCall(cudaFreeArray(data));
    data = NULL;
  }
}

__global__ static
void ResizeLanczos3D_Kernel(float *dOut, 
                            int x0, int y0, int wc, int hc, int wo, int ho, 
                            int r)
{
  int   const xo = blockDim.x * blockIdx.x + threadIdx.x;
  int   const yo = blockDim.y * blockIdx.y + threadIdx.y;

  if (xo >= wo || yo >= ho) return;

  float const xi = (xo + 0.5f) * wc / wo;
  float const yi = (yo + 0.5f) * hc / ho;

  int   const xiStart = (wc>wo) ? ceilf(xi - r*wc/wo) : ceilf(xi - r);
  int   const yiStart = (hc>ho) ? ceilf(yi - r*hc/ho) : ceilf(yi - r);

  int   const nx = (wc>wo) ? floorf(2*r*wc/wo) : 2*r-1;
  int   const ny = (hc>ho) ? floorf(2*r*hc/ho) : 2*r-1;

  float const xlStart = (wc>wo) ? ((xiStart+0.5f-xi)*wo/wc):(xiStart+0.5f-xi);
  float const ylStart = (hc>ho) ? ((yiStart+0.5f-yi)*ho/hc):(yiStart+0.5f-yi);

  float const dxl = (wc>wo) ? ((float)wo)/wc : 1;
  float const dyl = (hc>ho) ? ((float)ho)/hc : 1;

  float norm = 0;
  float val  = 0;
  {
    int   yii = yiStart;
    float yli = ylStart;
    for (int j=0; j<=ny; j++, yii++, yli+=dyl) {
      int         xii = xiStart;
      float       xli = xlStart;
      float const ly  = lanczos(yli, r);

      for (int i=0; i<=nx; i++, xii++, xli+=dxl) {
        float const lx = lanczos(xli, r);
        float const l = lx * ly;
        norm += l;
        val  += (l * tex3D(t3, x0+xii, y0+yii, threadIdx.z));
      }
    }
  }
  val = max(0.f, min(255.f, val/norm));
  
  dOut[yo*wo*3 + xo*3 + threadIdx.z] = val;
}

__host__
float *pdcuda::ImageResizerLanczos3D::resize(int w, int h) const
{
  return croppedResize(0,0, this->w,this->h, w,h);
}

__host__
float *pdcuda::ImageResizerLanczos3D::croppedResize(int x0, int y0,
                                                    int wc, int hc,
                                                    int w,  int h) const
{
  static int const BLOCKDIM = 8;
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float>();

  ScopedLock ll(t3_mut);

  t3.addressMode[0] = cudaAddressModeClamp;
  t3.addressMode[1] = cudaAddressModeClamp;
  t3.addressMode[2] = cudaAddressModeClamp;
  t3.filterMode = cudaFilterModePoint;
  t3.normalized = false;    // access with integer texture coordinates

  cutilSafeCall(cudaBindTextureToArray(t3, data, channelDesc));
  
  float *dResized = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dResized, w*h*3*sizeof(float)));

  dim3 const block(BLOCKDIM, BLOCKDIM, 3);
  dim3 const grid((int)ceil(w/float(block.x)), (int)ceil(h/float(block.y)));
  cudaGetLastError(); // reset error status
  ResizeLanczos3D_Kernel<<<grid,block>>>(dResized, x0,y0, wc,hc, w,h, 
                                         this->windowRadius);
  cutilSafeCall(cudaGetLastError());

  cutilSafeCall(cudaUnbindTexture(t3));

  return dResized;
}

///////////////////////////////////////////////////////////////////////////////
// ImageResizerBilinear2D

__host__
pdcuda::
ImageResizerBilinear2D::ImageResizerBilinear2D(float const *data, 
                                               int w, int h,
                                               bool dataOnHost, 
                                               bool dataColumnMajor)
  : ImageResizer(w,h), data(NULL)
{
  this->data = rgbToCudaArray2d(data, w, h, dataOnHost, dataColumnMajor);
}

__host__
pdcuda::
ImageResizerBilinear2D::~ImageResizerBilinear2D() 
{
  if (data) {
    cutilSafeCall(cudaFreeArray(data));
    data = NULL;
  }
}

__global__ static
void ResizeBilinear2D_Kernel(float *dOut, 
                             int x0, int y0, int wc, int hc, int wo, int ho)
{
  int const xo = blockDim.x * blockIdx.x + threadIdx.x;
  int const yo = blockDim.y * blockIdx.y + threadIdx.y;

  if (xo >= wo || yo >= ho) return;

  float const xi = xo * wc / wo + 0.5f;
  float const yi = yo * hc / ho + 0.5f;

  float4 const val = tex2D(t2, x0+xi, y0+yi);

  dOut[yo*wo*3 + xo*3    ] = val.x;
  dOut[yo*wo*3 + xo*3 + 1] = val.y;
  dOut[yo*wo*3 + xo*3 + 2] = val.z;
}

__host__
float *pdcuda::ImageResizerBilinear2D::resize(int w, int h) const
{
  return croppedResize(0,0, this->w,this->h, w,h);
}

__host__
float *pdcuda::ImageResizerBilinear2D::croppedResize(int x0, int y0,
                                                     int wc, int hc,
                                                     int w,  int h) const
{
  static int const BLOCKDIM = 8;
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

  ScopedLock ll(t2_mut);

  t2.addressMode[0] = cudaAddressModeClamp;
  t2.addressMode[1] = cudaAddressModeClamp;
  t2.filterMode = cudaFilterModeLinear;
  t2.normalized = false;    // access with integer texture coordinates

  cutilSafeCall(cudaBindTextureToArray(t2, data, channelDesc));
  
  float *dResized = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dResized, w*h*3*sizeof(float)));

  dim3 const block(BLOCKDIM, BLOCKDIM, 1);
  dim3 const grid((int)ceil(w/float(block.x)), (int)ceil(h/float(block.y)));
  cudaGetLastError(); // reset error status
  ResizeBilinear2D_Kernel<<<grid,block>>>(dResized, x0,y0, wc,hc, w,h);
  cutilSafeCall(cudaGetLastError());

  cutilSafeCall(cudaUnbindTexture(t2));

  return dResized;
}
