#include <pdcuda/imtransfer.h>
#include <cuda_runtime.h>
#include <cutil/cutil_inline.h>
#include <cutil/error_checker.h>
#include <exception>
#include <stdexcept>
#include <iostream>
#include <vector>

using namespace std;
using namespace pdcuda;

#define NotYetImplemented \
  throw logic_error("NOT YET IMPLEMENTED (" __FILE__ ")")

// NOTE: the __host__ functions in this file are not exception-safe.
// They will leak memory if they throw an exception.

///////////////////////////////////////////////////////////////////////////////
// RGB (host or device) -> RGBA (device)

#define TBLOCKDIM 16

__global__ 
void TransposedRgbToRgba_Kernel(float4 *out, float const *in, int w, int h) 
{
  // Adapted from the nVidia transpose SDK example
  __shared__ union {
    float   f[TBLOCKDIM * TBLOCKDIM * 4];
    float4 f4[TBLOCKDIM * TBLOCKDIM];
  } rgbaData;
  
  // read the matrix tile into shared memory, convert to RGBA, and 
  // transpose all at once
  unsigned const xBase = blockIdx.x * TBLOCKDIM;
  unsigned const yBase = blockIdx.y * TBLOCKDIM;
  unsigned const sx    = threadIdx.y;
  if (xBase+sx < w) {
    unsigned const maxI  = min(TBLOCKDIM*3, (h-blockIdx.y*TBLOCKDIM)*3);
    for (unsigned i=threadIdx.x; i<maxI; i+=TBLOCKDIM) {
      unsigned const sy = i/3;
      unsigned const sc = i%3;
      rgbaData.f[sy*TBLOCKDIM*4 + sx*4 + sc] = 
        in[(xBase+sx)*h*3 + yBase*3 + i];
    }
  }  
  __syncthreads();
  
  // write the transposed matrix tile to global memory
  unsigned const xOut = (xBase + threadIdx.x);
  unsigned const yOut = (yBase + threadIdx.y);
  if (xOut < w && yOut < h) {
    float4 f4 = rgbaData.f4[threadIdx.y*TBLOCKDIM + threadIdx.x];
    f4.w = 0; // clear alpha channel
    out[yOut*w + xOut] = f4;
  }
}

// Change at will (preferably a multiple of 0.5*warpSize).
#define RGBTORGBA_BLOCKDIM_X 32

// This kernel takes an packed row-major RGB image and expands it to
// be a packed row-major RGBA image, using 0 as the alpha channel value. 
//
// It could probably be optimized a little more to use 128-bit I/O and
// to avoid some thread divergence.  It takes about 9ms to run on a
// GTX 260+ for a 1920x1080 image.
//
// Another possible optimization would be to read directly into RGBA
// shared memory.  This would reduce the shared memory requirements by
// almost 50%, but it would make the indexing math more complex when
// reading from global memory.
//
// Assumes that 
//    blockDim.x = RGBTORGBA_BLOCKDIM_X 
//    blockDim.y = 1 
//    blockDim.z = 1
__global__
static void RgbToRgba_Kernel(float *out, float const *in, int w, int h)
{
  __shared__ float rgbFlat [RGBTORGBA_BLOCKDIM_X * 3];
  __shared__ float rgbaData[RGBTORGBA_BLOCKDIM_X * 4];

  if (blockDim.y*blockIdx.y + threadIdx.y >= h) return;
  
  // Collaboratively load the RGB data
  {
    int   const  xcmax   = w*3;
    int          colIdx  = blockDim.x*blockIdx.x*3 + threadIdx.x;
    float const *gmem    = in + blockDim.y*blockIdx.y*w*3 + colIdx;
    rgbFlat[threadIdx.x]                           = (colIdx<xcmax) ? *gmem:-1;
    colIdx += RGBTORGBA_BLOCKDIM_X;
    gmem   += RGBTORGBA_BLOCKDIM_X;
    rgbFlat[threadIdx.x+RGBTORGBA_BLOCKDIM_X]      = (colIdx<xcmax) ? *gmem:-1;
    colIdx += RGBTORGBA_BLOCKDIM_X;
    gmem   += RGBTORGBA_BLOCKDIM_X;
    rgbFlat[threadIdx.x+(RGBTORGBA_BLOCKDIM_X<<1)] = (colIdx<xcmax) ? *gmem:-1;
  }
  __syncthreads();

  // Shared-memory conversion to RGBA
  for (int c=0; c<3; c++) {
    rgbaData[threadIdx.x*4 + c] = rgbFlat[threadIdx.x*3 + c];
  }
  rgbaData[threadIdx.x*4 + 3] = 0;
  __syncthreads();

  // Collaboratively store the RGBA data
  {
    int   const  xcmax   = w*4;
    int          colIdx  = blockDim.x*blockIdx.x*4 + threadIdx.x;
    float       *gmem    = out + blockDim.y*blockIdx.y*w*4 + colIdx;
    if (colIdx<xcmax) *gmem = rgbaData[threadIdx.x];
    colIdx += RGBTORGBA_BLOCKDIM_X;
    gmem   += RGBTORGBA_BLOCKDIM_X;
    if (colIdx<xcmax) *gmem = rgbaData[threadIdx.x+ RGBTORGBA_BLOCKDIM_X];
    colIdx += RGBTORGBA_BLOCKDIM_X;
    gmem   += RGBTORGBA_BLOCKDIM_X;
    if (colIdx<xcmax) *gmem = rgbaData[threadIdx.x+(RGBTORGBA_BLOCKDIM_X<<1)];
    colIdx += RGBTORGBA_BLOCKDIM_X;
    gmem   += RGBTORGBA_BLOCKDIM_X;
    if (colIdx<xcmax) *gmem = rgbaData[threadIdx.x+RGBTORGBA_BLOCKDIM_X*3];
  }
}


__host__
float4 *pdcuda::rgbToDeviceRgba(float const *data, int w, int h,
                                bool dataOnHost, bool dataColumnMajor) 
{
  // TODO: use streams
  unsigned int const rgbSize  = w * h * 3 * sizeof(float);
  unsigned int const rgbaSize = w * h * sizeof(float4);

  // Transfer RGB data to device if needed
  float const *dData = NULL;
  if (dataOnHost) {
    cutilSafeCall(cudaMalloc((void**)(void*)&dData, rgbSize));
    cutilSafeCall(cudaMemcpy((float*)dData, data, rgbSize, 
                             cudaMemcpyHostToDevice));
  } else {
    dData = data;
  }

  // Convert to RGBA, transposing too if needed
  float4 *dRet = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dRet, rgbaSize));
  cudaGetLastError(); // reset error status
  if (dataColumnMajor) {
    dim3 const block(TBLOCKDIM, TBLOCKDIM, 1);
    dim3 const grid((int)ceil(w/float(block.x)), (int)ceil(h/float(block.y)));
    TransposedRgbToRgba_Kernel<<<grid,block>>>(dRet, dData, w, h);
  } else {
    dim3 const block(RGBTORGBA_BLOCKDIM_X,1,1);
    dim3 const grid((int)ceil(w/float(block.x)), (int)ceil(h/float(block.y)));
    RgbToRgba_Kernel<<<grid, block>>>((float*)dRet, dData, w, h);
  }
  cutilSafeCall(cudaGetLastError());

  if (dataOnHost) cutilSafeCall(cudaFree(const_cast<float *>(dData)));

  return dRet;
}

///////////////////////////////////////////////////////////////////////////////
// RGB (host or device) -> cudaArray<float4,2>

__host__
cudaArray *pdcuda::rgbToCudaArray2d(float const *data, int w, int h,
                                    bool dataOnHost, bool dataColumnMajor)
{
  float4 *dRgba = rgbToDeviceRgba(data, w, h, dataOnHost, dataColumnMajor);
  cudaArray *dCudaArray = rgbaToCudaArray2d(dRgba, w, h, false, false);
  cutilSafeCall(cudaFree(dRgba));
  return dCudaArray;
}

///////////////////////////////////////////////////////////////////////////////
// RGBA (host or device) -> cudaArray<float4,2>

#define F4TBLOCKDIM 16

// Adapted from the CUDA SDK transpose example to use float4 instead
// of float elements.
//
// This kernel is optimized to ensure all global reads and writes are
// coalesced, and to avoid bank conflicts in shared memory.  This
// kernel is up to 11x faster than the naive kernel below.  Note that
// the shared memory array is sized to (BLOCK_DIM+1)*BLOCK_DIM.  This
// pads each row of the 2D block in shared memory so that bank
// conflicts do not occur when threads address the array column-wise.
__global__ void Float4Transpose_Kernel(float4 *odata, float4 const *idata, 
                                       int width, int height)
{
  __shared__ float4 block[F4TBLOCKDIM][F4TBLOCKDIM+1];
  
  // read the matrix tile into shared memory
  unsigned xIndex = blockIdx.x * F4TBLOCKDIM + threadIdx.x;
  unsigned yIndex = blockIdx.y * F4TBLOCKDIM + threadIdx.y;
  if ((xIndex < width) && (yIndex < height)) {
    unsigned index_in = yIndex * width + xIndex;
    block[threadIdx.y][threadIdx.x] = idata[index_in];
  }

  __syncthreads();

  // write the transposed matrix tile to global memory
  xIndex = blockIdx.y * F4TBLOCKDIM + threadIdx.x;
  yIndex = blockIdx.x * F4TBLOCKDIM + threadIdx.y;
  if ((xIndex < height) && (yIndex < width)) {
    unsigned index_out = yIndex * height + xIndex;
    odata[index_out] = block[threadIdx.x][threadIdx.y];
  }
}

__host__
cudaArray *pdcuda::rgbaToCudaArray2d(float4 const *data, int w, int h,
                                     bool dataOnHost, bool dataColumnMajor)
{
  // Configure the input data
  bool                 freeIn   = false;
  float4 const        *in       = NULL;
  enum cudaMemcpyKind  xferKind = (enum cudaMemcpyKind)-1;
  if (dataColumnMajor) {
    float4 *inTmp = NULL;
    unsigned SZ = w*h*sizeof(float4);
    cutilSafeCall(cudaMalloc((void**)(void*)&inTmp, SZ));
    
    dim3 const block(F4TBLOCKDIM, F4TBLOCKDIM, 1);
    dim3 const grid((int)ceil(h/float(F4TBLOCKDIM)), 
                    (int)ceil(w/float(F4TBLOCKDIM)));
    if (dataOnHost) {
      float4 *dData = NULL;
      cutilSafeCall(cudaMalloc((void**)(void*)&dData, SZ));
      cutilSafeCall(cudaMemcpy(dData, data, SZ, cudaMemcpyHostToDevice));
      cudaGetLastError(); // reset error status
      Float4Transpose_Kernel<<<grid,block>>>(inTmp, dData, h, w);
      cutilSafeCall(cudaGetLastError());
      cutilSafeCall(cudaFree(dData));
    } else {
      cudaGetLastError(); // reset error status
      Float4Transpose_Kernel<<<grid,block>>>(inTmp, data, h, w);
      cutilSafeCall(cudaGetLastError());
    }

    in       = inTmp;
    freeIn   = true;
    xferKind = cudaMemcpyDeviceToDevice;
  } else {
    in       = data;
    xferKind = dataOnHost ? cudaMemcpyHostToDevice : cudaMemcpyDeviceToDevice;
  }

  // Create the cudaArray
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();
  cudaArray *dCudaArray = NULL;
  cutilSafeCall(cudaMallocArray(&dCudaArray, &channelDesc, w,h));
  
  // Transfer data to the cudaArray
  cutilSafeCall(cudaMemcpy2DToArray(dCudaArray, 0,0, in,
                                    w*sizeof(float4), w*sizeof(float4), h, 
                                    xferKind));

  // Cleanup
  if (freeIn) cutilSafeCall(cudaFree(const_cast<float4 *>(in)));

  return dCudaArray;
}

///////////////////////////////////////////////////////////////////////////////
// RGB (host or device) -> cudaArray<float,3>

#define TOPLANAR_BLOCKDIM 8

#define BC 0
#define BX (TOPLANAR_BLOCKDIM * blockIdx.x)
#define BY (TOPLANAR_BLOCKDIM * blockIdx.y)
#define TC threadIdx.x
#define TX threadIdx.y
#define TY threadIdx.z
 
__global__
void XycToCyx_Kernel(float *dCYX, float const *dRgb, int w, int h) 
{
  // NOTE: some care could probably be taken to avoid bank conflicts
  // by padding a dimension.
  __shared__ float s[TOPLANAR_BLOCKDIM * TOPLANAR_BLOCKDIM * 3];

  // Do the transpose while reading.  Here TX refers to the Y
  // coordinate and TY refers to the X coordinate, so we can have
  // coalesced memory access.  blockDim.y and blockDim.x must be
  // equal for this to work.
  if (BX+TY < w && BY+TX < h) {
    s[TC * TOPLANAR_BLOCKDIM * TOPLANAR_BLOCKDIM + 
      TX * TOPLANAR_BLOCKDIM + 
      TY] = 
      dRgb[(BX+TY)*h*3 + (BY+TX)*3 + (BC+TC)];
  }
  __syncthreads();

  // Figure out the indexing that will minimize non-coalesced writes
  int thrdI = (threadIdx.x + 
               threadIdx.y*blockDim.x + 
               threadIdx.z*blockDim.x*blockDim.y);
  int const oc = thrdI / (TOPLANAR_BLOCKDIM * TOPLANAR_BLOCKDIM);
  thrdI %= (TOPLANAR_BLOCKDIM * TOPLANAR_BLOCKDIM);
  int const oy = thrdI / TOPLANAR_BLOCKDIM;
  int const ox = thrdI % TOPLANAR_BLOCKDIM;

  // Write out our element
  if (BX+ox < w && BY+oy < h) {
    dCYX[(BC+oc)*w*h + (BY+oy)*w + (BX+ox)] = 
      s[oc*(TOPLANAR_BLOCKDIM*TOPLANAR_BLOCKDIM) + oy*TOPLANAR_BLOCKDIM + ox];
  }
}

__global__
void YxcToCyx_Kernel(float *dCYX, float const *dRgb, int w, int h) 
{
  // NOTE: some care could probably be taken to avoid bank conflicts
  // by padding a dimension.
  __shared__ float s[TOPLANAR_BLOCKDIM * TOPLANAR_BLOCKDIM * 3];
  
  if (BX+TX < w && BY+TY < h) {
    s[TC * TOPLANAR_BLOCKDIM * TOPLANAR_BLOCKDIM + 
      TY * TOPLANAR_BLOCKDIM + 
      TX] = 
      dRgb[(BY+TY)*w*3 + (BX+TX)*3 + (BC+TC)];
  }
  __syncthreads();

  // Figure out the indexing that will minimize non-coalesced writes
  int thrdI = (threadIdx.x + 
               threadIdx.y*blockDim.x + 
               threadIdx.z*blockDim.x*blockDim.y);
  int const oc = thrdI / (TOPLANAR_BLOCKDIM * TOPLANAR_BLOCKDIM);
  thrdI %= (TOPLANAR_BLOCKDIM * TOPLANAR_BLOCKDIM);
  int const oy = thrdI / TOPLANAR_BLOCKDIM;
  int const ox = thrdI % TOPLANAR_BLOCKDIM;

  // Write out our element
  if (BX+ox < w && BY+oy < h) {
    dCYX[(BC+oc)*w*h + (BY+oy)*w + (BX+ox)] = 
      s[oc*(TOPLANAR_BLOCKDIM*TOPLANAR_BLOCKDIM) + oy*TOPLANAR_BLOCKDIM + ox];
  }
}

#undef TY
#undef TX
#undef TC
#undef BY
#undef BX
#undef BC

__host__
bool pdcuda::valid3dRgbArraySize(int w, int h) {
  if (w == 0 || h == 0) return false;
  if (w >= MAX_3D_TEXTURE_EXT) return false;
  if (h >= MAX_3D_TEXTURE_EXT) return false;
  if (h == 1) return true; // 2D-degenerate sizes are more permissive
  if ((w & 0x000F) != 0) return false;
  return true;
}

__host__
cudaArray *pdcuda::rgbToCudaArray3d(float const *data, int w, int h,
                                    bool dataOnHost, bool dataColumnMajor)
{
  ErrorChecker::condition(valid3dRgbArraySize(w,h), __FILE__, __LINE__);

  unsigned const rgbSize = w * h * 3 * sizeof(float);

  // Make sure the data are on the device
  float const *dData = NULL;
  if (dataOnHost) {
    cutilSafeCall(cudaMalloc((void**)(void*)&dData, rgbSize));
    cutilSafeCall(cudaMemcpy((float*)dData, data, rgbSize, 
                             cudaMemcpyHostToDevice));
  } else {
    dData = data;
  }

  // Allocate the 3D array
  cudaArray *dCudaArray = NULL;
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float>();
  int        const pwI = w; 
  int        const phI = h; 
  int        const pdI = 3; 
  cudaExtent const ext = make_cudaExtent(pwI, phI, pdI);
  cutilSafeCall(cudaMalloc3DArray(&dCudaArray, &channelDesc, ext));

  // Permute the dimensions as needed
  float *dCYX = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dCYX, w*h*3*sizeof(float)));
  if (dataColumnMajor) {
    cudaGetLastError(); // reset error status
    dim3 const block(3, TOPLANAR_BLOCKDIM, TOPLANAR_BLOCKDIM);
    dim3 const grid((int)ceil(w/float(TOPLANAR_BLOCKDIM)), 
                    (int)ceil(h/float(TOPLANAR_BLOCKDIM)));
    XycToCyx_Kernel<<<grid,block>>>(dCYX, dData, w, h);
  } else {
    cudaGetLastError(); // reset error status
    dim3 const block(3, TOPLANAR_BLOCKDIM, TOPLANAR_BLOCKDIM);
    dim3 const grid((int)ceil(w/float(TOPLANAR_BLOCKDIM)), 
                    (int)ceil(h/float(TOPLANAR_BLOCKDIM)));
    YxcToCyx_Kernel<<<grid,block>>>(dCYX, dData, w, h);
  }
  cutilSafeCall(cudaGetLastError());
  
  // Transfer the data
  cudaMemcpy3DParms copyParams = {0};
  copyParams.extent   = ext;
  copyParams.kind     = cudaMemcpyDeviceToDevice;
  copyParams.dstArray = dCudaArray;
  copyParams.srcPtr   = make_cudaPitchedPtr(
    dCYX, ext.width*sizeof(float), ext.width, ext.height);
  cutilSafeCall(cudaMemcpy3D(&copyParams));

  // Cleanup
  cutilSafeCall(cudaFree(dCYX));
  if (dataOnHost) cutilSafeCall(cudaFree(const_cast<float *>(dData)));

  return dCudaArray;
}
