#include <pdcuda/rhogdense.h>
#include <stdio.h>
#include <iostream>
#include <cutil/cutil_inline.h>
#include <cutil/error_checker.h>
#include <pdcuda/dyndispatch.h>

using namespace std;
using namespace pdcuda;

#define DISCRETE_HISTOGRAM         0
#define PRECISION_HISTOGRAM        1 // what Dalal used
#define PRECISION_ORIENT_HISTOGRAM 2 // discrete in space, precise in orient.
// TODO: make this be a template parameter
#define HISTOGRAM_TYPE PRECISION_HISTOGRAM

// Use #define: nvcc 2.1 can't handle "static float const" values.
#define OriExtent    ((SemiCirc) ? 180.f : 360.f)
//static float const HalfBin      = OrientBin / 2.f;
#define InvBandwidth ((OrientBin) / (OriExtent))
//static float const CenterBound  = OriExtent / 2.0f;

static int const RDThreadsPerBlock = 9;

__constant__ float gaussianStencil[BlockSizeY][BlockSizeX];

#if HISTOGRAM_TYPE == DISCRETE_HISTOGRAM
// Custom bin wrapping, optimized for RhogDense_Kernel
__device__ 
inline 
float wrapIt(int bin) {
  return (bin < OrientBin) ? bin : 0;
}
#endif

#if HISTOGRAM_TYPE != DISCRETE_HISTOGRAM
// Iterative modulus, optimized for cases when val is within or close
// to the [0,divisor) range.  divisor must be positive.
template <typename T>
__device__ 
inline T itermod(T val, T divisor) {
  while (val >= divisor) val -= divisor;
  while (val < 0) val += divisor;
  return val;
}
#endif

// A thread computes an entire block descriptor.  Blocks are used to
// collaborate on loading and storing data.
//
// Constant memory: spatial stencil
// Texture memory:  unused
// Shared memory:   paged dMag & dOri; block descriptors
//
// Assumes:
//   blockDim = {RDThreadsPerBlock,1,1}
//   w >= BlockSizeX
//   h >= BlockSizeY
//   RDThreadsPerBlock > 1
//
// Ruled-out optimizations:
//
//  1) 1D linear texture memory for dMag and dOri: results in a slight
//  performance decrease since our read accesses are already very
//  well ordered.  
// 
//  2) Use 128-bit loads for collaborative loading.  This optimization
//  does work, but it only yields a 4.7% speed improvement for the
//  optimal block size (9,1,1).  Also, it requires that the input
//  gradient image width be a multiple of 4 (otherwise we run into
//  alignment issues).
//
//  3) Use the "register trick" to store sBD in registers intead of
//  shared memory.  sBD needs too much space, so it causes even more
//  register pressure under this optimization than the shared memory
//  pressure it currently exerts.  
//
// Profiling data (3840x2160 image on a GTX 280 / Core i7 965 with 9
// threads per block):
//
//  49,145.98 us GPU time
//      18.8% occupancy
//    1996    shared memory bytes per block (limiting factor)
//      38    registers / thread
//       8    warps / multiprocess (limiting factor)
//
//   branches:           2,567,964
//   divergent branches:   118,695 (4.6%)
//   bank conflicts:     1,922,345
//
template <int FineStrideX, int FineStrideY>
__global__ static
void RhogDense_Kernel(float *dBD, float const *dMag, float const *dOri,
                      int w, int h, 
                      float epsilon, float maxVal, float epsilon2)
{
  // This layout saves us from duplicating the shared values, but may
  // lead to a lot of bank conflicts.
  //   sMag[threadIdx.x*FineStrideX + cx*CellSizeX + x]
  __shared__ float sMag[(RDThreadsPerBlock-1)*FineStrideX + BlockSizeX];
  __shared__ float sOri[(RDThreadsPerBlock-1)*FineStrideX + BlockSizeX];
  // This layout should avoid all bank conflicts on writes to gmem
  //   sBD[((cy*NumCellX + cx)*OrientBin + bin)*RDThreadsPerBlock + threadIdx.x]
  __shared__ float sBD[BlockDescLen * RDThreadsPerBlock];

  // This will cause the whole CUDA block to exit early, so we don't 
  // need to worry about __syncthreads() problems.
  if (blockDim.y*blockIdx.y + threadIdx.y + BlockSizeY > h) return;

  // Collaboratively clear the histograms
  for (int i=threadIdx.x; i<BlockDescLen*RDThreadsPerBlock; i+=RDThreadsPerBlock) {
    sBD[i] = 0;
  }
  // No __syncthreads() needed...it'll happen later
 
  // Process a horizontal strip of cells at the same time
  for (int cy=0; cy<NumCellY; cy++) {
    for (int y=0; y<CellSizeY; y++) {
      // Collaboratively load a horizontal strip of gradients. 
      {
        int const maxX = min((RDThreadsPerBlock-1)*FineStrideX + BlockSizeX,
                             w - blockDim.x*blockIdx.x*FineStrideX);
        int const inBase =             
          (blockDim.y*blockIdx.y*FineStrideY + cy*CellSizeY + y) * w + 
          (blockDim.x*blockIdx.x*FineStrideX);
        for (int sx=threadIdx.x; sx<maxX; sx+=RDThreadsPerBlock) {
          sMag[sx] = dMag[inBase+sx];
          sOri[sx] = dOri[inBase+sx];
        }
      }
      __syncthreads();

      // Compute histograms for all cells in the strip
      for (int cx=0; cx<NumCellX; cx++) {
#if HISTOGRAM_TYPE != PRECISION_HISTOGRAM
        int const histBase = 
          ((cy*NumCellX + cx)*OrientBin)*RDThreadsPerBlock + threadIdx.x;
#endif
        for (int x=0; x<CellSizeX; x++) {
          int const inIdx = threadIdx.x*FineStrideX + cx*CellSizeX + x;
#if HISTOGRAM_TYPE == DISCRETE_HISTOGRAM
          int const bin   = wrapIt((int)(sOri[inIdx] * InvBandwidth)); 
          sBD[histBase + bin*RDThreadsPerBlock] += 
            sMag[inIdx] * gaussianStencil[cy*CellSizeY + y][cx*CellSizeX + x];
#elif HISTOGRAM_TYPE == PRECISION_ORIENT_HISTOGRAM
          float const oPreciseBin = (sOri[inIdx] * InvBandwidth) - 0.5f;
          float const oDelta      = (oPreciseBin - (int)floorf(oPreciseBin));
          int   const oFloorBin   = itermod<int>(floorf(oPreciseBin), OrientBin);
          int   const oCeilBin    = itermod<int>(ceilf(oPreciseBin),OrientBin);

          float const baseVote = sMag[inIdx] * 
            gaussianStencil[cy*CellSizeY + y][cx*CellSizeX + x];

          sBD[histBase + oFloorBin*RDThreadsPerBlock] += (1-oDelta) * baseVote;
          sBD[histBase + oCeilBin*RDThreadsPerBlock]  +=    oDelta  * baseVote;
#elif HISTOGRAM_TYPE == PRECISION_HISTOGRAM
          // Figure out how to spread the votes over a 2x2x2 space
          float const oPreciseBin = (sOri[inIdx] * InvBandwidth) - 0.5f;
          int   const oFloorBin   = itermod<int>(floorf(oPreciseBin), OrientBin);
          int   const oCeilBin    = itermod<int>(ceilf(oPreciseBin),OrientBin);
          float const oDelta      = (oPreciseBin - (int)floorf(oPreciseBin));
          int   const ofo         = oFloorBin * RDThreadsPerBlock; // orientation floor bin offset
          int   const oco         = oCeilBin * RDThreadsPerBlock;
          
          float const xPreciseBin = cx + (x+0.5f) / (float)CellSizeX - 0.5f;
          float       xDelta      = (xPreciseBin - (int)floorf(xPreciseBin));
          int   const xFloorBin   = (int)floorf(xPreciseBin);
          int   const xCeilBin    = (int)ceilf(xPreciseBin);
          int   const xfo         = xFloorBin*OrientBin*RDThreadsPerBlock;
          int   const xco         =  xCeilBin*OrientBin*RDThreadsPerBlock;

          float const yPreciseBin = cy + (y+0.5f) / (float)CellSizeY - 0.5f;
          float       yDelta      = (yPreciseBin - (int)floorf(yPreciseBin));
          int   const yFloorBin   = (int)floorf(yPreciseBin);
          int   const yCeilBin    = (int)ceilf(yPreciseBin);
          int   const yfo         = yFloorBin*NumCellX*OrientBin*RDThreadsPerBlock;
          int   const yco         =  yCeilBin*NumCellX*OrientBin*RDThreadsPerBlock;

          // Total vote mass for the current input pixel
          float const v = sMag[inIdx] *
            gaussianStencil[cy*CellSizeY + y][cx*CellSizeX + x];

          // Actually perform the voting
          if (xFloorBin >= 0 && yFloorBin >= 0) {
            sBD[threadIdx.x+yfo+xfo+ofo] += (1-oDelta)*(1-xDelta)*(1-yDelta)*v;
            sBD[threadIdx.x+yfo+xfo+oco] += (  oDelta)*(1-xDelta)*(1-yDelta)*v;
          }
          if (xCeilBin < NumCellX && yFloorBin >= 0) {
            sBD[threadIdx.x+yfo+xco+ofo] += (1-oDelta)*(  xDelta)*(1-yDelta)*v;
            sBD[threadIdx.x+yfo+xco+oco] += (  oDelta)*(  xDelta)*(1-yDelta)*v;
          }
          if (xFloorBin >= 0 && yCeilBin < NumCellY) {
            sBD[threadIdx.x+yco+xfo+ofo] += (1-oDelta)*(1-xDelta)*(  yDelta)*v;
            sBD[threadIdx.x+yco+xfo+oco] += (  oDelta)*(1-xDelta)*(  yDelta)*v;
          }
          if (xCeilBin < NumCellX && yCeilBin < NumCellY) {
            sBD[threadIdx.x+yco+xco+ofo] += (1-oDelta)*(  xDelta)*(  yDelta)*v;
            sBD[threadIdx.x+yco+xco+oco] += (  oDelta)*(  xDelta)*(  yDelta)*v;
          }
#else 
#         error "BAD HISTOGRAM_TYPE"
#endif
        }
      }
      __syncthreads();
    }
  }

  // Normalize (L2Hys)
  {
    // Find L2 norm
    float norm = 0.f;
    for (int idx=threadIdx.x; idx<BlockDescLen*RDThreadsPerBlock;
         idx+=RDThreadsPerBlock) {
      norm += sBD[idx] * sBD[idx];
    }
    norm = 1 / (sqrtf(norm) + epsilon*BlockDescLen);

    // Apply L2 norm, clip values, find hysNorm
    float hysNorm = 0.f;
    for (int idx=threadIdx.x; idx<BlockDescLen*RDThreadsPerBlock;
         idx+=RDThreadsPerBlock) {
      float const clipped = min(maxVal, sBD[idx] * norm);
      sBD[idx] = clipped;
      hysNorm += (clipped * clipped);
    }
    hysNorm = 1 / (sqrtf(hysNorm) + epsilon2);

    // Apply hysNorm
    for (int idx=threadIdx.x; idx<BlockDescLen*RDThreadsPerBlock;
         idx+=RDThreadsPerBlock) {
      sBD[idx] *= hysNorm;
    }
  }
  __syncthreads();

  // Collaboratively store the orientation data
  //   sBD[bi*RDThreadsPerBlock + threadIdx.x]
  //   gBD[bi*BH*BW + y*BW + x]
  {
    int const BW = horizBlocks(w, FineStrideX);
    int const BH = vertBlocks(h, FineStrideY);
      
    int       sIdx    = threadIdx.x;
    int const sStride = RDThreadsPerBlock;

    int const x       = blockDim.x*blockIdx.x + threadIdx.x;
    int const y       = blockDim.y*blockIdx.y;
    int       dIdx    = y*BW + x;
    int const dStride = BH * BW;

    if (x < BW) {
      for (int i=0; i<BlockDescLen; i++) {
        dBD[dIdx] = sBD[sIdx];
        dIdx += dStride;
        sIdx += sStride;
      }
    }
  }
}

#define SQR(x) ((x)*(x))

// Initialize the gaussianStencil.  Although technically this function
// is not threadsafe, it probably wouldn't cause any problems to call
// it from multiple threads since the cudaMemcpyToSymbol calls are
// probably serialized and the values don't change if the code is
// executed several times.
__host__ static 
void initStencil() {
  static bool first = true;
  if (!first) return;

  float stencil[BlockSizeY][BlockSizeX];
  float var2[2] = { 2 * SQR( BlockSizeX / (2*WeightScale) ),
                    2 * SQR( BlockSizeY / (2*WeightScale) ) };
  float center[2] = { (BlockSizeX-1) / 2.f, 
                      (BlockSizeY-1) / 2.f };
  for (int y=0; y<BlockSizeY; y++) {
    for (int x=0; x<BlockSizeX; x++) {
      stencil[y][x] = exp( -(SQR(x-center[0]) / var2[0] +
                             SQR(y-center[1]) / var2[1]) );
    }
  }
  cutilSafeCall(cudaMemcpyToSymbol(gaussianStencil, stencil, sizeof(stencil)));

  // Avoid thread safety issues by flagging only after we've
  // completely initialized the constant memory.
  first = true; 
}

struct RhogDenseFunctor {
  RhogDenseFunctor(float *dBD, float const *dMag, float const *dOri, 
                   int w, int h, float epsilon, float maxVal, float epsilon2)
    : dBD(dBD), dMag(dMag), dOri(dOri), w(w), h(h), epsilon(epsilon), 
      maxVal(maxVal), epsilon2(epsilon2)
  {};

  template <int FineStrideX, int FineStrideY>
  void dispatch() {
    int const BW = horizBlocks(w, FineStrideX);
    int const BH = vertBlocks(h, FineStrideY);
    cudaGetLastError(); // reset error status
    dim3 const block(RDThreadsPerBlock,1,1);
    dim3 const grid((int)ceil(BW/float(block.x)), 
                    (int)ceil(BH/float(block.y)));  
    RhogDense_Kernel<FineStrideX, FineStrideY>
      <<<grid,block>>>(dBD, dMag,dOri, w,h, epsilon,maxVal,epsilon2);
    cutilSafeCall(cudaGetLastError());
  }

private:
  float       *dBD;
  float const *dMag;
  float const *dOri;
  int w, h; 
  float epsilon, maxVal, epsilon2;
};

__host__
void pdcuda::computeRhogDense(float *dBD, 
                              float const *dMag, float const *dOri,
                              int w, int h, int fineStrideX, int fineStrideY,
                              float epsilon, float maxVal, float epsilon2)
{
  initStencil();
  RhogDenseFunctor f(dBD, dMag,dOri, w,h, epsilon,maxVal,epsilon2);
  IntDispatcher2D<MaxFineStrideX, MaxFineStrideY, RhogDenseFunctor>::dispatch
    (fineStrideX, fineStrideY, f);
}

///////////////////////////////////////////////////////////////////////////////

__constant__ float cClassifier[WinDescLen];
__constant__ float cBias;

// Each thread is responsible for computing the classification score
// for a single detector window.
//
// Constant memory: classification vector & bias 
// Texture memory:  unused
// Shared memory:   paged block descriptor inputs
//
// Assumes:
//   blockDim = {NumBlockX,NumBlockY,1}   
//   WinStrideX == BlockStrideX == fineStrideX 
//   WinStrideY == BlockStrideY == fineStrideY
//   bw >= NumBlockX
//   bh >= NumBlockY
//
// Logical indexing:
//   cClassifier(dxOffs,dyOffs,didx)
//   dBD(bidx,dy,dx)
//   dScores(wy,wx)
//
// Memory layout:
//   cClassifier[dxOffs*NumBlockY*BlockDescLen + dyOffs*BlockDescLen + didx]
//   dBD[bidx*bh*bw + wy*bw + dx]
//   dScores[wy * WW + wx]
//
// Future improvements:
//
// It should be possible to generalize this code to handle cases
// whenever (WinStride%fineStride==0) and (BlockStride%fineStride==0).
// Doing so would make the code a bit more complex because we'd have
// to do strided reads from dBD and strided writes to dScores, which
// would involve encoding 4D coordinates into the 2D blockDim
// (assuming we want to keep the task of a single thread the same and
// the collaborative loading of a block the same).
__global__ static
void ClassifyWindows_Kernel(float *dScores, float const *dBD, int bw, int bh)
{
  __shared__ float sBD[NumBlockY*2-1][NumBlockX*2-1];

  int const fineStrideX = WinStrideX;
  int const fineStrideY = WinStrideY;

  int const WW = horizWindows(bw, fineStrideX);
  int const WH = vertWindows(bh, fineStrideY);

  int const wx = (blockDim.x*blockIdx.x + threadIdx.x);
  int const wy = (blockDim.y*blockIdx.y + threadIdx.y);

  // Initialize the classification score
  float score = -cBias;

  for (int cx=0; cx<NumCellX; cx++) {
    for (int cy=0; cy<NumCellY; cy++) {
      for (int o=0; o<OrientBin; o++) {
        int const dBase = (((cy)*NumCellX + cx)*OrientBin + o)*bh*bw;
        int const cOffs = (((cx)*NumCellY + cy)*OrientBin + o);

        // Collaboratively load the bin data.  We grab one bin from
        // all of the block descriptors that affect this CUDA block.
        //
        // This implementation requires that
        // blockDim={NumBlockX,NumBlockY,1}.  It does not have
        // well-coalesced reads for typical values of NumBlockX and
        // FineStrideX, but we avoid the very expensive integer
        // division and modulo operations this way.
        {
          // We have two types of checks here: one that limits the
          // apron area (those that look at threadIdx) and one type
          // that handles looking beyond the edge of the image (those
          // that look at bw and bh).

          // main part
          sBD[threadIdx.y][threadIdx.x] = dBD[dBase + wy*bw + wx];
          if (threadIdx.x < NumBlockX-1) {
            // right apron
            sBD[threadIdx.y][threadIdx.x + NumBlockX] = 
              (wx+NumBlockX < bw) ? dBD[dBase + wy*bw + (wx+NumBlockX)] : 0;
          }
          if (threadIdx.y < NumBlockY-1) {
            // lower apron
            sBD[threadIdx.y + NumBlockY][threadIdx.x] = 
              (wy+NumBlockY < bh) ? dBD[dBase + (wy+NumBlockY)*bw + wx] : 0;
            if (threadIdx.x < NumBlockX-1) {
              // lower-right apron
              sBD[threadIdx.y + NumBlockY][threadIdx.x + NumBlockX] = 
                ((wy+NumBlockY < bh) && (wx+NumBlockX < bw)) ?
                dBD[dBase + (wy+NumBlockY)*bw + (wx+NumBlockX)] : 0;
            }
          }
        }
        __syncthreads();

        // Partial dot products with the classification vector
        for (int bxOffs=0; bxOffs<NumBlockX; bxOffs++) {
          for (int byOffs=0; byOffs<NumBlockY; byOffs++) {
            float const bdVal = sBD[byOffs + threadIdx.y][bxOffs + threadIdx.x];
            float const cVal  = cClassifier[bxOffs*NumBlockY*BlockDescLen + 
                                            byOffs*BlockDescLen + 
                                            cOffs];
            score += bdVal * cVal;
          }
        }
        __syncthreads();
      }
    }
  }

  // Save the final classification score
  if (wx<WW && wy<WH) {
    dScores[wy*WW + wx] = score;
  }
}

// Last dim is most contiguous
//   dScores: (wy,wx)
//       dBD: (cx,cy,bin,by,bx)
__host__
void pdcuda::classifyWindows(float *dScores, float const *dBD,
                             int bw, int bh, int fineStrideX, int fineStrideY, 
                             float const *hClassifier, float bias)
{
  ErrorChecker::condition(bw >= NumBlockX,             __FILE__, __LINE__);
  ErrorChecker::condition(bh >= NumBlockY,             __FILE__, __LINE__);
  ErrorChecker::condition(WinStrideX   == fineStrideX, __FILE__, __LINE__);
  ErrorChecker::condition(BlockStrideX == fineStrideX, __FILE__, __LINE__);
  ErrorChecker::condition(WinStrideY   == fineStrideY, __FILE__, __LINE__);
  ErrorChecker::condition(BlockStrideY == fineStrideY, __FILE__, __LINE__);

  cutilSafeCall(cudaMemcpyToSymbol(cClassifier, hClassifier, 
                                   sizeof(cClassifier)));
  cutilSafeCall(cudaMemcpyToSymbol(cBias, &bias, sizeof(bias)));

  cudaGetLastError(); // reset error status
  dim3 const block(NumBlockX, NumBlockY, 1);
  dim3 const grid((int)ceil(horizWindows(bw,fineStrideX) / float(block.x)), 
                  (int)ceil(vertWindows(bh,fineStrideY) / float(block.y)));  
  ClassifyWindows_Kernel<<<grid,block>>>(dScores, dBD, bw, bh);
  cutilSafeCall(cudaGetLastError());
}
