/*
 * ============================================================================
 *
 *       Filename:  windescriptor.cpp
 *
 *    Description:  Provides implementation to windescriptor.h
 *
 *        Version:  1.0
 *        Created:  Sunday 03 June 2007 01:19:54  IST IST
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Navneet Dalal
 *        Company:  Lear / INRIA
 *
 * ============================================================================
 */

#include <iostream>
#include <sstream>
#include <algorithm>
#include <functional>

#include <lear/types.h>

#include <lear/blitz/ext/globalfunc.h>
#include <lear/blitz/ext/tvmutil.h>
#include <lear/blitz/tvmio.h>

#include <lear/io/ioext.h>
#include <lear/io/fileheader.h>

#include <lear/image/rescale.h>
#include <lear/image/imageio.h>

#include <lear/cvision/windescriptor.h>

#ifdef HAVE_CUDA
#include <pdcuda/rhogdense.h>
#include <pdcuda/arrays.h>
#include <pdcuda/transposedarrays.h>
#include <pdcuda/gradients.h>
#include <pdcuda/imresize.h>
#endif

using namespace std;

typedef lear::WinDescriptor::DescCont::const_iterator   DescIter;
typedef lear::WinDescriptor::GridCont::const_iterator   GridIter;
typedef lear::WinDescriptorBase::FeatType               FeatType;

//-----------------------------------------------------------------------------

lear::WinDescriptorBase::WinDescriptorBase(IndexType t_extent_)
  :
  image_extent_(-1,-1),
  extent_(t_extent_)
{
}

// Since we have some pure virtual methods, we must declare at least
// one non-pure virtual method declared elsewhere so that gcc can pick
// a spot to stick the vtable.  Making the virtual destructor
// non-inline solves the problem of having no non-inline non-pure
// virtual methods.
lear::WinDescriptorBase::~WinDescriptorBase() {}

lear::FileHeader lear::WinDescriptorBase::header() const {
  return lear::FileHeader("WinDesc",100);
}

//-----------------------------------------------------------------------------

std::ostream& operator<<( std::ostream& o, lear::WinDescriptorBase const &w) 
{ w.print(o); return o;}

lear::BiOStream& operator<<( lear::BiOStream& o, lear::WinDescriptorBase const &w) 
{ w.print(o); return o;}


//-----------------------------------------------------------------------------

lear::WinDescriptor::WinDescriptor(IndexType t_extent_, 
                                   DescCont const &t_desc_,
                                   GridCont const &t_grid_,
                                   int cachesize) 
  : lear::WinDescriptorBase(t_extent_), 
    desc_(t_desc_), 
    grid_(t_grid_),
    initlength_(0),
    length_(0),
    numItem_(range_cast<int>(desc_.size())),
    indexrange_(numItem_),
    rgbIsCurr(true)
{
  if (range_cast<int>(grid_.size()) != numItem_) {    
    ThrowException("Number of position grid objects " 
                   << " is not equal to descriptor grid objects");
  }
  using namespace std;

  // compute length_ using formula
  DescIter d = desc_.begin();
  GridIter g = grid_.begin();

  int gridlen = 0;
  for (int j=0; j<numItem_; ++j, ++d, ++g) {
    indexrange_(j) = IndexType(length_, range_cast<int>((*d)->size()*g->size()));

    length_ += (*d)->size()*g->size();
    gridlen +=g->size();

    // fill cache also
    cache_.push_back(CacheType((*d)->size(), cachesize));
  }
  initlength_ = length_;
}

void lear::WinDescriptor::setimage(IProcessor::GrayImage const &image) 
{
  currGray.resize(image.extent());
  currGray  = image;
  rgbIsCurr = false;
}

void lear::WinDescriptor::setimage(IProcessor::RGBImage const &image) 
{
  currRgb.resize(image.extent());
  currRgb   = image;
  rgbIsCurr = true;
}

void lear::WinDescriptor::getimage(RGBImage &image, IndexType const &sz) 
{
  image.resize(sz);
  if (rgbIsCurr) {
    image = lear::rescale(currRgb, sz);
  } else {
    ThrowException("Not yet implemented");
    //image = lear::rescale(currGray, sz);
  }
}

void lear::WinDescriptor::preprocess(IndexType topleft, 
                                     IndexType extent,
                                     IndexType newsize) 
{
  if (rgbIsCurr) {
    template_preprocess(currRgb, topleft, extent, newsize);
  } else {
    template_preprocess(currGray, topleft, extent, newsize);
  }
}

template <class PixelType>
lear::WinDescriptor::Preprocessor &
lear::WinDescriptor::template_preprocess(blitz::Array<PixelType,2> const &in,
                                         IndexType topleft, 
                                         IndexType extent,
                                         IndexType newsize) 
{
  using namespace std;
  preprocessor.clear();

  blitz::Array<PixelType,2> image;

  // cropping
  if (blitz::sum(topleft != 0) && blitz::sum(extent != in.extent())) {
    //image =  lear::crop(in, topleft, topleft+extent);
    blitz::Range xrange(topleft[0], topleft[0]+extent[0]-1);
    blitz::Range yrange(topleft[1], topleft[1]+extent[1]-1);
    image.reference
      (blitz::Array<PixelType,2>(const_cast<blitz::Array<PixelType,2> &>(in), 
                                 xrange, yrange));
  } else {
    image.reference(in);
  }

  // rescaling
  if (blitz::sum(newsize != image.extent())) {
    blitz::Array<PixelType,2> tmp(image);
    image.resize(newsize);
    image = lear::rescale(tmp, newsize);
  }
  
  // the rest of the preprocessing
  IndexType image_extent_ = image.extent();
  if (blitz::isLess(image.extent(),extent_)) {
    //                  ostringstream mesg;
    //                  mesg << "Descriptor extent "
    //                      << "( " << setw(4) << extent_[0] << ", " 
    //                      << setw(4) << extent_[1] << " )" 
    //                      << " is bigger than image size "
    //                      << "( " << setw(4) << image.rows() << ", " 
    //                      << setw(4) << image.cols() << " )";

    //                  ThrowException(mesg.str());

    blitz::Array<PixelType,2> nimage = 
      extendBorder(image, extent_/2-1, extent_/2, image_extent_/2);
    for (DescIter i = desc_.begin(); i != desc_.end(); ++i) {
      preprocessor.push_back((*i)->preprocess(nimage));
    }
    image_extent_ = nimage.extent();
  } else {
    for (DescIter i = desc_.begin(); i != desc_.end(); ++i) {
      preprocessor.push_back((*i)->preprocess(image));
    }
  }
  //    for_each(cache_.begin(), cache_.end(),
  //            bind2nd(mem_fun_ref(&CacheType::clear),image_extent_));
  for (CacheCont::iterator iter=cache_.begin(); iter!=cache_.end(); iter++){
    iter->clear(image_extent_);
  }
  return preprocessor;
}

lear::WinDescriptorBase::FeatType 
lear::WinDescriptor::compute(IndexType gridTopLeft) const {
  typedef CacheCont::iterator  CacheIter;

  FeatType vec(initlength_);
  //FeatType::iterator dest = vec.begin();
  ElemType *dest = &*vec.begin();

  DescIter d=desc_.begin(); 
  GridIter g=grid_.begin(); 
  CacheIter c =  cache_.begin();
  Preprocessor::const_iterator p = preprocessor.begin();
  for (; d!=desc_.end(); ++d, ++p, ++c, ++g) {
    DescOp op(*d,*p);
    for (GridType::const_iterator i=g->begin(); 
         i != g->end(); ++i) {
      FeatType f = (*c)(*i+gridTopLeft,op);

      // This copy is performed because the operations return a feature
      // descriptor.  To avoid the copy, I think we'd need to refactor the
      // operation to stuff its results into a slice of vec.  We'd pass the 
      // slice to the cache iter instead of making a copy.  
      //dest = std::copy(f.begin(),f.end(),dest);

      // memcpy is almost instantaneous.  std::copy takes about 50% of the
      // time that the above CacheIter::operator() call requires.
      size_t const nelm   = f.size();
      size_t const nbytes = sizeof(ElemType) * nelm;
      memcpy(dest, &*f.begin(), nbytes);
      dest += nelm;
    }
  }
  return vec;
}

void lear::WinDescriptor::print(std::ostream& o) const {
  using namespace std;
  o << setw(14) << title() 
    << "   Grids Count " << setw(7) << numItem_ << " |\n" 
    << "  | Length    " << setw(9) << length() 
    << "    Extent    " << setw(4) << right << extent_[0] 
    << "x" << setw(4) << left << extent_[1] << right << " |\n";

  int count = 1;
  for (GridIter i = grid_.begin(); i != grid_.end(); ++i,++count) {
    o << "  | Grid Num  " << setw(9) << count 
      << "    Size      " << setw(9) << i->size() <<" |\n";
  }
  o << "  |____________________________________________|\n";
}

void lear::WinDescriptor::print(lear::BiOStream& o) const {
  using namespace std;

  o << header();
  o << range_cast<int32>(numItem_);
  for (DescIter i = desc_.begin(); i != desc_.end(); ++i) {
    o << *(*i);
  }

  o << range_cast<int32>(numItem_);
  for (GridIter i = grid_.begin(); i != grid_.end(); ++i) {
    o << *i;
  }

  o << extent_ << length_;
}

//-----------------------------------------------------------------------------

#ifdef HAVE_CUDA

#include <cutil/cutil_inline.h>

#include <pdcuda/transposedarrays.h>
#include <pdcuda/arrays.h>
#include <pdcuda/gradients.h>
#include <pdcuda/rhogdense.h>

#define CHECK_VAL(v,expected)                                           \
  if (!((v) == (expected))) {                                           \
    ThrowException("Invalid value for " << #expected <<                 \
                   "  Expected: " << expected << "  Actual: " << v);    \
  }

lear::WinDescriptorCuda::WinDescriptorCuda(IndexType        t_extent_, 
                                           DescCont  const &t_desc_,
                                           IndexType        fineStride_,
                                           bool             quantizePos,
                                           bool             useLanczos)
  : WinDescriptorBase(t_extent_), fineStride_(fineStride_), 
    quantizePos(quantizePos), hBDdirty(true), hBDw(-1), hBDh(-1), 
    useLanczos(useLanczos)
{  
  // For now, we extract the processing parameters from the CPU
  // processing units.  For the moment, this seems to be the simplest
  // way to allow GPU processing.  Since the current CUDA
  // implementation locks down many parameter settings, we check that
  // the requested parameters match.

  using namespace pdcuda;

  CHECK_VAL(extent()[0], WinSizeX);
  CHECK_VAL(extent()[1], WinSizeY);
  CHECK_VAL(length(), (size_t)WinDescLen);

  size_t const NumDescriptorTypes = 1;
  CHECK_VAL(t_desc_.size(), NumDescriptorTypes);
  RHOGDense const *desc = *t_desc_.begin();
  CHECK_VAL(desc->extent()[0],   BlockSizeX);
  CHECK_VAL(desc->extent()[1],   BlockSizeY);
  CHECK_VAL(desc->length(),      BlockDescLen);
  CHECK_VAL(desc->stride()[0],   BlockStrideX);
  CHECK_VAL(desc->stride()[1],   BlockStrideY);
  CHECK_VAL(desc->numcell()[0],  NumCellX);
  CHECK_VAL(desc->numcell()[0],  NumCellY);
  CHECK_VAL(desc->orientbin(),   OrientBin);
  CHECK_VAL(desc->semicirc(),    SemiCirc);
  CHECK_VAL(desc->weightScale(), WeightScale);

  if (typeid(*desc->normalizer()) !=
      typeid(L2HysNormalizer<RealType>)) {
    ThrowException("Only L2Hys normalizers are supported right now.");
  }
  L2HysNormalizer<RealType> const *normalizer = 
    static_cast<L2HysNormalizer<RealType> const *>(desc->normalizer());
  epsilon  = normalizer->epsilon();
  maxvalue = normalizer->maxValue();
  epsilon2 = normalizer->epsilon2();

  if (typeid(*desc->processor()) != typeid(GradProcessor_NoSmooth)) {
    ThrowException("Only non-smoothing gradient preprocessors are "
                   "supported right now.");
  }
  GradProcessor_NoSmooth const *preprocessor = 
    static_cast<GradProcessor_NoSmooth const *>(desc->processor());
  CHECK_VAL(preprocessor->remapBefore()->toMethod(), ImageSqrtRemap::Method);
  if ((preprocessor->remapAfter() != NULL) &&
      (typeid(*preprocessor->remapAfter()) != typeid(ImageNoRemap))) {
    ThrowException("No gradient remappers are supported right now.");
  }
  CHECK_VAL(preprocessor->semiCirc(), SemiCirc);
}

#undef CHECK_VAL

void lear::WinDescriptorCuda::setimage(GrayImage const &image) {
  RGBImage rgb(image.extent());
  for (int x=0; x<image.extent()[0]; x++) {
    for (int y=0; y<image.extent()[0]; y++) {
      rgb(x,y) = image(x,y);
    }
  }
  setimage(rgb);
}

void lear::WinDescriptorCuda::setimage(RGBImage const &image) {
  using namespace pdcuda;
  int const wi = image.extent()[0];
  int const hi = image.extent()[1];
  image_extent_ = image.extent();
  if (useLanczos) {
    resizer.reset(new ImageResizerLanczos2D((float const *)image.data(), 
                                            wi, hi, 
                                            lanczosWindowRadius, true, true));
  } else {
    resizer.reset(new ImageResizerBilinear2D((float const *)image.data(), 
                                             wi, hi, true, true));
  }
  hBDdirty = true;
}

void lear::WinDescriptorCuda::getimage(RGBImage &image, IndexType const &sz) 
{
  assert(blitz::all(sz >= 0));
  image.resize(sz);
  pdcuda::TransposedDeviceArray2D<blitz::TinyVector<float,3> > r(&image);
  pdcuda::ScopedDeviceArray<float> dResized(resizer->resize(sz[0], sz[1]));
  cutilSafeCall(cudaMemcpy(r.devicePtr(), dResized.get(),
                           sz[0]*sz[1]*3*sizeof(float),
                           cudaMemcpyDeviceToDevice));
  r.fromDevice();
}


void lear::WinDescriptorCuda::preprocess(IndexType topleft, 
                                         IndexType extent,
                                         IndexType newsize) 
{
  using namespace pdcuda;
  using namespace blitz;

  int const w = newsize[0];
  int const h = newsize[1];
  ScopedDeviceArray<float> dResized
    (resizer->croppedResize(topleft[0],topleft[1], extent[0],extent[1], w,h));
  ScopedDeviceArray<float> dMag(w*h);
  ScopedDeviceArray<float> dOri(w*h);
  computeSpatialGradients(dMag.get(), dOri.get(), dResized.get(), w, h);  
  dBD.reset(computeRhogDense(dMag.get(), dOri.get(), w,h, 
                             fineStride_[0], fineStride_[1],
                             epsilon, maxvalue, epsilon2));
  hBDdirty = true;
  hBDw     = horizBlocks(w,fineStride_[0]);
  hBDh     = vertBlocks(h,fineStride_[1]);
}

FeatType lear::WinDescriptorCuda::compute(IndexType gridTopLeft) const 
{
  using namespace pdcuda;

  // Resize block descriptor data structure if needed
  if (hBDdirty) {
    hBD.resize(NumCellY, NumCellX, OrientBin, hBDh, hBDw);
    cutilSafeCall(cudaMemcpy(hBD.data(), dBD.get(), hBD.size()*sizeof(float),
                             cudaMemcpyDeviceToHost));
    hBDdirty = false;
  }

  // Quantize gridTopLeft, if needed
  if (quantizePos && blitz::sum(fineStride_ != 1)) {
    for (int i=0; i<2; i++) {
      gridTopLeft[i] = (int( floor( (gridTopLeft[i] / float(fineStride_[i])) + 
                                    0.5f ) ) *
                        fineStride_[i]);
    }
  }

  // Convert to hBD coordinates
  IndexType baseBlockPos;
  for (int i=0; i<2; i++) {
    if (gridTopLeft[i] % fineStride_[i] != 0) {
      ThrowException("The position " << gridTopLeft[i] << " does not lie on a "
                     "computed block boundary (regulated by "
                     "fineStride["<<i<<"]="<<fineStride_[i]<<").");
    }
    if (((gridTopLeft[i] + WinSize[i] - BlockSize[i]) / fineStride_[i] + 1) >
        hBD.extent()[3+(1-i)]) {
      ThrowException("The window at " << gridTopLeft[i] << " lies too close "
                     "to the edge of the block descriptor image (extent["
                     <<i<<"]="<<hBD.extent()[3+(1-i)]*fineStride_[i]<<").");
    }
    baseBlockPos[i] = gridTopLeft[i] / fineStride_[i];
  }

  // Extract the feature
  FeatType feature(this->length());
  RealType *data = feature.data();
  for (int bx=0; bx<NumBlockX; bx++) {
    for (int by=0; by<NumBlockY; by++) {
      for (int cx=0; cx<NumCellX; cx++) { 
        for (int cy=0; cy<NumCellY; cy++) {
          for (int o=0; o<OrientBin; o++) {
            *data++ = hBD(cy, cx, o, 
                          baseBlockPos[1] + by*BlockStrideY/fineStride_[1], 
                          baseBlockPos[0] + bx*BlockStrideX/fineStride_[0]);
          }
        }
      }
    }
  }

  return feature;
}

size_t lear::WinDescriptorCuda::length() const { 
  return pdcuda::WinDescLen; 
}

#define PW(X,W) setw(W)<<(X)
#define PXY(X,Y) right << PW((X),4) << "x" << left << PW((Y),4) << right
#define PEXT(EXT) PXY((EXT)[0],(EXT)[1])
#define PB(B,W) PW((B)?"true":"false", W)

void lear::WinDescriptorCuda::print(std::ostream& o) const {
  using namespace std;
  o << PW(title(),20)                           <<"   Grids Count "<<PW(1,7)<< " |\n"
    << "  | Length    "<<PW(length(),9)        <<"    Extent    "<<PEXT(extent_)<<" |\n"
    << "  |                                            |\n"
    << "  | Grid Num  "<<PW(1,9)               <<"    Size      "<<PW(pdcuda::NumBlockX*pdcuda::NumBlockY,9)<<" |\n"
    << "  | Extent    "<<PEXT(pdcuda::NumBlock)<<"    FineStride"<<PEXT(fineStride_)<<" |\n"
    << "  | Epsilon   "<<PW(epsilon,9)         <<"    MaxValue  "<<PW(maxvalue,9)<<" |\n"
    << "  | Epsilon2  "<<PW(epsilon2,9)        <<"    Lanczos?  "<<PB(useLanczos,9)<<" |\n"
    << "  | Lanczos radius "<<PW(lanczosWindowRadius,4)<<PW(" ",4+10+9)<<" |\n"
    << "  |____________________________________________|\n";
}

void lear::WinDescriptorCuda::print(lear::BiOStream& o) const {
  using namespace std;
  using namespace lear;
  using namespace pdcuda;

  o << header();

  o << range_cast<int32>(1);
  o << RHOGDense(CellSize, NumCell, BlockStride, OrientBin,
                 WeightScale, SemiCirc,
                 new GradProcessor_NoSmooth(SemiCirc, NULL, 
                                            new ImageSqrtRemap(), NULL), 
                 new L2HysNormalizer<RealType>(epsilon, maxvalue));
  
  o << range_cast<int32>(1);
  o << lear::DenseGrid<2>::get(WinSize, BlockSize, BlockStride);

  o << extent() << (int32)length();
}

blitz::Array<lear::WinDescriptorBase::RealType,2> 
lear::WinDescriptorCuda::classifyAllRowMajor(LinearClassify const &c) const
{
  using namespace pdcuda;

  int const bw = hBDw;
  int const bh = hBDh;
  int const ww = horizWindows(bw, fineStride_[0]);
  int const wh = vertWindows(bh,  fineStride_[1]);

  blitz::Array<RealType,1> classifier(c.classifier());

  ScopedDeviceArray<float> dScores(ww*wh);
  classifyWindows(dScores.get(), dBD.get(), bw, bh, 
                  fineStride_[0], fineStride_[1], classifier.data(), c.bias());

  blitz::Array<RealType,2> scores(wh,ww); // ROW MAJOR
  cutilSafeCall(cudaMemcpy(scores.data(), dScores.get(), 
                           ww*wh*sizeof(float), cudaMemcpyDeviceToHost));
  return scores;
}

#endif
