/*
* =============================================================================
*
*       Filename:  windetect.cpp
*
*    Description:  Provides implementation to windetect.h interface.
*
*        Version:  1.0
*        Created:  Sunday 03 June 2007 10:35:48  IST IST
*       Revision:  none
*       Compiler:  gcc
*
*         Author:  Navneet Dalal
*        Company:  Lear / INRIA.
*
* =============================================================================
*/

#ifdef WIN32
// For now, suppress warnings related to unsafe functions, ->bool
// conversions, and precision loss.
#pragma warning(disable:4996 4800 4244)
#endif

#include <list>
#include <fstream>
#include <iostream>

#include <blitz/array.h>
#include <blitz/tinyvec.h>

#include <lear/types.h>

#include <lear/numericutil/rng.h>
#include <lear/blitz/ext/tvmutil.h>
#include <lear/blitz/ext/domainiter.h>

#include <lear/io/biostream.h>
#include <lear/io/fileheader.h>
#include <lear/blitz/tvmio.h>
#include <lear/blitz/blitzio.h>

#include <lear/image/rescale.h>
#include <lear/image/imageio.h>

#include <lear/cvision/rhogdense.h>
#include <lear/cvision/iprocessor.h>
#include <lear/cvision/dnormalizer.h>

#include <lear/cvision/densegrid.h>
#include <lear/cvision/imageslider.h>
#include <lear/cvision/scalepyramid.h>
#include <lear/cvision/perspectivescalepyramid.h>
#include <lear/cvision/windescriptor.h>

#include <lear/image/imageio.h>

#include <lear/interface/windetect.h>

#include "../lib/cfg.h"
#ifdef HAVE_CUDA
#include <pdcuda/rhogdense.h>
#endif

#ifdef HAVE_TRACKER
#include <tracker/JTracker.h>
#include <tracker/TrkOutputStream.h>
#include <tracker/stringmanip.h>
#endif

using namespace std;
using namespace blitz;
using namespace lear;

typedef RHOGDenseParam::RealType        RealType;
typedef blitz::Array<RealType,1>        Array1DType;
typedef blitz::Array<RealType,2>        Array2DType;
typedef blitz::Array<RealType,3>        Array3DType;

typedef blitz::TinyVector<int,2>        IndexType;
typedef blitz::TinyVector<RealType,2>   Scale;

typedef lear::DenseGrid<2>::GridType    GridType;

typedef lear::WinDescriptorBase            WinDescBase;
typedef lear::WinDescriptorBase::DescCont  DescContainer;
typedef lear::WinDescriptor::GridCont      GridContainer;

const char WinDetect::DefaultLabel='O';

/**
* A static object of this type will be created as we don't want to export 
* complicated RHOGDense, IProcessor, Normalizer, WinDescriptor interface.
*/
struct WinDetectDescHolder {
  bool initialized;
  DescContainer                descarray;
  GridContainer                gridarray;

  WinDescBase*                 windesc;

  WinDetectDescHolder() :
  initialized(false), windesc(NULL)
  {}

  ~WinDetectDescHolder() {
    delete windesc;
    for (DescContainer::iterator i=descarray.begin(); i!=descarray.end(); ++i) {
      delete *i; 
    }
  }
  static lear::RHOGDense* init(RHOGDenseParam const &param, int verbose=0);
};
static WinDetectDescHolder          descholder;

// creates gridtype and windescriptor.
void WinDetect::init(RHOGDenseParam const *param) 
{
  std::vector<RHOGDenseParam const *> v(1);
  v[0] = param;
  init(v);
}

void WinDetect::init(const std::vector<RHOGDenseParam const *>& param) 
{
  if (descholder.initialized) {
    ThrowException("Init is supposed to be called only once. This is a repeat initialization.");
  }
  IndexType size(size_x, size_y);

  for (unsigned i= 0; i< param.size(); ++i) {
    WinDescBase::DescType* desc = 
      WinDetectDescHolder::init(*param[i], verbose);
    descholder.descarray.push_back(desc);
    descholder.gridarray.push_back
      (lear::DenseGrid<2>::get(size, desc->extent(), desc->stride()));

    boost::shared_ptr<RHOGDenseParam> newParam(new RHOGDenseParam());
    *newParam.get() = *param[i];
    this->params.push_back(newParam);
  }

#ifdef HAVE_CUDA
  if (useCuda) {
    IndexType const fineStride((finestride_x>0) ? finestride_x : winstride_x,
                               (finestride_y>0) ? finestride_y : winstride_y);
    descholder.windesc = 
      new WinDescriptorCuda(size, descholder.descarray, 
                            fineStride, quantizepos);
  } else {
#endif
    descholder.windesc = 
      new WinDescriptor(size, descholder.descarray, 
                        descholder.gridarray, cachesize);
#ifdef HAVE_CUDA
  }
#endif

  descholder.initialized = true;
  if (verbose > 1) { cout << *descholder.windesc << endl; }
} 

lear::RHOGDense* WinDetectDescHolder::init(RHOGDenseParam const &param, int verbose) 
{ 
  using namespace lear;

  ImageNoRemap* mapperBefore = NULL;// remap image before computing gradient
  ImageNoRemap* mapperAfter = NULL;// remap gradient magnitude 
  switch (param.preprocessing2use) 
  {
    // gradient operator
  case RHOGDenseParam::RGB_Log_Grad:
    mapperBefore = new ImageLogRemap();
    break;

  case RHOGDenseParam::RGB_Sqrt_Grad:
    mapperBefore = new ImageSqrtRemap();
    break;

  case RHOGDenseParam::Lab_Sqrt_Grad:
    mapperBefore = new ImageLabSqrtRemap();
    break;

  case RHOGDenseParam::Lab_Grad:
    mapperBefore = new ImageLabRemap();
    break;

  case RHOGDenseParam::RGB_Grad:
  default:
    break;
  };

  IProcessor* preprocessor=NULL;
  if (param.gscale < 1e-3) {
    // It owns and deletes all pointers. So no worries.
    preprocessor = new GradProcessor_NoSmooth(param.semicirc, NULL, mapperBefore, mapperAfter);
  } else {
    // It owns and deletes all pointers. So no worries.
    preprocessor = new GradProcessor(param.gscale, param.semicirc, NULL, mapperBefore, mapperAfter);
  }
  DescNormalizer<RealType>* normalizer=NULL;
  switch (param.norm2use) 
  {
  case RHOGDenseParam::NoNorm:
    normalizer = new DescNormalizer<RealType>();
    break;

  case RHOGDenseParam::NormL1:
    normalizer = new L1Normalizer<RealType>(param.epsilon);
    break;

  case RHOGDenseParam::NormL2:
    normalizer = new L2Normalizer<RealType>(param.epsilon);
    break;

  case RHOGDenseParam::NormL2Hys:
    normalizer = new L2HysNormalizer<RealType>(param.epsilon, param.maxvalue);
    break;

  case RHOGDenseParam::NormL1Sqrt:
  default:
    normalizer = new L1SqrtNormalizer<RealType>(param.epsilon);
    break;
  }
  IndexType cellsize(param.cellsize_x, param.cellsize_y); 
  IndexType numcell(param.numcell_x, param.numcell_y); 
  IndexType stride(param.descstride_x, param.descstride_y); 

  WinDescBase::DescType* desc = new RHOGDense(
    cellsize, numcell, stride, 
    param.orientbin, param.wtscale, param.semicirc, 
    preprocessor, normalizer);// owns preprocessor & normalizer and frees them when they not required

  if (verbose > 1) std::cout << *desc << std::endl;
  return desc;
}

// Convert image to blitz array format
static WinDescBase::ImageType getImage(
  const unsigned char* image, int width, int height, int step) 
{
  using namespace lear;

  WinDescBase::ImageType bimage(width, height) ;
  if (step==0)
    step = 3*width;

  for (int j= 0; j< height; ++j) {
    for (int i= 0; i< width; ++i) {
      const unsigned char* pixel = image+j*step + 3*i;
      bimage(i,j) = IProcessor::RGBType(pixel[0], pixel[1], pixel[2]);
    }
  }
  //ImageIO::write("image.png", bimage);
  return bimage;
} 

bool WinDetect::computefeature(
  float* result, int xloc, int yloc, 
  const unsigned char* image, int width, int height, int step) const
{
  if (!descholder.initialized) {
    ThrowException("Init is supposed to be called before calling computefeature");
  }
  WinDescBase* windesc = descholder.windesc;

  IndexType descextent = windesc->extent();

  if (xloc >= 0 && yloc >= 0 && 
      xloc+descextent[0] <= width && 
      yloc+descextent[1] <= height)
  {
    windesc->setimage(getImage(image,width,height, step)); 

    windesc->preprocess(width, height);
    Array1DType desc = windesc->compute(IndexType(xloc,yloc)); 
    std::copy(desc.begin(), desc.end(), result);
    return true;
  }
  return false;
}

std::vector<bool> WinDetect::computefeature(
  std::vector<float*>& result, 
  std::vector<int> const &xlocs, std::vector<int> const &ylocs, 
  const unsigned char* image, int width, int height, int step) const
{
  std::vector<bool> status(xlocs.size());
  if (!descholder.initialized) {
    ThrowException("Init is supposed to be called before calling computefeature");
  }

  WinDescBase* windesc = descholder.windesc;

  IndexType descextent = windesc->extent();

  windesc->setimage(getImage(image,width,height, step)); 
  windesc->preprocess(width, height);

  for (unsigned i= 0; i< xlocs.size(); ++i) {
    IndexType topleft (xlocs[i], ylocs[i]);
    if (topleft[0] >=0 && topleft[1] >=0 && 
      topleft[0]+descextent[0] <= width && 
      topleft[1]+descextent[1] <= height)
    {
      Array1DType desc = windesc->compute(topleft); 
      std::copy(desc.begin(), desc.end(), result[i]);
      status[i] = true;
    } else 
      status[i] = false;
  }
  return status;
}


size_t WinDetect::featurelength() const
{
  if (!descholder.initialized) {
    ThrowException("Init is supposed to be called before calling featurelength");
  }
  return descholder.windesc->length();
}

// Used to create jitter in windows in order to learn more robust
// detector.  This is test code with not so good results from it. So
// you may ignore it.
class JitterWindow {
public:
  typedef std::vector<IndexType>::const_iterator iterator;

  JitterWindow(IndexType winextent_, IndexType jitter_, int steps_) 
    : winextent_(winextent_), jitter_(jitter_), steps_(steps_), 
      samples_(steps_*8+1) // GED: What's 8: fudge factor? 8-neighbors? cell size? block stride? win stride?
  {
    ptlist_.reserve(samples_);
  }

  void generatepointlist(IndexType imageextent, IndexType startloc) {
    ptlist_.clear();

    for (blitz::DomainIter<2> i(-steps_,steps_); i; ++i) {
      IndexType start = startloc + (*i)*jitter_;
      IndexType end(start+winextent_);
      if ((blitz::sum(start>=0)==2) && (blitz::sum(end<=imageextent)==2)) {
        ptlist_.push_back(start);
      }
    }
    samples_ = ptlist_.size();

    if (samples_ > 1) {
      cerr << "MULTIPLE JITTER SAMPLES!" << endl;
      exit(1);
    }
  }

  iterator begin()   const { return ptlist_.begin(); }
  iterator end()     const { return ptlist_.end();   }
  int      samples() const { return samples_;        }

private:
  IndexType winextent_;
  IndexType jitter_;
  int steps_,samples_;
  std::vector<IndexType> ptlist_;
};

#ifdef BUILD_APP
static void writeExtra(WinDetectDump const &param,
                       lear::BiOStream& to,// output file stream
                       IndexType const lbound, IndexType const extent,
                       RealType const scale, std::string const &filename)
{
  to << lbound << extent << scale;
  if (param.dumpfulldetail>1) to << filename;
}

void WinDetectDump::writePresetExamples(std::string const &hardfile,
                                        std::string const &outfile,
                                        bool suppressNonJittered)
{
  if (!descholder.initialized) {
    ThrowException("Init is supposed to be called before we can use writePresetExamples");
  }
  WinDescBase* windesc = descholder.windesc;
  if (verbose > 1) { std::cout << *this << std::endl; }

  if (verbose > 2) {
    std::cout << "Writing hard file for detection score > " 
              << hardscore << std::endl;
  }
  int totallines = 0;
  {
    std::ifstream from(hardfile.c_str());
    if (!from) {
      ThrowException("Unable to open hard test case file " + hardfile);
    }
    char line[1024];
    while (from.getline(line,1024)) {
      totallines++;
    }
  }
  std::ifstream from(hardfile.c_str());
  if (!from) {
    ThrowException("Unable to open hard test case file " + hardfile);
  }

  if (from && verbose > 2) {
    cout << "From stream ("<<hardfile<<") is good again and can read" << std::endl;
  }
  IndexType size(size_x, size_y); // size of window, in pixels
  IndexType jitterstep(jitterstep_x, jitterstep_y);

  JitterWindow jitwin(size, jitterstep, jitterwinstep);
  if (memorylimit) {
    size_t descsize = windesc->length()*sizeof(RealType);// in bytes now

    if (verbose > 1) {
      cout << "Upper memory limit is " << memorylimit << " MB"
           << ", Descriptor size (in bytes) is " << descsize << endl;
    }

    size_t maxdescriptor = size_t(memorylimit)*1024*1024/descsize;
    if (poscases) maxdescriptor -= poscases;
    if (negcases) maxdescriptor -= negcases;

    samples = totallines*jitwin.samples()/maxdescriptor;
    if (verbose > 1) {
      cout << "Total lines are " << totallines 
           << ", jitter windows are " << jitwin.samples()
           << ", maximum descriptors are " << maxdescriptor 
           << ", samples value " << samples << endl;
    }
  }

  lear::BiOStream to(outfile.c_str());
  if (!to) {
    ThrowException("Unable to open output file " + outfile);
  }
  lear::FileHeader header("RawDesc",100);
  to << header;

  to << *windesc;

  lear::BiOStream::pos_type const countFilePos = to.tellp();
  int count = 0;
  // number of elements
  to << count;

  IndexType newsize;
  RealType prevscale=0;
  std::string prevfile="";

  IndexType topleft(topleft_x, topleft_y);

  bool hasTopLeft = false;
  if (blitz::sum(topleft>=0) == 2) {
    hasTopLeft = true;
    if (verbose > 2) {
      cout << "Top-Left corner specified ( " 
           << topleft[0] << ", " << topleft[1] << " ) " << endl;
    }
  }

  bool negcase = (label!='P');
  size_t skipped = 1;
  WinDescBase::ImageType origimage;
  while (from) {
    std::string file;
    from >> file;
    IndexType lbound, extent;
    RealType scale, score;
    // TODO: I think the scale is superfluous
    from >> lbound[0] >> lbound[1] >> extent[0] >> extent[1] >> scale >> score;
    if (!from) break;
    if (file.empty()) continue;

    if ((negcase && score < hardscore) || (!negcase && score > hardscore)) {
      continue;
    }

    bool newimage;
    if (prevfile != file){
      lear::ImageIO::read(file,origimage);
      newimage = true;
      prevfile = file;
    } else {
      newimage = false;
    }
    if (verbose > 3) {
      cout << "Processing file " << file << " at " 
           << "(" << right << setw(4) << lbound[0] << "," 
           << left << setw(4) << lbound[1] << ") + " 
           << "(" << right << setw(4) << extent[0] << "x" 
           << left << setw(4) << extent[1] << "), " 
           << "with scale " << scale << " and score " << score
           << endl;
    }
    IndexType start(lbound/scale);
    //if (hasTopLeft) {
    //    start += topleft;
    //} else 
    {
      // extent difference, can have one-two pixel error
      IndexType exdiff = extent/scale - size; 
      exdiff = max(exdiff,0); // in case, there is small error due 
      // to rounding off while writing extent
      start += exdiff/2;
    }
    IndexType end(start+size);

    bool reprocess = false;
    if (newimage || std::abs(scale - prevscale) > 1e-3) {
      prevscale = scale;
      reprocess = true;
    }
    if (newimage) {
      windesc->setimage(origimage);
    }
    IndexType newsize(ceil(origimage.extent()/scale));
    if (reprocess) windesc->preprocess(newsize);

    // if we are looking only on a sub window of hard example
    
    if (blitz::sum(end >= newsize)) continue;
    
    jitwin.generatepointlist(newsize, start);
    for (JitterWindow::iterator ji = jitwin.begin(); 
         ji!=jitwin.end();++ji) {
      if (samples && skipped <= samples){
        ++skipped;
        if (verbose >= 9) {
          cout << "window skipped due to sampling" << endl;
        }
        continue;
      } else {
        skipped = 1;
      }
      
      if (verbose >= 8)  {
        std::cout << "Computing descriptor at location " 
                  << *ji << ", " << newsize << std::endl;
      }
      
      if (suppressNonJittered && !negcase && blitz::all(*ji==start)) continue;
      if (blitz::sum(*ji + size >= newsize)) continue;

      to << windesc->compute(*ji);
      ++count;
    }
    
    if (verbose > 5) {
      cout << "Processed location " << file << std::endl;
    }
  }
  to.seekp(countFilePos);
  to << count;
  to.close();

  if (verbose > 2) cout << "Written " << count << " features" << std::endl;
}

void WinDetectDump::writeWinDesc(InputImageIterator &infiles,
                                 std::string const &outfile,
                                 std::string const &testlocs)
{
  using namespace lear;
  
  if (!descholder.initialized) {
    ThrowException("Init is supposed to be called before we can use writeWinDesc");
  }
  WinDescBase* windesc = descholder.windesc;
  if (verbose > 1) { std::cout << *this << std::endl; }

  lear::BiOStream to(outfile.c_str());
  if (!to) {
    ThrowException("Unable to open output file " + outfile);
  }
  lear::FileHeader header("RawDesc",100+(dumpfulldetail*10));
  to << header;

  to << *windesc;

  lear::BiOStream::pos_type start = to.tellp();
  lear::int32 count = 0;
  // number of elements
  to << count; 

  std::ofstream testlocsstr;
  bool dotestlocs = testlocs.size();
  if (dotestlocs) {
    testlocsstr.open(testlocs.c_str());
    if (!testlocsstr) {
      std::cerr << "Could not open file to write tested/dumped locations " << testlocsstr << ". Ignoring ..." << std::endl;
      dotestlocs = false;
    }
  }


  lear::numericutil::RandomGenerator rng( randomSeed ? static_cast<unsigned long> (time(0)) : 0 );

  IndexType topleft(topleft_x, topleft_y);
  bool const hasTopLeft = (blitz::sum(topleft>=0) == 2);
  if (hasTopLeft && verbose > 2) {
    cout << "Top-Left corner specified ( " 
         << topleft[0] << ", " << topleft[1] << " ) " << endl;
  }

  IndexType fullsize(fullsize_x, fullsize_y);
  bool const hasFullSize = (blitz::sum(fullsize>=0) == 2);
  if (hasFullSize && verbose > 2) {
    cout << "Full-Window size specified ( " 
         << fullsize[0] << ", " << fullsize[1] << " ) " << endl;
  }

  IndexType const size(size_x, size_y);
  IndexType const winstride(winstride_x, winstride_y);
  IndexType const winsize = (hasFullSize && hasTopLeft)? fullsize: size;

  IndexType jitterstep(jitterstep_x, jitterstep_y);
  JitterWindow jitwin(size, jitterstep, jitterwinstep);
  while (infiles.next()) {
    WinDescBase::ImageType const &image = infiles.currImage();

    if (verbose > 3) {
      cout << "Processing file " << infiles.currName() << endl;
    }
    if (blitz::sum(image.extent()<winsize)) {
      std::cerr <<  "Image extent "
        << "( " << std::setw(4) << image.rows() << ", " 
        << std::setw(4) << image.cols() << " )"
        << " is smaller than descriptor size "
        << "( " << std::setw(4) << winsize[0] << ", " 
        << std::setw(4) << winsize[1] << " ). Ignoring ..." ;
      continue;
    }

    windesc->setimage(image);
    if (pyramid) {
      typedef lear::ImageSlider<2>                SliderType;
      typedef lear::ScalePyramid<2>               PyramidType;

      const PyramidType pyramid(image.extent(),size,
        scaleratio, 
        endscale, 
        startscale);

      if (verbose > 6) {
        cout << "Scale levels" << setw(2) << pyramid.size() <<
          ", Start scale = " << 
          setw(3) << setprecision(1) << pyramid.startScale() <<
          ", End scale = " <<
          setw(3) << setprecision(1) << pyramid.endScale() <<
          endl;
      }

      int imagewindows = 0;
      for (PyramidType::iterator piter = pyramid.begin(); 
        piter != pyramid.end(); ++piter) 
      {
        windesc->preprocess(*piter);
        SliderType slider(*piter, size, winstride);
        for (SliderType::iterator siter = slider.begin(); 
          siter != slider.end(); ++siter) 
        {
          to << windesc->compute(*siter); ++imagewindows;
          if (dotestlocs){
            testlocsstr << infiles.currName() 
              << ' ' << setprecision(4) << setw(10) << (*siter)[0] 
              << ' ' << setprecision(4) << setw(10) << (*siter)[1] 
              << ' ' << setprecision(4) << setw(10) << size[0]*piter.scale() 
              << ' ' << setprecision(4) << setw(10) << size[1]*piter.scale() 
              << std::endl;
          }
        }
      }
      if (verbose > 1) {
        cout << "Image " << infiles.currName() << " | Descriptors " 
             << setw(8) << imagewindows << endl;
      }
      count += imagewindows;
    } else { // no pyramid
      windesc->preprocess(image.extent());
      blitz::TinyVector<unsigned long,2> ex;
      ex = image.extent() - winsize;
      if (hasTopLeft) ex -= topleft;

      for (size_t i=0; i<samples; ++i) {
        IndexType tl;
        if (samples > 1) {
          tl[0] = static_cast<int>( rng.uniform(ex[0]) );
          tl[1] = static_cast<int>( rng.uniform(ex[1]) );

          if (hasTopLeft) tl += topleft;
        } else if (hasTopLeft) {
          tl = topleft;
          if (hasFullSize) tl += (image.extent() - fullsize)/2;
        } else {
          tl = (image.extent() - size)/2;
        }

        jitwin.generatepointlist(image.extent(), tl);
        for (JitterWindow::iterator ji = jitwin.begin(); 
          ji!=jitwin.end();++ji) 
        {
          to << windesc->compute(*ji); ++count;
          if (dumpfulldetail) {
            writeExtra(*this, to,*ji, size, 1, infiles.currName());
          }
          if (dotestlocs){
            testlocsstr << infiles.currName() 
              << ' ' << setprecision(4) << setw(10) << (*ji)[0] 
              << ' ' << setprecision(4) << setw(10) << (*ji)[1] 
              << ' ' << setprecision(4) << setw(10) << size[0]
              << ' ' << setprecision(4) << setw(10) << size[1]
              << std::endl;
          }
        }
      }
    }

    if (verbose > 5) cout << "Processed file " << infiles.currName() << std::endl;
  }
  to.seekp(start);
  to << count;

  if (verbose > 2) cout << "Written " << count << " features" << std::endl;
}
#endif



#include <string>
#include <cstdio>
#include <fstream>

#include <lear/util/fileutil.h>
#include <lear/blitz/ext/globalfunc.h>

#include <lear/classifier/resultholder.h>
#include <lear/classifier/th_processresult.h>
#include <lear/classifier/ms_processresult.h>

/**
* A static object. It hides the complexity of determining the best object
* locations from the main interface.
*/
struct WinDetectClassifyHolder {
  bool initialized;
  lear::MeanShift_ProcessResult *holder;

  WinDetectClassifyHolder() : initialized(false), holder(NULL) {}
  ~WinDetectClassifyHolder() { delete holder; }
};
static WinDetectClassifyHolder      classifierholder;


static inline void fillDetectInfo(lear::DetectInfo &d,
                                  RealType const score,
                                  RealType const scale,
                                  IndexType const lbound, // scaled top-left corner
                                  IndexType const extent) // scaled size
{
  d.score = float(score);
  d.scale = float(scale);
  for (int i=0; i<2; i++) {
    RealType const sz = extent[i]*scale;
    RealType const center = scale * (lbound[i] + extent[i]/2);
    d.lbound[i] = int(ceil(center - sz/2));
    d.extent[i] = int(floor(sz));
    d.orig_lbound[i] = lbound[i];
  }

  /*
  typedef blitz::TinyVector<RealType,2> Real2DType;
  Real2DType const s(extent*scale);
  Real2DType const c(scale*(lbound+extent/2));

  return lear::DetectInfo(
    static_cast<float>(score),
    static_cast<float>(scale),
    blitz::ceil(c - s/2), blitz::floor(s), lbound);
  */
}


void DetectedRegion::print(std::ostream& o) const
{
  using namespace std;
  o   << setw(6) << x << ' ' << setw(6) << y << ' '
      << setw(6) << width << ' ' << setw(6) << height << ' '
      << setw(9) << setprecision(5) << scale << ' ' 
      << setw(13) << setprecision(8) << scientific << score<<fixed;
}

std::ostream& operator<<(std::ostream& o, DetectedRegion const &region){
  region.print(o);
  return o;
}

void WinDetectClassify::initclassifier() 
{ 
  using namespace lear;
  using std::setw; using std::setprecision;

  if (classifierholder.initialized) {
    ThrowException("Init is supposed to be called only once. This is a repeat initialization.");
  }

  typedef blitz::TinyVector<RealType,2>                SigmaType;
  SigmaType score2prob(score2prob_a, score2prob_b);

  typedef blitz::TinyVector<RealType,3>                NonmaxType;
  NonmaxType nonmaxSigma(nonmaxsigma_x,nonmaxsigma_y, nonmaxsigma_scale); 

  IndexType size(size_x, size_y);

  MeanShift_ProcessResult* holder = new MeanShift_ProcessResult(
    size, lightthreshold, threshold, 
    score2prob, nonmaxSigma, softmax, verbose);

  if (verbose > 1)
    std::cout << "Processor " << holder->toString() << std::endl; 

  classifierholder.holder = holder;
  classifierholder.initialized = true;
}

void WinDetectClassify::test(LinearClassify const &classifier,
                             std::list<DetectedRegion>& detections,
                             const unsigned char* imagedata, 
                             int width, int height, int step) const
{
  using namespace lear;
  if (!descholder.initialized) {
    ThrowException("Init is supposed to be called before we can use test");
  }
  if (!classifierholder.initialized) {
    ThrowException("Init is supposed to be called before we can use test");
  }

  WinDescBase* windesc = descholder.windesc;
  MeanShift_ProcessResult& holder = *(classifierholder.holder);
  if (verbose > 1) 
  { std::cout << *this << std::endl; }

  // create classifier
  if ((size_t)classifier.length() != (size_t)windesc->length()) {
    std::cerr << "Classifier length " << 
      classifier.length() << std::endl;
    ThrowException("Dimension mismatch between window feature vector "
                   "and SVM learned feature vectors");
  }


  IndexType winsize(size_x, size_y);
  IndexType winstride(winstride_x, winstride_y);

  IndexType tmargin(margin_x, margin_y);
  IndexType tavsize(avsize_x, avsize_y);
  bool addmargin ( blitz::sum((tmargin*tavsize) > 0));

  WinDescBase::ImageType image;
  IndexType toadd = 0;

  WinDescBase::ImageType origimage = getImage(imagedata,width,height, step);
  if (addmargin) {
    IndexType newext = origimage.extent();
    IndexType toaddX = 0, toaddY = 0;
    if (tavsize[0])
      toaddX = newext*tmargin/tavsize[0];
    if (tavsize[1])
      toaddY = newext*tmargin/tavsize[1];

    toadd = blitz::max(toaddX,toaddY);
    newext += 2*toadd;

    image.reference(extendBorder(
      origimage, newext/2-1, newext/2, origimage.extent()/2));
  } else {
    image.reference(origimage);
  }

  int imagewindows = 0; 
  holder.clear();

  typedef lear::ImageSlider<2>                SliderType;
  typedef lear::ScalePyramid<2>               PyramidType;

  const PyramidType pyramid(image.extent(),winsize,
    scaleratio, endscale, startscale);

  if (verbose > 6) {
    cout << "Scale levels" << setw(2) << pyramid.size() <<
      ", Start scale = " << 
      setw(6) << setprecision(2) << pyramid.startScale() <<
      ", End scale = " <<
      setw(6) << setprecision(2) << pyramid.endScale() <<
      endl;
  }

  windesc->setimage(image);
  for (PyramidType::iterator piter = pyramid.begin(); 
    piter != pyramid.end(); ++piter) 
  {
    windesc->preprocess(*piter);
    SliderType slider(*piter,winsize,winstride);
    for (SliderType::iterator siter = slider.begin(); 
      siter != slider.end(); ++siter) 
    {
      IndexType tl=*siter;

      Array1DType desc = windesc->compute(tl); 

      DetectInfo r;
      fillDetectInfo(r, classifier(desc.data()),piter.scale(),
                     tl, windesc->extent());
      r.lbound -=toadd;

      holder(r);
      ++imagewindows;
    }
  }

  if (verbose > 3) {
    cout << "Processed " << setw(5) << imagewindows << " windows" <<  endl;
  }
  detections.clear();
  holder.doit();

  MeanShift_ProcessResult::const_iterator s = holder.begin();
  MeanShift_ProcessResult::const_iterator e = holder.end();
  for (; s!= e; ++s) {
    detections.push_back(DetectedRegion( s->score, s->scale, 
      s->lbound[0], s->lbound[1], s->extent[0], s->extent[1]));
  }
}



#ifdef BUILD_APP
#include <lear/classifier/mask_processresult.h>
#include <lear/classifier/hist_processresult.h>
#include <lear/classifier/hard_ppresult.h>
#include <lear/classifier/list_ppresult.h>
#include <lear/classifier/markimage.h>
#include <lear/classifier/drawbestthresh.h>
#include <lear/classifier/aligninimage.h>
#include <lear/classifier/tracker_ppresult.h>

//ofstream F("t.csv");

static inline
void recordOneClassification(RealType              score,
                             IndexType      const &size, 
                             IndexType      const &windescExt,
                             IndexType      const &tl,     // scaled img coords
                             Scale          const &scale, 
                             IndexType      const &toadd,  // original img coords
                             lear::ResultHolder   &holder, 
                             bool                  dotestlocs, 
                             std::string    const &f, 
                             int                  &imagewindows, 
                             std::ofstream        &testlocsstr)
{
  using namespace lear;
  DetectInfo r;
  fillDetectInfo(r, score, blitz::max(scale), tl, windescExt);
  r.lbound -= toadd;

  //F << tl[0] << "," << tl[1] << "," << scale << "," << r.score << "\n"; F.flush();
  
  holder(r);
  if (dotestlocs){
    using std::setw; using std::setprecision;
    testlocsstr << f
      << ' ' << setprecision(4) << setw(10) << tl[0]*scale - toadd[0]
      << ' ' << setprecision(4) << setw(10) << tl[1]*scale - toadd[1]
      << ' ' << setprecision(4) << setw(10) << size[0]*scale
      << ' ' << setprecision(4) << setw(10) << size[1]*scale
      << std::endl;
  }
  ++imagewindows;
}

static inline
void runOneDescriptor(
  IndexType              const &size, 
  WinDescBase            const *windesc, 
  IndexType              const &tl,     // scaled img coords
  LinearClassify         const &classifier, 
  Scale                  const &scale, 
  IndexType              const &toadd,  // original img coords
  lear::ResultHolder           &holder, 
  bool                          dotestlocs, 
  std::string            const &f, 
  int                          &imagewindows, 
  std::ofstream                &testlocsstr)
{
  recordOneClassification(classifier(windesc->compute(tl).data()), 
                          size, windesc->extent(),
                          tl, scale, toadd, holder, dotestlocs, f,
                          imagewindows, testlocsstr);
}

// This function is a quick-and-dirty extractions from runImageSlider.  We 
// might want to encapsulate these huge parameter lists in a struct.
inline bool runDegenerateImageSlider(
  WinDescBase::ImageType       &image, 
  IndexType              const &initialExt, 
  bool                          hasTopLeft, 
  bool                          hasFullSize,
  IndexType              const &topleft, 
  IndexType              const &fullsize, 
  IndexType              const &size, 
  IndexType              const &winsize,
  WinDescBase                  *windesc, 
  LinearClassify         const &classifier, 
  IndexType              const &toadd, 
  lear::ResultHolder           &holder, 
  bool                          dotestlocs, 
  std::string            const &f, 
  int                          &imagewindows, 
  std::ofstream                &testlocsstr) 
{
  IndexType tl;
  if (hasTopLeft) {
    tl = topleft;
    if (hasFullSize) tl += (initialExt - fullsize)/2;
  } else {
    tl = (initialExt - size)/2;
  }
  if (blitz::any(initialExt < winsize+tl)) {
    std::cerr <<  "Image extent "
      << "( " << std::setw(4) << initialExt[0] << ", " 
      << std::setw(4) << initialExt[1] << " )"
      << " is smaller than descriptor size "
      << "( " << std::setw(4) << winsize[0] << ", " 
      << std::setw(4) << winsize[1] << " ). Ignoring ..." ;
    return false;
  }
  windesc->setimage(image);
  windesc->preprocess(initialExt);

  runOneDescriptor(size, windesc, tl, classifier, 1, 
    toadd, holder, dotestlocs, f, imagewindows, testlocsstr);

  return true;
}

// This function is a quick-and-dirty extractions from runImageSlider.  We 
// might want to encapsulate these huge parameter lists in a struct.
inline void runPyramidImageSlider(
  WinDescBase::ImageType       &image, 
  IndexType              const &initialExt,
  bool                          hasTopLeft, 
  bool                          hasFullSize,
  IndexType              const &topleft, 
  IndexType              const &size, 
  IndexType              const &winsize,
  IndexType              const &winstride,
  WinDescBase                  *windesc, 
  LinearClassify         const &classifier, 
  IndexType              const &toadd, 
  lear::ResultHolder           &holder, 
  bool                          dotestlocs, 
  std::string            const &f, 
  int                          &imagewindows, 
  std::ofstream                &testlocsstr,
  RealType                      scaleratio,
  RealType                      endscale,
  RealType                      startscale,
  int                           verbose) 
{
  typedef lear::ImageSlider<2>   SliderType;
  typedef lear::ScalePyramid<2>  PyramidType;

  PyramidType const pyramid(initialExt, winsize,
    scaleratio, endscale, startscale);

  if (verbose > 6) {
    cout << "Scale levels" << setw(2) << pyramid.size() <<
      ", Start scale = " << 
      setw(6) << setprecision(2) << pyramid.startScale() <<
      ", End scale = " <<
      setw(6) << setprecision(2) << pyramid.endScale() <<
      endl;
  }

  windesc->setimage(image);
  for (PyramidType::iterator piter = pyramid.begin(); 
       piter != pyramid.end(); ++piter) 
  {
    windesc->preprocess(*piter);
    SliderType slider(*piter, winsize, winstride);
    for (SliderType::iterator siter = slider.begin(); 
         siter != slider.end(); ++siter) 
    {
      IndexType tl = *siter;
      if (hasTopLeft && hasFullSize) {
        tl += topleft;
      }
      runOneDescriptor(size, windesc, tl, classifier, piter.scale(), 
        toadd, holder, dotestlocs, f, imagewindows, testlocsstr);
    }
  }
}

typedef lear::PerspectiveScalePyramid::TopRows TopRows;

static inline
void runPerspectiveRowsGivenScale(TopRows        const &sTopRow, 
                                  IndexType      const  winsize, 
                                  IndexType      const  sExt, 
                                  IndexType      const  uLCropped, 
                                  Scale          const &scale, 
                                  IndexType      const  winstride, 
                                  IndexType      const  size, 
                                  WinDescBase          *windesc, 
                                  LinearClassify const &classifier, 
                                  lear::ResultHolder   &holder, 
                                  bool                  dotestlocs, 
                                  std::string    const &f, 
                                  int                  &imagewindows, 
                                  std::ofstream        &testlocsstr)
{
  typedef lear::ImageSlider<2>  SliderType;

  for (TopRows::const_iterator riter=sTopRow.begin(); 
       riter!=sTopRow.end(); ++riter)
  {
    int const sTopRow = *riter;
    // Skip windows that run off the image vertically
    if (sTopRow < 0) continue;
    if (sTopRow + winsize[1] > sExt[1] + uLCropped[1]/scale[1]) continue;

    // Just scan this individual row
    IndexType const sLRow = IndexType(0, sTopRow) - uLCropped/scale[0];
    SliderType slider(IndexType(sExt[0], winsize[1]), 
      winsize, winstride);
    for (SliderType::iterator siter = slider.begin(); 
      siter != slider.end(); ++siter) 
    {
      IndexType sLWin = *siter + sLRow;
      if ((sLWin[0] < 0) || (sLWin[1] < 0)) continue;
      if (sLWin[0]+winsize[0] > sExt[0]) continue;
      if (sLWin[1]+winsize[1] > sExt[1]) continue;

      runOneDescriptor(size, windesc, sLWin, classifier, scale, 
        -uLCropped, holder, dotestlocs, f, imagewindows, testlocsstr);
    }
  }
}

#ifdef HAVE_CUDA
static inline
void runPerspectiveRowsGivenScale_CUDA(IndexType      const  uLCropped, 
                                       Scale          const &scale, 
                                       IndexType      const  size, 
                                       WinDescriptorCuda    *windescCuda,
                                       LinearClassify const &classifier, 
                                       lear::ResultHolder   &holder, 
                                       bool                  dotestlocs, 
                                       std::string    const &f, 
                                       int                  &imagewindows, 
                                       std::ofstream        &testlocsstr)
{
  // For now, we'll be roughly consistent with the non-CUDA version.
  // We sample along the X axis at the zoomed scale, but sample at the
  // original scale for the Y axis.  Note that the non-CUDA version
  // centers the slider.  We don't because that would make the call
  // and implementation of windescCuda->classifyAllRowMajor more
  // challenging (since windows wouldn't be aligned on WinStride
  // boundaries).
  //
  // Note that the non-CUDA version often evaluates fewer windows for
  // small scale precision values because it can skip scales for some
  // rows.
  blitz::Array<RealType,2> scores(windescCuda->classifyAllRowMajor(classifier));
  int const ww = scores.extent()[1];
  int const wh = scores.extent()[0];
  RealType const ystride = (1.0f / scale[1]);
  for (RealType yf=0; yf<wh; yf+=ystride) { 
    int const y = int(yf);
    for (int x=0; x<ww; x++) {
      IndexType const sLWin(x * pdcuda::WinStrideX, y * pdcuda::WinStrideY);
      RealType const score = scores(y,x);  // ROW MAJOR for now
      recordOneClassification(score, size, windescCuda->extent(), sLWin, scale,
        -uLCropped, holder, dotestlocs, f, 
        imagewindows, testlocsstr);
    }
  }
}
#endif

// This function is a quick-and-dirty extractions from runImageSlider.  We 
// might want to encapsulate these huge parameter lists in a struct.
static inline 
void runPerspectiveSlider(
  WinDescBase::ImageType               &image, 
  IndexType                      const &initialExt,
  bool                                  hasTopLeft, 
  bool                                  hasFullSize,
  IndexType                      const &size, 
  IndexType                      const &winsize,
  int                            const borderPixels,     // winsize[1] = borderPixels*2 + person height in pixels
  IndexType                      const &winstride,
  RealType                       const  pedestrianHeight, // canonical pedestrian height (world coordinates)
  blitz::TinyVector<RealType,3>  const &rodrigues,        // world-to-camera Rodrigues rotation vector
  blitz::TinyVector<RealType,3>  const &translation,      // world-to-camera translation
  blitz::TinyVector<RealType,2>  const &focalLength,      // focal length of the camera (pixels)
  WinDescBase                          *windesc, 
  LinearClassify                 const &classifier, 
  IndexType                      const &toadd, 
  lear::ResultHolder                   &holder, 
  bool                                  dotestlocs, 
  std::string                    const &f, 
  int                                  &imagewindows, 
  std::ofstream                        &testlocsstr,
  RealType                              scalePrecision,
  RealType                              scaleratio,
  RealType                              endscale,
  RealType                              startscale,
#ifdef HAVE_CUDA
  bool                                  useCuda,
#endif
  int                                   verbose)  
{
  typedef lear::PerspectiveScalePyramid PyramidType;
  typedef lear::ImageSlider<2>          SliderType;

  assert(blitz::all(initialExt > 0));

  PyramidType const pyramid(
    initialExt, winsize, borderPixels, winstride, 
    pedestrianHeight, rodrigues, translation, focalLength, 
    scalePrecision, scaleratio, endscale, startscale);

  if (verbose > 6) {
    cout << 
      "Perspective-corrected scale-space pyramid" << 
      ", Distinct scale levels: " << setw(2) << pyramid.numDistinctScales() <<
      ", Smallest scale = " << 
      setw(6) << setprecision(2) << pyramid.smallestScale() <<
      ", Largest scale = " <<
      setw(6) << setprecision(2) << pyramid.largestScale() <<
      endl;
  }

  if (toadd[0] != 0 || toadd[1] != 0) {
    ThrowException("toadd not supported right now");
  }
  if (hasTopLeft && hasFullSize) {
    ThrowException("topleft not supported right now");
  }

#ifdef HAVE_CUDA
  WinDescriptorCuda *windescCuda = dynamic_cast<WinDescriptorCuda*>(windesc);
  if (windescCuda == NULL && useCuda) {
    ThrowException("windesc object is the wrong type");
  }
#endif

  Scale const initialScale(image.extent()[0] / RealType(initialExt[0]),
                           image.extent()[1] / RealType(initialExt[1]));

  windesc->setimage(image);
  PyramidType::ScaleToTopRows const &scaleData = pyramid.scaleToTopRows();
  typedef PyramidType::ScaleToTopRows::const_iterator PIter;
  typedef PyramidType::TopRows TopRows;
  for (PIter piter=scaleData.begin(); piter!=scaleData.end(); ++piter) {
    // Variable prefixes:
    //   u -- unscaled coordinates (image coords)
    //   c -- unscaled & cropped coordinates (cropped coords)
    //   s -- cropped & scaled coordinates (pyimg coords)
    //   L -- lower bound
    //   U -- upper bound
    // (uY - uLCropped)/scale = sY
    Scale   const  scale   = piter->first * initialScale;
    TopRows const &sTopRow = piter->second;

    // Don't pad the image for windows that overlap the boundaries
    // TODO: allow some padding to improve filter outputs
    IndexType const uLCropped(0, 
      max(0, int(floor(*sTopRow.begin() * scale[1]))));
    IndexType const uUCropped(image.extent()[0],
                              min(initialExt[1], 
                                  int(ceil((*sTopRow.rbegin() + 
                                            winsize[1]) * scale[1]))));
    IndexType const eCropped(uUCropped - uLCropped);
    IndexType const sExt(eCropped/scale);

    if (verbose > 7) {
      cout <<"  Examining scale " << setprecision(6) << scale  
           << ", (" << right << setw(4) << uLCropped[0] << "x"
           << right << setw(4) << uLCropped[1] << ") to "
           << "(" << right << setw(4) << uUCropped[0]
           << "x" << right << setw(4) << uUCropped[1] << "), "
           << "final size is (" << right << setw(4) << sExt[0] << "x" 
           << right << setw(4) << sExt[1] << ")" <<  endl;
      cout.flush();
    }
    windesc->preprocess(uLCropped, eCropped, sExt);

#ifdef HAVE_CUDA
    if (useCuda) {
      runPerspectiveRowsGivenScale_CUDA(uLCropped, scale, size, 
                                        windescCuda,  classifier, holder, 
                                        dotestlocs, f, imagewindows, 
                                        testlocsstr);
    } else {
#endif
      runPerspectiveRowsGivenScale(sTopRow, winsize, sExt, uLCropped, scale, 
                                   winstride, size, windesc, classifier, 
                                   holder, dotestlocs, f, imagewindows, 
                                   testlocsstr);
#ifdef HAVE_CUDA
    }
#endif
  }
}

#ifdef HAVE_TRACKER
lear::Tracker_PPResult *
WinDetectClassify::createTrackerPPResult(IndexType const &initialExt, 
                                         double fpsFromVid) const {
  auto_ptr<JTracker> tracker(new JTracker());
  stringstream ss;
  ss << "--maxTrk " << maxtrk << " "
     << "--maxCC " << maxcc << " "
     << "--ivxy " << ivxy << " "
     << "--ivdxy " << ivdxy << " "
     << "--iva " << iva << " "
     << "--maxvxy " << maxvxy << " "
     << "--maxvdxy " << maxvdxy << " "
     << "--maxva " << maxva << " "
     << "--minvxy " << minvxy << " "
     << "--minvdxy " << minvdxy << " "
     << "--minva " << minva << " "
     << "--addThreshold " << addThreshold << " "
     << "--delThreshold " << delThreshold << " "
     << "--searchTime " << searchTime << " "
     << "--pLearningRate " << pLearningRate << " "
     << "--vLearningRate " << vLearningRate << " "
     << "--aLearningRate " << aLearningRate << " "
     << "--ccAreaThreshold " << 0 << " " // startscale already does this
    ;  
  if (tracker->setup(ss.str().c_str()) != Trk_OK) {
    ThrowException("Could not configure tracker: " << ss.str());
  }    
    
  auto_ptr<ITrkOutputStream> trkoutput;
  if (strcasecmp(trkformat.c_str(), "josh") == 0) {
    trkoutput.reset(new TrkJoshOutputStream());
  } else if (strcasecmp(trkformat.c_str(), "stauffer") == 0) {
    trkoutput.reset(new TrkStaufferOutputStream());
  } else {
    ThrowException("Unknown trk format");
  }
  string setup = strprintf("--filename %s", trkfile.c_str());
  if (trkoutput->setup(setup.c_str()) != Trk_OK) {
    ThrowException("Could not configure trkoutput: " << setup);
  }

  double const myfps = ((fps==-1) 
                        ? ((fpsFromVid==-1) ? 
                           (30/1.001) : // default: assume NTSC
                           fpsFromVid) 
                        : fps);
    
  return new Tracker_PPResult(tracker.release(), trkoutput.release(), myfps,
                              initialExt);
}
#endif

void WinDetectClassify::runImageSlider(LinearClassify     const &classifier, 
                                       InputImageIterator       &infiles,
                                       std::string        const &maskfile,
                                       std::string        const &outfile,  
                                       bool               const  writeListFilename,
                                       std::string        const &outimage, 
                                       std::string        const &outhist, 
                                       std::string        const &falsetxt, 
                                       std::string        const &testlocs
  ) const
{ 
  using namespace lear;
  using std::setw; using std::setprecision;
  if (!descholder.initialized) {
    ThrowException("Init is supposed to be called before we can use runImageSlider");
  }

  WinDescBase* windesc = descholder.windesc;
  if (verbose > 1) 
  { std::cout << *this << std::endl; }

  // create classifier
  if ((size_t)classifier.length() != (size_t)windesc->length()) {
    std::cout << "Classifier length " << 
      classifier.length() << std::endl;
    ThrowException("Dimension mismatch between window feature vector "
                   "and SVM learned feature vectors");
  }

  IndexType const topleft(topleft_x, topleft_y);
  bool      const hasTopLeft = (blitz::sum(topleft>=0) == 2);
  if (hasTopLeft && verbose > 2) {
    cout << "Top-Left corner specified ( " 
         << topleft[0] << ", " << topleft[1] << " ) " << endl;
  }

  IndexType const fullsize(fullsize_x, fullsize_y);
  IndexType const fullstride(fullstride_x, fullstride_y);

  bool const hasFullSize = (blitz::sum(fullsize>=0) == 2);
  if (hasFullSize && verbose > 2) {
    cout<< "Full-Window size specified ( " 
      << fullsize[0] << ", " << fullsize[1] << " ) " << endl;
  }

  IndexType const size(size_x, size_y);
  IndexType const stride(winstride_x, winstride_y);
  IndexType const winsize = (hasFullSize && hasTopLeft) ? fullsize: size;

  bool const hasFullStride = (blitz::sum(fullstride>=0) == 2);
  if (hasFullStride && verbose > 2) {
    cout<< "Full-Window stride specified ( " 
      << fullstride[0] << ", " << fullstride[1] << " ) " << endl;
  }
  IndexType const winstride =
    (hasFullSize && hasTopLeft && hasFullStride)? fullstride: stride;

  if (verbose > 2) {
    switch (sliderType) 
    {
      case NO_SLIDER:
        cout << "No pyramid specified. Compare center/top-left window only." << endl;
        break;
      case PYRAMID_SLIDER:
        cout << "Using full scale-space pyramid to search for detections." << endl;
        break;
      case PERSPECTIVE_SLIDER:
        cout << "Using a perspective-adjusted scale-space pyramid to search for detections." << endl;
        break;
      default:
        ThrowException("Unknown slider type");
    }
  }

  bool const doFalseOut  = !falsetxt.empty() && label != DefaultLabel;
  bool const infileIsDir = true; // inlist.size() > 1; // TODO: get rid of verifyImageOut stuff : it's all heuristic anyway
  bool const doImageOut  = lear::MarkImage::verifyImageOut(outimage,infileIsDir);
  bool const doBestOut   = lear::DrawBestThreshold::verifyImageOut(outimage, infileIsDir);
  bool const doMasking   = (maskfile.size() != 0);

  std::ofstream testlocsstr;
  bool dotestlocs = testlocs.size();
  if (dotestlocs) {
    testlocsstr.open(testlocs.c_str());
    if (!testlocsstr) {
      std::cerr << "Could not open file to write tested/dumped locations " << testlocsstr << ". Ignoring ..." << std::endl;
      dotestlocs = false;
    }
  }


  ResultHolder holder;

  if (doBestOut && !aligninimage && showbestscores) {
    holder.push_back(new ProcessResult());
    holder.push2last(new DrawBestThreshold(outimage, infileIsDir, "raw"));
  }
  if (doMasking) {
    IndexType const initialExt(initialsize_x, initialsize_y);
    holder.push_back(new Mask_ProcessResult(maskfile, initialExt));
    if (doBestOut && !aligninimage && showbestscores) {
      holder.push2last(new DrawBestThreshold(outimage, infileIsDir, "masked"));
    }
  }
  if (outhist.empty() || (doImageOut && aligninimage)) {
    holder.push_back(new Threshold_ProcessResult(lightthreshold, label!='P'));
  } else { // write histogram now for efficiency reasons
    holder.push_back(
      new Hist_ProcessResult(outhist,lightthreshold, label!='P', -20, 40));
  }
  if (doFalseOut) {
    try {
      Hard_PPResult* p = new Hard_PPResult(falsetxt);
      holder.push2last(p);
      if (verbose > 2) {
        cout << "Detected false +/- will be written to "<< 
          falsetxt << endl;
      }
    } catch (lear::Exception& ) {
      cerr << "Unable to open file " << falsetxt <<
        " for writing false +/-. Disabling false out." << endl;
    }
  }
  typedef blitz::TinyVector<RealType,2>                SigmaType;
  SigmaType  const score2prob(score2prob_a, score2prob_b);

  typedef blitz::TinyVector<RealType,3>                NonmaxType;
  NonmaxType const nonmaxSigma(nonmaxsigma_x,nonmaxsigma_y, nonmaxsigma_scale); 

  int list_marker=1;
  if (!no_nonmax) {
    list_marker=2;
    holder.push_back(new MeanShift_ProcessResult(size, lightthreshold, threshold, 
                                                 score2prob, nonmaxSigma, softmax,
                                                 verbose));
  } 
  if (doImageOut && !aligninimage) {
    holder.push2last(new MarkImage(outimage, infileIsDir, !showscore));
  }
  if (doBestOut && !aligninimage && showbestscores) {
    holder.push2last(new DrawBestThreshold(outimage, infileIsDir, "final"));
  }
  if (outfile.size()) {
    holder.push2last(new List_PPResult(outfile, list_marker, 
                                       writeListFilename, verbose));
  }

  IndexType const alignmargin(alignmargin_x, alignmargin_y);
  if (doImageOut && aligninimage ) {
    holder.push2last(new AlignInImage(
      outimage, infileIsDir,alignmargin,size,outhist,-20,40));
  }

#ifdef HAVE_TRACKER
  if (trkfile.size() != 0) {
    IndexType const initialExt(initialsize_x, initialsize_y);
    holder.push2last(createTrackerPPResult(initialExt, infiles.fps()));
  }
#endif

  if (verbose > 3) {
    holder.print(cout);
  }

  unsigned long totalwindows=0; 

  IndexType const tinitialsize(initialsize_x, initialsize_y);
  IndexType const tmargin(margin_x, margin_y);
  IndexType const tavsize(avsize_x, avsize_y);
  //margin = maximum(margin,0);
  //avsize = maximum(avsize,0);
  bool const presize(blitz::any(tinitialsize != -1));
  bool const addmargin(blitz::any(tmargin*tavsize));
  if (addmargin) {
    if (doMasking) {
      ThrowException("doMasking support has not been added yet when addmargin "
                     "is also being used.  Figure out how to fix "
                     "Mark_ProcessResult so it handles the margins too (or "
                     "delete this check and add documentation if the current "
                     "behavior is actually correct).");
    }
    if (trkfile.size()) {
      ThrowException("Kalman tracking support has not been added yet when "
                     "addmargin is also being used for the same reason that "
                     "addmargin and masking is not supported yet.");
    }
  }
  int numImages = 0;
  while (infiles.next()) {
    WinDescBase::ImageType const &origimage = infiles.currImage();
    IndexType toadd = 0;

    IndexType initialExt = origimage.extent();
    if (presize) {
      for (int d=0; d<2; d++) {
        if (tinitialsize[d] != -1) initialExt[d] = tinitialsize[d];
      }
    }

    WinDescBase::ImageType image;
    if (addmargin) {
      WinDescBase::ImageType presized;
      if (presize) {
        // margin addition has not been pushed into WinDescriptor, so
        // we're forced to do the initial resizing right now.
        image.resize(initialExt);
        presized = lear::rescale(origimage, initialExt);
      } else {
        presized.reference(origimage);
      }

      IndexType newext = presized.extent();
      IndexType toaddX = 0, toaddY = 0;
      if (tavsize[0]) toaddX = newext*tmargin/tavsize[0];
      if (tavsize[1]) toaddY = newext*tmargin/tavsize[1];

      toadd = blitz::max(toaddX,toaddY);
      newext += 2*toadd;

      image.reference(extendBorder(
        presized, newext/2-1, newext/2, presized.extent()/2));

      initialExt = image.extent();
    } else {
      image.reference(origimage);
    }

    if (verbose > 5) cout << "Processing file " << infiles.currName() << endl;

    int imagewindows = 0; 
    holder.clear();

    switch (sliderType) 
    {
      case NO_SLIDER:
        if (!runDegenerateImageSlider(image, initialExt, hasTopLeft, hasFullSize, topleft, 
              fullsize, size, winsize, windesc, classifier, toadd, holder, 
              dotestlocs, infiles.currName(), imagewindows, testlocsstr)) {
          continue;
        }
        break;
      case PYRAMID_SLIDER:
        runPyramidImageSlider(image, initialExt, hasTopLeft, hasFullSize, topleft, 
          size, winsize, winstride, windesc, classifier, toadd, holder, 
          dotestlocs, infiles.currName(), imagewindows, testlocsstr, 
          scaleratio, endscale, startscale, verbose);  
        break;
      case PERSPECTIVE_SLIDER:
        runPerspectiveSlider(image, initialExt, hasTopLeft, hasFullSize, /*topleft, */
                             size, winsize, borderPixels, winstride, 
                             pedestrianHeight, rodrigues, translation, 
                             focalLength, windesc, classifier, toadd, holder, 
                             dotestlocs, infiles.currName(), imagewindows, 
                             testlocsstr, scalePrecision, scaleratio, endscale,
                             startscale, 
#ifdef HAVE_CUDA
                             useCuda,
#endif                             
                             verbose);
        break;
      default:
        ThrowException("Unknown slider type");
    }
    totalwindows+=imagewindows;

    if (blitz::any(initialExt != origimage.extent())) {
      WinDescBase::ImageType resized;
      windesc->getimage(resized, initialExt);
      holder.write(resized, infiles.currName()); //just write to files
    } else {
      holder.write(origimage, infiles.currName()); //just write to files
    }
    if (verbose > 3) {
      cout << "Processed " << setw(5) << imagewindows 
        << " windows in file " << infiles.currName() <<  endl;
    }

    numImages++;
    if (maximages >= 0 && numImages >= maximages) break;
  }
  if (verbose > 0)  {
    cout << "Tested " << totalwindows << " windows from " 
         << numImages << " images." << endl;
  }
}

#endif

void WinDetect::print(std::ostream& o) const 
{
  o << "WinDetect ::\n"
    "  | Size    " << setw(3) << right << size_x   << "x" << setw(3) << left <<   size_y << right << 
    "          Stride  " << setw(3) << right << winstride_x << "x" << setw(3) << left << winstride_y << right << "   |\n" <<
    "  | TopLeft " << setw(3) << right << topleft_x<< "x" << setw(3) << left << topleft_y<< right << 
    "          FullSize" << setw(3) << right <<fullsize_x<< "x" << setw(3) << left <<fullsize_y<< right << "   |\n" <<
    "  | ScaleRatio " << setw(4) << left << scaleratio << " StartScale " << setw(3) << left << startscale << " EndScale " << setw(3) << left << endscale  << "|\n"
    "  | Label " << setw(4) << left << label  << "  CacheSize  " << setw(4) << left << cachesize  << "  Verbose  " << setw(4) << left << verbose << " |\n"
    "  |--------------------------------------------|\n";
}
#ifdef BUILD_APP
void WinDetectDump::print(std::ostream& o) const
{
  o << "WinDetectDump :: \n"
    "  | Pyramid " << pyramid << "  RandSeed " << randomSeed << "  Samples " << setw(2) << left << samples  << "  DumpDet " << dumpfulldetail << " |\n"
    "  | MemLimit " << setw(6) << left << memorylimit << "          HardScore " << setw(6) << left << hardscore << "    |\n" 
    "  | PosCases " << setw(6) << left << poscases    << "          NegCases  " << setw(6) << left << negcases << "    |\n"
    "  | JitterStep X:" << setw(4) << left << jitterstep_x << "  Y:" << setw(4) << left << jitterstep_y<< "   WinStep:" << setw(3) << jitterwinstep <<  "      |\n"
    "  |----------------------------------------------|\n";
  WinDetect::print(o);
}
#endif
void WinDetectClassify::print(std::ostream& o) const 
{
  o << "WinDetectClassify ::\n"
    "  | SliderType " << sliderType << "   NonMax " << !no_nonmax << "   AlignImg " << aligninimage  << "   ShowSc " << showscore << " |\n"
    "  | Thres   " << setw(4) << left << threshold << "  LtThres   " << setw(4) << left << lightthreshold   << "  SoftMax   " << setw(4) << left << softmax << " |\n"
    "  | Sc2Prob   A:" << setw(4) << left << score2prob_a << "  B:" << setw(4) << left << score2prob_b<< "                     |\n"
    "  | NonMaxSig X:" << setw(4) <<  nonmaxsigma_x<< "  Y:" << setw(4) << nonmaxsigma_y << "  Y: " << setw(4) << nonmaxsigma_scale  << "            |\n"
    "  | AvgSize " << setw(3) << right << avsize_x << "x" << setw(3) << left << avsize_y << right << 
    "          Margin  " << setw(3) << right << margin_x << "x" << setw(3) << left << margin_y << right << "     |\n" <<
    "  | AlMargin" << setw(3) << right << alignmargin_x<< "x" << setw(3) << left << alignmargin_y<< right << 
    "          FStride " << setw(3) << right <<fullstride_x<< "x" << setw(3) << left <<fullstride_y<< right << "     |\n" <<
    "  |----------------------------------------------|\n";
  WinDetect::print(o);
}
