#ifndef PDCUDA_TESTUTILS_H
#define PDCUDA_TESTUTILS_H

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <limits>
#include <set>
#include <map>
#include <string>
#include <limits>

#include <blitz/array.h>

#include <lear/image/imageio.h>

#include <cuda_runtime_api.h>
#include <cutil/cutil_inline.h>
#include <cutil/cutil.h>
#include <cutil/stopwatch.h>
#include <cutil/error_checker.h>

#include "types_gold.h"

#ifdef _WIN32
#  define isnan _isnan
#endif

namespace pdcuda 
{
  /////////////////////////////////////////////////////////////////////////////
  // Forward Declarations

  template <typename BoundFcn>
  void TTimeIt(char const *name, int n, BoundFcn f);
  
  // Macros for timing repeated function calls
  //   n:        number of times the bound function should be called
  //   F:        function object
  //   p1..p<N>: arguments to which the function should be bound
#define TimeIt1(n,F,p1                     ) TTimeIt(#F, n, boost::bind(&F, p1                     )) 
#define TimeIt2(n,F,p1,p2                  ) TTimeIt(#F, n, boost::bind(&F, p1,p2                  )) 
#define TimeIt3(n,F,p1,p2,p3               ) TTimeIt(#F, n, boost::bind(&F, p1,p2,p3               )) 
#define TimeIt4(n,F,p1,p2,p3,p4            ) TTimeIt(#F, n, boost::bind(&F, p1,p2,p3,p4            )) 
#define TimeIt5(n,F,p1,p2,p3,p4,p5         ) TTimeIt(#F, n, boost::bind(&F, p1,p2,p3,p4,p5         )) 
#define TimeIt6(n,F,p1,p2,p3,p4,p5,p6      ) TTimeIt(#F, n, boost::bind(&F, p1,p2,p3,p4,p5,p6      )) 
#define TimeIt7(n,F,p1,p2,p3,p4,p5,p6,p7   ) TTimeIt(#F, n, boost::bind(&F, p1,p2,p3,p4,p5,p6,p7   )) 
#define TimeIt8(n,F,p1,p2,p3,p4,p5,p6,p7,p8) TTimeIt(#F, n, boost::bind(&F, p1,p2,p3,p4,p5,p6,p7,p8)) 


  // Fancier timer collection
  class TimerSet {
  public:
    inline TimerSet();
    inline ~TimerSet();
    inline void                            tic(std::string const &name);
    inline void                            toc();
    inline void                            printTimes();
    inline void                            printAverageTimes();
    inline std::vector<std::string> const &timerNames() const;
    inline StopWatchC                     *timer(std::string const &name);
    inline StopWatchC                     *timer(int i);
    inline size_t                          maxNameLen() const;
  private:
    std::set<unsigned>         timerIds; // for memory cleanup
    std::map<std::string, int> iFromName;
    std::vector<StopWatchC*>   swFromI;
    std::vector<std::string>   nameFromI;
    int                        currI;
    static int const           NO_I = -1;
  };

  // RAII class for auto-calling TimerSet::toc.  This is useful when
  // you want to use multiple TimerSet instances and you don't want to
  // have to worry about explicitly calling toc when switching between
  // timer sets.
  class TicToc {
  public:
    TicToc(TimerSet &ts, std::string const &n) : ts(ts) { ts.tic(n); }
    ~TicToc() { ts.toc(); }
  private:
    TimerSet &ts;
  };


  template <typename T1, typename T2, int N>
  inline float rms(blitz::Array<T1,N> const *a, blitz::Array<T2,N> const *b);

  template <typename T1, typename T2, int N>
  inline float rms(std::vector<blitz::TinyVector<T1,N> > const *a, 
                   std::vector<blitz::TinyVector<T2,N> > const *b);

  template <typename T1, typename T2, int N>
  inline float rmsRing(blitz::Array<T1,N> const *a, 
                       blitz::Array<T2,N> const *b, 
                       double ringSz);

  // Helper macros for testing RMS differences between pairs of blitz
  // arrays.
#define CheckSimilar(a,b, tol) \
  ErrorChecker::condition(((a)->size() == (b)->size()) && \
                          (rms((a),(b)) <= (tol)),\
                          __FILE__, __LINE__)
#define CheckSimilarRing(a,b, tol, ringSz) \
  ErrorChecker::condition(((a)->size() == (b)->size()) && \
                          (rmsRing((a),(b),(ringSz)) <= (tol)),\
                          __FILE__, __LINE__)

  inline void imwriteScaledTransposed(std::string const &fname, 
                                      blitz::Array<float,2> const *scores,
                                      float minVal=std::numeric_limits<float>::quiet_NaN(),
                                      float maxVal=std::numeric_limits<float>::quiet_NaN());







  /////////////////////////////////////////////////////////////////////////////
  // Implementations

  // Helper template for the TimeIt<N> macros
  template <typename BoundFcn>
  void TTimeIt(char const *name, int n, BoundFcn f) {
    printf("%s: ", name);      
    fflush(stdout);          
    unsigned    timerId = StopWatch::create();     
    StopWatchC &timer   = StopWatch::get(timerId); 
    timer.start();      
    for (int i=0; i<n; ++i) {
      f();
    }            
    timer.stop();
    printf("mean time = %f (ms)\n", timer.getTime()/n);   
    StopWatch::destroy(timerId);
  }


  inline TimerSet::TimerSet() : currI(NO_I) {};

  inline TimerSet::~TimerSet() {
    std::set<unsigned>::iterator i;
    for (i=timerIds.begin(); i!=timerIds.end(); i++) {
      StopWatch::destroy(*i);
    }
  }
  
  void inline TimerSet::tic(std::string const &name) {
    toc();
    
    std::map<std::string, int>::iterator i = iFromName.find(name);
    if (i == iFromName.end()) {
      unsigned timerId = StopWatch::create();
      timerIds.insert(timerId);
      currI = iFromName[name] = (int)swFromI.size();
      nameFromI.push_back(name);
      StopWatchC *sw = &StopWatch::get(timerId);
      swFromI.push_back(sw);
    } else {
      currI = i->second;
    }

    swFromI[currI]->start();
  }

  inline void TimerSet::toc() {
    if (currI != NO_I) {
      cutilSafeCall(cudaThreadSynchronize());
      swFromI[currI]->stop();
      currI = NO_I;
    }
  }

  inline void TimerSet::printTimes() {
    toc();

    size_t w = maxNameLen();
    for (size_t i=0; i<swFromI.size(); i++) {
      std::cout << "    " << std::right << std::setw(w) 
                << nameFromI[i] << ": " << std::flush;
      printf("%8.2f (ms)\n", swFromI[i]->getTime());
      fflush(stdout);
    }
  }

  inline void TimerSet::printAverageTimes() {
    toc();

    size_t w = maxNameLen();
    for (size_t i=0; i<swFromI.size(); i++) {
      std::cout << "    " << std::right << std::setw(w) 
                << nameFromI[i] << ": " << std::flush;
      printf("%8.2f (ms)\n", swFromI[i]->getAverageTime());
      fflush(stdout);
    }
  }

  inline std::vector<std::string> const &TimerSet::timerNames() const {
    return nameFromI;
  }

  inline StopWatchC *TimerSet::timer(std::string const &name) {
    if (iFromName.find(name) == iFromName.end()) return NULL;
    return swFromI[iFromName[name]];
  }

  inline StopWatchC *TimerSet::timer(int i) {
    if (i >= 0 && (size_t)i >= iFromName.size()) return NULL;
    return swFromI[i];
  }

  inline size_t TimerSet::maxNameLen() const {
    size_t w = 0;
    for (size_t i=0; i<nameFromI.size(); i++) {
      if (nameFromI[i].size() > w) w = nameFromI[i].size();
    }
    return w;
  }


  // Computes the RMS difference between two blitz arrays for linear
  // spaces.
  template <typename T1, typename T2, int N>
  inline float 
  rms(blitz::Array<T1,N> const *a, blitz::Array<T2,N> const *b) {
    double ssa = 0;
    double ssb = 0;
    double ssd = 0;
    typename blitz::Array<T1,N>::const_iterator ia = a->begin();
    typename blitz::Array<T2,N>::const_iterator ib = b->begin();
    for (int i=0; i<a->size(); ++i, ++ia, ++ib) {
      double const d = (*ia - *ib);
      ssd += d*d;
      ssa += *ia * *ia;
      ssb += *ib * *ib;
     }
    double const ret = sqrt(ssd / a->size());
    std::cout << "rms = " << ret 
              << " (mean |a|="<<sqrt(ssa / a->size())
              <<", mean |b|="<<sqrt(ssb / b->size())<<")"
              <<std::endl;
    return ret;
  }

  // ...for vectors of TinyVectors
  template <typename T1, typename T2, int N>
  inline float 
  rms(std::vector<blitz::TinyVector<T1,N> > const *a, 
      std::vector<blitz::TinyVector<T2,N> > const *b)
  {
    double ssa = 0;
    double ssb = 0;
    double ssd = 0;
    if (a->size() != b->size()) {
      return std::numeric_limits<double>::quiet_NaN();
    }

    for (size_t i=0; i<a->size(); i++) {
      for (int j=0; j<N; j++) {
        double const d = (a[i][j] - b[i][j]);
        ssd += d*d;
        ssa += a[i][j] * a[i][j];
        ssb += a[i][j] * a[i][j];
      }
    }
    double const ret = sqrt(ssd / a->size());
    std::cout << "rms = " << ret 
              << " (mean |a|="<<sqrt(ssa / a->size())
              <<", mean |b|="<<sqrt(ssb / b->size())<<")"
              <<std::endl;
    return ret;
  }

  // Computes the RMS difference between two blitz arrays for input
  // spaces with a circular topology (like angles).
  template <typename T1, typename T2, int N>
  inline float 
  rmsRing(blitz::Array<T1,N> const *a, blitz::Array<T2,N> const *b, 
          double ringSz)
  {
    double ssd = 0;
    typename blitz::Array<T1,N>::const_iterator ia = a->begin();
    typename blitz::Array<T2,N>::const_iterator ib = b->begin();
    for (int i=0; i<a->size(); ++i, ++ia, ++ib) {
      double d = fmod((*ia - *ib) + ringSz, ringSz);
      d = std::min(d, ringSz - d);
      ssd += d*d;
    }
    double const ret = sqrt(ssd / a->size());
    std::cout << "rms ring = " << ret << std::endl;
    return ret;
  }


  inline void imwriteScaledTransposed(std::string const &fname, 
                                      blitz::Array<float,2> const *scores,
                                      float minVal, float maxVal)
  {
    using namespace lear;

    if (isnan(minVal)) {
      minVal = blitz::min(*scores); 
    }
    if (isnan(maxVal)) {
      maxVal = blitz::max(*scores); 
    }

    int const wi = scores->extent()[0];
    int const hi = scores->extent()[1];

    blitz::Array<float,2> forWriting(IndexType(hi,wi));
    for (int xi=0; xi<wi; xi++) {
      for (int yi=0; yi<hi; yi++) {
        float val = (*scores)(xi,yi);
        if (val >= maxVal) {
          val = 255;
        } else if (val <= minVal) {
          val = 0;
        } else {
          val = (val + minVal) * 255 / (maxVal - minVal);
        }
        forWriting(yi,xi) = val;
      }
    }

    lear::ImageIO::write(fname, forWriting);
  }


};

#endif
