#ifndef TRANSPOSEDARRAYS_H
#define TRANSPOSEDARRAYS_H

#include <blitz/array.h>
#include <cutil/cutil_inline.h>
#include <cutil/error_checker.h>

namespace pdcuda
{

  // Variable naming conventions:
  //
  //   We use a hungarian notation prefixes on variables to indicate
  //   the storage location:
  //    
  //   d: "d"evice pointer
  //   h: "h"ost pointer or array
  //   c: "c"onstant memory
  //   t: "t"ransposed array (a host array and a transposed device 
  //        array bound together).
  //   s: "s"hared memory array (device only)
  //
  //   Sometimes we have an additional prefix to indicate who produced
  //   an array.  These are typically found in the app/test*.cpp
  //   files.  If this extra prefix is used, it's always followed by
  //   the storage location prefix.
  //
  //   c: "C"PU
  //   g: "G"PU
  //   t: "t"emporary data
  //
  //   Examples:
  //     dFoo: a device pointer to the "foo" array 
  //     ghFoo: a host (CPU) pointer to the "foo" array, where the array 
  //       contents were generated on the GPU


  // Wrapper for synchronizing column-major 2D blitz arrays with row-major
  // device (GPU) arrays.  
  //
  // This code is host-only.
  template<typename T> 
  class TransposedDeviceArray2D
  {
  public:
    typedef blitz::Array<T,2> HArray;
    typedef blitz::TinyVector<int,2> Extent;

    TransposedDeviceArray2D(HArray *hArray, 
                            bool copyNow=false, bool isMutable=true);
    TransposedDeviceArray2D(HArray *hArray, int d1, int d2, 
                            bool copyNow=false, bool isMutable=true);

    TransposedDeviceArray2D(HArray const *hArray, bool copyNow=false);
    TransposedDeviceArray2D(HArray const *hArray, int d1, int d2, 
                            bool copyNow=false);

    TransposedDeviceArray2D(int d1, int d2);

    virtual ~TransposedDeviceArray2D();

    void          toDevice();
    void          fromDevice();
    T            *devicePtr()  const { return dData; }
    HArray       *hostArray()        { return isMutable ? hArray : NULL; }
    HArray const *hostArray()  const { return hArray; }

    Extent extent() const { return ext; }

  private:
    void release();
    void allocMem();

    bool     isMutable;
    Extent   ext;
    HArray  *hArray;
    HArray   transposed;
    T       *dData;
    bool     ownsHArray;
  };

  //-------------------------------------------------------------------------
  // Implementation

  // Wrap the given host array.  This class does not take ownership of
  // the array.  
  template<typename T> 
  TransposedDeviceArray2D<T>::
  TransposedDeviceArray2D(HArray *hArray, bool copyNow, bool isMutable) :
    isMutable(isMutable), ext(hArray->extent()), hArray(hArray), dData(NULL),
    ownsHArray(false)
  {
    allocMem();
    if (copyNow) toDevice();
  }

  // Wrap the given host array and resize it.  This class does not
  // take ownership of the array.  
  template<typename T> 
  TransposedDeviceArray2D<T>::
  TransposedDeviceArray2D(HArray *hArray, int d1, int d2,
                          bool copyNow, bool isMutable) :
    isMutable(isMutable), hArray(hArray), dData(NULL),
    ownsHArray(false)
  {
    hArray->resize(d1,d2);
    ext = hArray->extent();
    allocMem();
    if (copyNow) toDevice();
  }

  template<typename T> 
  TransposedDeviceArray2D<T>::
  TransposedDeviceArray2D(HArray const *hArray, bool copyNow) :
    isMutable(false), ext(hArray->extent()), 
    hArray(const_cast<HArray*>(hArray)), dData(NULL),
    ownsHArray(false)
  {
    allocMem();
    if (copyNow) toDevice();
  }

  template<typename T> 
  TransposedDeviceArray2D<T>::
  TransposedDeviceArray2D(HArray const *hArray, int d1, int d2,
                          bool copyNow) :
    isMutable(false), hArray(static_cast<HArray*>(hArray)), dData(NULL),
    ownsHArray(false)
  {
    hArray->resize(d1,d2);
    ext = hArray->extent();
    allocMem();
    if (copyNow) toDevice();
  }

  // Creates the host array as well (and owns it)
  template<typename T> 
  TransposedDeviceArray2D<T>::
  TransposedDeviceArray2D(int d1, int d2) :
    isMutable(true), hArray(new HArray(d1,d2)), dData(NULL),
    ownsHArray(true)
  {
    ext = hArray->extent();
    allocMem();
  }

  // Automatically free the device array and temporary scratch space
  template<typename T> 
  TransposedDeviceArray2D<T>::~TransposedDeviceArray2D() 
  {
    release();
  }

  // Free up device memory and temporary scratch space
  template<typename T> 
  void TransposedDeviceArray2D<T>::release()
  {
    if (ownsHArray) delete hArray;
    hArray = NULL;
    memset(&ext, 0, sizeof(ext));
    if (dData) cutilSafeCall(cudaFree(dData));
    dData  = NULL;
  }

  // Transpose, allocate, and transfer a 2D host array to the GPU device
  template<typename T> 
  void TransposedDeviceArray2D<T>::toDevice()
  {
    ErrorChecker::condition(hArray->extent()[0]==ext[0], __FILE__,__LINE__);
    ErrorChecker::condition(hArray->extent()[1]==ext[1], __FILE__,__LINE__);

    memset(transposed.data(), 0, sizeof(T)*transposed.size());

    int const d1 = ext[0];
    int const d2 = ext[1];
    for (int i1=0; i1<d1; i1++) {
      for (int i2=0; i2<d2; i2++) {
        transposed(i2,i1) = (*hArray)(i1,i2);
      }
    }

    cutilSafeCall(cudaMemcpy(dData, transposed.data(), sizeof(T)*d1*d2,
                             cudaMemcpyHostToDevice));
  }
  
  // Allocate an array on the GPU device and temporary transposition
  // scratch space.
  template<typename T> 
  void TransposedDeviceArray2D<T>::allocMem()
  {
    cutilSafeCall(cudaMalloc((void**)(void*)&dData, 
                             sizeof(T)*hArray->size()));      
    ext = hArray->extent();
    transposed.resize(ext[1], ext[0]);
  }

  // Transfer data back from the device to the wrapped host array and
  // transpose it.
  template<typename T> 
  void TransposedDeviceArray2D<T>::fromDevice()
  {
    ErrorChecker::condition(hArray->extent()[0]==ext[0], __FILE__,__LINE__);
    ErrorChecker::condition(hArray->extent()[1]==ext[1], __FILE__,__LINE__);
    ErrorChecker::condition(isMutable, __FILE__, __LINE__);

    int const d1 = hArray->extent(0);
    int const d2 = hArray->extent(1);
    blitz::Array<T,2> transposed(d2,d1);

    cutilSafeCall(cudaMemcpy(transposed.data(), dData, sizeof(T)*d1*d2,
                             cudaMemcpyDeviceToHost));

    for (int i1=0; i1<d1; i1++) {
      for (int i2=0; i2<d2; i2++) {
        (*hArray)(i1,i2) = transposed(i2,i1);
      }
    }
  }
  
};

#endif
