/*
  Tests the consistency of our GPU implementation of color gradient
  calculation is consistent with Dalal's CPU implementation that
  computes the gradient of sqrt(RGB) using a non-smoothing gradient
  filter.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <string>
#include <memory>
#include <iostream>

#include <blitz/array.h>
#include <boost/bind.hpp>

#include <lear/cvision/iprocessor.h>
#include <lear/image/imageio.h>

#include <pdcuda/gradients.h>
#include <pdcuda/transposedarrays.h>
#include <pdcuda/testutils.h>

#include <pdcuda/gradients_gold.h>

using namespace std;
using namespace pdcuda;

//////////////////////////////////////////////////////////////////////////////
// GPU Implementation

void computeColorGradientsGPU(Magnitude *m, OrientationF *o, RGBImage const &i) 
{
  int const w = i.extent()[0];
  int const h = i.extent()[1];

  typedef blitz::TinyVector<float,3> Pixel;
  pdcuda::TransposedDeviceArray2D<float> mag(m, w,h);
  pdcuda::TransposedDeviceArray2D<float> ori(o, w,h);
  pdcuda::TransposedDeviceArray2D<Pixel> rgb(&i, true);
  
  pdcuda::computeSpatialGradients(mag.devicePtr(), ori.devicePtr(), 
                                  (float const *)rgb.devicePtr(), w, h);  
  mag.fromDevice();
  ori.fromDevice();
  cutilSafeCall(cudaThreadSynchronize());
}

//////////////////////////////////////////////////////////////////////////////
// Main & command-line argument processing

// USAGE:
//   testcolorgradients
//   testcolorgradients infile
//   testcolorgradients infile rmsMag 
//   testcolorgradients infile rmsMag rmsOri 
//   testcolorgradients infile rmsMag rmsOri nIter
void parseArgs(int argc, char const **argv,
               string &infile, float &rmsMag, float &rmsOri, int &nIter)
{
  if (argc > 1) infile = argv[1];
  if (argc > 2) rmsMag = atof(argv[2]);
  if (argc > 3) rmsOri = atof(argv[3]);
  if (argc > 4) nIter  = atoi(argv[4]);
}

int main(int argc, char const **argv) {
  string infile("input.png");
  float  rmsMag = 1.f;
  float  rmsOri = 2.f;
  int    nIter  = 20;
  parseArgs(argc, argv, infile, rmsMag, rmsOri, nIter);

  RGBImage in;
  lear::ImageIO::read(infile, in);

  Magnitude    cMag;
  Magnitude    gMag;
  Orientation  cOri;
  OrientationF gOri;
  computeColorGradientsGPU(&gMag, &gOri, in); // warmup the device
  TimeIt3(nIter, computeColorGradientsGPU, &gMag, &gOri, in);
  TimeIt3(nIter, computeColorGradientsCPU, &cMag, &cOri, in);

  lear::ImageIO::write("cOri.png", Orientation(cOri*255/180.));
  lear::ImageIO::write("gOri.png", Orientation(gOri*255/180.));
  lear::ImageIO::write("cMag.png", Magnitude(cMag*16));
  lear::ImageIO::write("gMag.png", Magnitude(gMag*16));

  printf("Checking for consistency...\n");
  CheckSimilar(&cMag, &gMag, rmsMag);
  CheckSimilarRing(&cOri, &gOri, rmsOri, 180);

  return 0;
}

