#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <algorithm>
#include <memory>
#define _USE_MATH_DEFINES
#include <math.h>
#include <sys/stat.h>

#include <boost/bind.hpp>
#include <boost/shared_array.hpp>

#include <lear/image/imageio.h>
#include <lear/image/rescale.h>
#include <lear/interface/windetect.h>
#include <lear/classifier/ms_processresult.h>
#include <lear/cvision/perspective.h>

#define double float // dirty trick to get the data as floats
# ifdef _WIN32
#  pragma warning (push)
#  pragma warning (disable:4305)
# endif
# include "../lib/persondetectorwt.tcc"
# ifdef _WIN32
#  pragma warning (pop)
# endif
#undef double

#ifdef _WIN32
#include <direct.h>
inline static void mkDir(char const *d) { _mkdir(d); }
#else
inline static void mkDir(char const *d) { mkdir(d, 0777); }
#endif

#include <pdcuda/arrays.h>
#include <pdcuda/transposedarrays.h>
#include <pdcuda/testutils.h>

#include <pdcuda/imtransfer.h>
#include <pdcuda/imresize.h>
#include <pdcuda/gradients.h>
#include <pdcuda/rhogdense.h>

#include <pdcuda/types_gold.h>
#include <pdcuda/gradients_gold.h>
#include <pdcuda/rhogdense_gold.h>

using namespace std;
using namespace pdcuda;

typedef blitz::TinyVector<RealType,2>                SigmaType;
typedef blitz::TinyVector<RealType,3>                NonmaxType;

float      const nonmaxScale = 1.3f;
SigmaType  const scoreToProb(1,1);
float      const minOptPrecision  = 0.9;
int        const maxOptTestMS     = 10000;
int        const whichTransferFunc(0); // 0=HardMax, 1=Sigmoid, 2=SoftMax
float      const minOverlapFrac =  0.5f; // PASCAL VOC uses 0.5f

//////////////////////////////////////////////////////////////////////////////
// Global detection

class ScoresGivenScale {
public:
  ScoresGivenScale(int w, int h, float scale) 
    : w(w), h(h), scale(scale), scores(new float[w*h]) {}

  ScoresGivenScale(ScoresGivenScale const &rhs) 
    : w(rhs.w), h(rhs.h), scale(rhs.scale), scores(rhs.scores) {}
   
  int   w, h;
  float scale; // orig/new
  boost::shared_array<float> scores;
};

void detectionsFromClassifierOutput(vector<DetectedRegion>         &detections,
                                    GrayImage                const *mask,
                                    vector<ScoresGivenScale> const &scores,
                                    float      svmThreshold, 
                                    SigmaType  scoreToProb,
                                    NonmaxType nonmaxSigma, 
                                    int        whichTransferFunc)
{
  using namespace lear;
  detections.clear();
  float const finalThreshold = svmThreshold; // for now
  MeanShift_ProcessResult holder(IndexType(WinSizeX, WinSizeY), 
                                 svmThreshold, 
                                 finalThreshold, 
                                 scoreToProb, 
                                 nonmaxSigma, 
                                 whichTransferFunc);

  for (size_t i=0; i<scores.size(); i++) {
    ScoresGivenScale const &sgs = scores[i];
    float      *s     = sgs.scores.get();
    float const scale = sgs.scale;
    for (int y=0; y<sgs.h; y++) {
      for (int x=0; x<sgs.w; x++) {
        float const score = *s++;
        int   const xOrig = (int)ceil(x*GoldFineStrideX*scale); // orig coords
        int   const yOrig = (int)ceil(y*GoldFineStrideY*scale);

        // TODO: we should dilate the mask 
        if ((!mask || (*mask)(xOrig, yOrig)) && (score >= svmThreshold)) {
          DetectInfo r(score, scale, 
                       IndexType(xOrig, yOrig),
                       IndexType((int)floor(WinSizeX*scale), 
                                 (int)floor(WinSizeY*scale)),
                       IndexType(x,y));
          //r.lbound -=toadd;
          holder(r);
        }
      }
    }
  }

  holder.doit();

  MeanShift_ProcessResult::const_iterator s = holder.begin();
  MeanShift_ProcessResult::const_iterator e = holder.end();
  for (; s!= e; ++s) {
    detections.push_back(DetectedRegion(s->score, s->scale, 
      s->lbound[0], s->lbound[1], s->extent[0], s->extent[1]));
  }
}

//////////////////////////////////////////////////////////////////////////////
// Detector evaluation

// Uses the PASCAL Visual Object Challenge protocol to evaluate
// detector mistakes: http://pascallin.ecs.soton.ac.uk/challenges/VOC/
//
// We match ground truth objects to detections by checking the
// intersection area / union area of their bounding boxes.  
//
//   ground truth object matches a detection: true positive
//   ground truth object has no match:        false negative
//   detection has no match:                  false positive
//  
// If a ground truth object is matched multiple times, all matches
// beyond the first count as false positives.
void vocMistakes(int &tp, int &fp, int &fn, 
                 vector<DetectedRegion> const &gt, 
                 vector<DetectedRegion> const &det,
                 float minOverlapFrac)
{
  assert(minOverlapFrac > 0);
  vector<bool> gtIsMatched(det.size(), 0);
  tp = fp = fn = 0;

  for (size_t deti=0; deti<det.size(); deti++) {
    DetectedRegion const &d = det[deti];

    // Find best match for this detection
    float  bestOverlapFrac = -numeric_limits<float>::infinity();
    size_t bestMatch       = (size_t)-1;
    for (size_t gti=0; gti<gt.size(); gti++) {
      if (gtIsMatched[gti]) continue;
      DetectedRegion const &g = gt[gti];

      float const bix1 = max(d.x,            g.x);
      float const biy1 = max(d.y,            g.x);
      float const bix2 = min(d.x + d.width , g.x + g.width );
      float const biy2 = min(d.y + d.height, g.y + g.height);
      float const intersectArea = (bix2-bix1+1) * (biy2-biy1+1);
      
      float const bux1 = min(d.x,            g.x);
      float const buy1 = min(d.y,            g.x);
      float const bux2 = max(d.x + d.width , g.x + g.width );
      float const buy2 = max(d.y + d.height, g.y + g.height);
      float const unionArea = (bux2-bux1+1) * (buy2-buy1+1);

      float const overlapFrac = intersectArea / unionArea;
      
      if ((intersectArea > 0) && (overlapFrac > bestOverlapFrac)) { // greedy
        bestOverlapFrac = overlapFrac;
        bestMatch       = gti;
      }
    }

    // Account for detection matching result
    if (bestOverlapFrac > minOverlapFrac) {
      gtIsMatched[bestMatch] = 1;
      tp++;
    } else {
      fp++;
    }
  }

  // Account for ground truth objects that weren't matched
  for (size_t i=0; i<gtIsMatched.size(); i++) {
    if (!gtIsMatched[i]) fn++;
  }
}

void mistakesFromScores(int   &mistakes,
                        float &precision,
                        float &recall,
                        vector<DetectedRegion>   const &gt,
                        GrayImage                const &mask,
                        vector<ScoresGivenScale> const &scores,
                        float      svmThreshold, 
                        SigmaType  scoreToProb,
                        NonmaxType nonmaxSigma, 
                        int        whichTransferFunc,
                        float      minOverlapFrac)
{
  vector<DetectedRegion> det;
  int tp=0, fp=0, fn=0;
  detectionsFromClassifierOutput(det, &mask, scores, 
                                 svmThreshold, scoreToProb, 
                                 nonmaxSigma, whichTransferFunc);
  vocMistakes(tp, fp, fn, gt, det, minOverlapFrac);
  mistakes  = fp + fn;
  precision = tp / (float)(tp+fp);
  recall    = tp / (float)(tp+fn);
}

void evalDetector(string const &fname,
                  float minSvmThresh, float maxSvmThresh, float svmThreshStep,
                  int minNonmaxWidth, int maxNonmaxWidth, int nonmaxWidthStep,
                  int minNonmaxHeight, int maxNonmaxHeight, int nonmaxHeightStep,
                  vector<DetectedRegion> const &gt, GrayImage const &mask, 
                  vector<ScoresGivenScale> const &cScores, 
                  vector<ScoresGivenScale> const &gScores)
{
  cout << "       svm       non-max      mistakes     precision     recall     eval time" << endl;
  cout << "    threshold  width height   cpu   gpu    cpu   gpu    cpu   gpu     (ms)   " << endl;
  cout << "    ---------  ----- ------  ----- -----  ----- -----  ----- -----  ---------" << endl;
  FILE *f = fname.size() ? fopen(fname.c_str(),"wt") : NULL;
  if (f) {
    fprintf(f, "svmThreshold,nonmaxWidth,nonmaxHeight,cpuMistakes,gpuMistakes,"
            "cpuPrecision,gpuPrecision,cpuRecall,gpuRecall,evalTime\n");
  }
  for (float svmThreshold=maxSvmThresh; svmThreshold>=minSvmThresh; 
       svmThreshold-=svmThreshStep) {
    float bestPrecision = 0;
    for (int nmWidth=minNonmaxWidth; nmWidth<=maxNonmaxWidth; 
         nmWidth+=nonmaxWidthStep) {
      for (int nmHeight=minNonmaxHeight; nmHeight<=maxNonmaxHeight; 
           nmHeight+=nonmaxHeightStep) {
        unsigned    timerId = StopWatch::create();     
        StopWatchC &timer   = StopWatch::get(timerId); 
        timer.start();      

        NonmaxType const nonmaxSigma(nmWidth, nmHeight, nonmaxScale);
        int   cMistakes;
        float cPrecision, cRecall;
        mistakesFromScores(cMistakes, cPrecision, cRecall, 
                           gt, mask, cScores, 
                           svmThreshold, scoreToProb, nonmaxSigma, 
                           whichTransferFunc, minOverlapFrac);
        int   gMistakes;
        float gPrecision, gRecall;
        mistakesFromScores(gMistakes, gPrecision, gRecall, 
                           gt, mask, gScores, 
                           svmThreshold, scoreToProb, nonmaxSigma, 
                           whichTransferFunc, minOverlapFrac);

        timer.stop();
        float const ms = timer.getTime();
        StopWatch::destroy(timerId);

        printf("    % 9.1f  % 5d % 5d  % 5d % 5d  % 5.2f % 5.2f  % 5.2f % 5.2f % 9.2f\n",
               svmThreshold, nmWidth, nmHeight, cMistakes, gMistakes,
               cPrecision, gPrecision, cRecall, gRecall, ms/1000);
        fflush(stdout);
        if (f) {
          fprintf(f, "%f,%d,%d,%d,%d,%f,%f,%f,%f,%f\n",
                  svmThreshold, nmWidth, nmHeight, cMistakes, gMistakes,
                  cPrecision, gPrecision, cRecall, gRecall, ms/1000);
        }

        if (cPrecision > bestPrecision) bestPrecision = cPrecision;
        if (gPrecision > bestPrecision) bestPrecision = gPrecision;
        // Early exit if we've hit an svmThreshold that's too time
        // consuming.
        if (ms > maxOptTestMS) break;
      }
    }
    // Low svmThresholds result in massive numbers of false
    // positives which slow down the meanshift and non-maxima
    // suppression (by orders of magnitude in time).  If we've hit a
    // threshold that's giving us unacceptable precision regardless
    // of other parameter settings, exit early.
    if (bestPrecision < minOptPrecision) break;
  }
  if (f) fclose(f);
  printf("\n");
  fflush(stdout);
}

//////////////////////////////////////////////////////////////////////////////
// Other helpers (visualization & reporting)

void highlightWithDetections(pdcuda::RGBImage *detOut, 
                             vector<DetectedRegion> const &det, 
                             int channel, float alpha)
{
  for (size_t i=0; i<det.size(); i++) {
    int const w = min(detOut->extent()[0], det[i].width);
    for (int x=0; x<w; x++) {
      int const h = min(detOut->extent()[1], det[i].height);
      for (int y=0; y<h; y++) {        
        float &val = (*detOut)(det[i].x+x, det[i].y+y)(channel);
        val = val*(1-alpha) + alpha*255;
      }
    }
  }
}

string getNextProfileName(vector<vector<string> > const &names) {
  // Find the first vector whose head does not appear in the tail of
  // any other vector.  This algorithm assumes there are no duplicate
  // names in any vector and that there are no empty vectors.
  for (size_t i=0; i<names.size(); i++) {
    string const &head(names[i][0]);
    for (size_t j=0; j<names.size(); j++) {
      if (i!=j) {
        vector<string> const &nl = names[j];
        if (find(++(nl.begin()), nl.end(), head) == nl.end()) {
          return head;
        }
      }
    }
  }
  return "LOGIC ERROR"; 
}

void printTimingProfile(char const *progname,
                        TimerSet &extraTimer, TimerSet &sharedTimer,
                        TimerSet &cpuTimer,   TimerSet &gpuTimer)
{
  // Gather all the data and statistics we'll need

  vector<string> profileNames;
  {
    vector<vector<string> > names;
    names.push_back(sharedTimer.timerNames());
    names.push_back(cpuTimer.timerNames());
    names.push_back(gpuTimer.timerNames());
    while (true) {
      for (int i=(int)names.size()-1; i>=0; i--) {
        if (!names[i].size()) names.erase(names.begin()+i);
      }
      if (!names.size()) break;

      string const &nextName = getNextProfileName(names);

      profileNames.push_back(nextName);
      for (size_t i=0; i<names.size(); i++) {
        if (names[i][0] == nextName) names[i].erase(names[i].begin());
      }
    }
  }
  size_t const headerLen = std::max(sharedTimer.maxNameLen(), 
                                    std::max(cpuTimer.maxNameLen(),
                                             gpuTimer.maxNameLen()));

  float totShared = 0.f;
  for (size_t i=0; i<sharedTimer.timerNames().size(); i++) {
    totShared += sharedTimer.timer((int)i)->getTime();
  }
  float totCpu = 0.f;
  for (size_t i=0; i<cpuTimer.timerNames().size(); i++) {
    totCpu += cpuTimer.timer((int)i)->getTime();
  }
  float totGpu = 0.f;
  for (size_t i=0; i<gpuTimer.timerNames().size(); i++) {
    totGpu += gpuTimer.timer((int)i)->getTime();
  }


  // Print out the profile

  cout << progname << endl;

  cout << "  misc. times:" << endl;

  extraTimer.printTimes();

  cout << endl;
  
  cout << "  timing profile:" << endl;

  cout << "    "
       << setw(headerLen) << right << "what"           << "  "
       << setw(9+7)       << left  << "CPU time (ms)"  << "  "
       << setw(9+7)       << left  << "GPU time (ms)"  << "  "
       << setw(7)         << left  << "speedup"        << endl;

  cout << "    "
       << string(headerLen, '-') << "  "
       << string(9+7, '-')       << "  "
       << string(9+7, '-')       << "  "
       << string(7, '-')         << endl;

  vector<string> const &sn(sharedTimer.timerNames());
  vector<string> const &cn(   cpuTimer.timerNames());
  vector<string> const &gn(   gpuTimer.timerNames());
  for (size_t i=0; i<profileNames.size(); i++) {
    string const &name = profileNames[i];
    cout << "    "
         << setw(headerLen) << right << name << "  " << flush;

    int hideSpeedup = 2;

    float ct = -1; // CPU Time
    if (find(cn.begin(), cn.end(), name) != cn.end()) {
      ct = cpuTimer.timer(name)->getTime();
      hideSpeedup--;
    } else if (find(sn.begin(), sn.end(), name) != sn.end()) {
      ct = sharedTimer.timer(name)->getTime();
    }
    if (ct == -1) {
      cout << string(9+7, ' ') << "  ";
    } else {
      printf("% 9.2f (% 3.f%%)  ", ct, 100*(ct/(totCpu+totShared)));
      fflush(stdout);
    }
      
    float gt = -1; // GPU Time
    if (find(gn.begin(), gn.end(), name) != gn.end()) {
      gt = gpuTimer.timer(name)->getTime();
      hideSpeedup--;
    } else if (find(sn.begin(), sn.end(), name) != sn.end()) {
      gt = sharedTimer.timer(name)->getTime();
    }
    if (gt == -1) {
      cout << string(9+7, ' ') << "  ";
    } else {
      printf("% 9.2f (% 3.f%%)  ", gt, 100*(gt/(totGpu+totShared)));
      fflush(stdout);
    }
    
    if (hideSpeedup > 0) {
      cout << endl;
    } else {
      printf("% 6.1fx\n", ct/gt);
      fflush(stdout);
    }
  }

  cout << endl;
  cout << "    "
       << setw(headerLen) << right << "TOTAL:" << "  " << flush;
  printf("% 9.2f         ", totCpu+totShared); 
  printf("% 9.2f         ", totGpu+totShared); 
  printf("% 6.1fx", (totCpu+totShared) / (totGpu+totShared)); 
  fflush(stdout);
  cout << endl;

  cout << endl;
}



//////////////////////////////////////////////////////////////////////////////
// Main

int main(int, char const **argv)
{
  string     const infile("input.jpg");
  string     const maskFile("input.mask.png");
  string     const headTipsFile("input.headtips.png");
  string     const calibFile("input.calib.txt");
  int        const windowRadius   = 2;
#ifdef WIN32
  float      const minScale       = 0.75f; 
#else
  float      const minScale       = 0.5f; 
#endif
  float      const maxScale       = 1.0f;
  int        const nScales        = 5;
  float      const scaleFactor    = pow(maxScale / minScale, 1.f / nScales);
  string     const outdir("tcpout");
  bool       const useLanczos     = false;//true;
  float      const svmThreshold   = 0;//-0.5f;
  int        const nonmaxWidth  = 4;
  int        const nonmaxHeight = 12;

  // Detector optimization takes a LONG time (~7s/test on a Core i7
  // 965 when svmThreshold=-1.5).
  bool       const optimizeDetector = false; 

  FeatureVector const hClassifier(PERSON_WEIGHT_VEC);
  float         const classifierBias = 6.6657914910925990525925044494215f; // from WinDetect.cpp

  TimerSet extraTimer, sharedTimer, cpuTimer, gpuTimer;

  {
    TicToc tt(extraTimer,"warm up GPU");
    TransposedDeviceArray2D<blitz::TinyVector<float,3> > r(1,1);
    r.toDevice();
    r.fromDevice();
  }

  
  {
    TicToc tt(extraTimer,"output setup");
    mkDir(outdir.c_str());
  }

  
  pdcuda::RGBImage in;
  { 
    TicToc tt(sharedTimer,"read input (CPU)");
    lear::ImageIO::read(infile, in);
  }
  int const wi = in.extent()[0];
  int const hi = in.extent()[1];


  GrayImage headTips, mask;
  vector<DetectedRegion> gt;
  {
    TicToc tt(extraTimer,"prepare ground-truth data");
    lear::ImageIO::read(headTipsFile, headTips);
    lear::ImageIO::read(maskFile, mask);
    lear::PerspectiveProjector proj(calibFile, IndexType(WinSizeX, WinSizeY),
                                    BorderPixels, PedestrianHeight);
    proj.detectedRegionsFromHeadTips(gt, headTips, mask);
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // GPU Pipeline

  gpuTimer.tic("GPU resizer setup");
  auto_ptr<ImageResizer> resizer;
  if (useLanczos) {
    resizer.reset(new ImageResizerLanczos2D((float const *)in.data(), wi, hi, 
                                            windowRadius, true, true));
  } else {
    resizer.reset(new ImageResizerBilinear2D((float const *)in.data(), wi, hi,
                                             true, true));
  }

  vector<ScoresGivenScale> gScores; // results go here
  for (float scale=minScale; scale<=maxScale; scale*=scaleFactor) {
    int const w = (int)(wi/scale);
    int const h = (int)(hi/scale);
    int const bw = horizBlocks(w, GoldFineStrideX);
    int const bh = vertBlocks(h, GoldFineStrideY);
    int const ww = horizWindows(bw, GoldFineStrideX);
    int const wh = vertWindows(bh, GoldFineStrideY);

    gpuTimer.tic("resize");
    ScopedDeviceArray<float> dResized(resizer->resize(w, h));

    gpuTimer.tic("gradients");
    ScopedDeviceArray<float> dMag(w*h);
    ScopedDeviceArray<float> dOri(w*h);
    computeSpatialGradients(dMag.get(), dOri.get(), dResized.get(), w, h);  

    gpuTimer.tic("normalized block descriptors");
    ScopedDeviceArray<float> dBD(computeRhogDense(
      dMag.get(),dOri.get(), w,h, GoldFineStrideY, GoldFineStrideY,
      DefaultEpsilon, DefaultMaxVal, Epsilon2));

    gpuTimer.tic("window classification");
    ScopedDeviceArray<float> dScore(ww*wh);

    classifyWindows(dScore.get(), dBD.get(), bw, bh, 
                    GoldFineStrideX, GoldFineStrideY, 
                    &hClassifier(0), classifierBias);

    ScoresGivenScale gScore(ww, wh, scale);
    cutilSafeCall(cudaMemcpy(gScore.scores.get(), dScore.get(), 
                             ww*wh*sizeof(float), cudaMemcpyDeviceToHost));
    gScores.push_back(gScore);

    gpuTimer.tic("scale cleanup");
  }

  vector<DetectedRegion> gDetections;
  {
    TicToc tt(gpuTimer, "detection (CPU)");
    NonmaxType nonmaxSigma(nonmaxWidth, nonmaxHeight, nonmaxScale);
    detectionsFromClassifierOutput(gDetections, NULL, gScores, 
                                   svmThreshold, scoreToProb, 
                                   nonmaxSigma, whichTransferFunc);
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // CPU Pipeline

  vector<ScoresGivenScale> cScores; // results go here
  for (float scale=minScale; scale<=maxScale; scale*=scaleFactor) {
    int const w = (int)(wi/scale);
    int const h = (int)(hi/scale);
    int const bw = horizBlocks(w, GoldFineStrideX);
    int const bh = vertBlocks(h, GoldFineStrideY);
    int const ww = horizWindows(bw, GoldFineStrideX);
    int const wh = vertWindows(bh, GoldFineStrideY);

    cpuTimer.tic("resize");
    pdcuda::RGBImage resized(lear::rescale(in, IndexType(w,h)));
    
    cpuTimer.tic("gradients");
    Magnitude   hMag;
    Orientation hOri;
    computeColorGradientsCPU(&hMag, &hOri, resized); // warmup the device
     
    cpuTimer.tic("normalized block descriptors");
    BlockDescriptors chBD;
    computeRhogDenseCPU(&chBD, &hMag, &hOri);

    cpuTimer.tic("window classification");
    ScoreMap chScores;
    computeClassificationScoresCPU(&chScores, &chBD, 
                                   &hClassifier, classifierBias, true);
    ScoresGivenScale cScore(ww, wh, scale);
    memcpy(cScore.scores.get(), &chScores(0,0), ww*wh*sizeof(float)); 
    cScores.push_back(cScore);

    cpuTimer.tic("scale cleanup");
    // Fall through: cpuTimer.toc will get called automatically by the
    // next cpuTimer.tic call, allowing us to capture the time for local
    // variable destructor calls.
  }

  vector<DetectedRegion> cDetections;
  {
    TicToc tt(cpuTimer, "detection (CPU)");
    NonmaxType nonmaxSigma(nonmaxWidth, nonmaxHeight, nonmaxScale);
    detectionsFromClassifierOutput(cDetections, NULL, cScores, 
                                   svmThreshold, scoreToProb, 
                                   nonmaxSigma, whichTransferFunc);
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  {
    TicToc tt(extraTimer,"classifier comparison");
    cout << "Classification score comparisons:" << endl;
    size_t i = 0;
    for (float scale=minScale; scale<=maxScale; scale*=scaleFactor, i++) {
      int const ww = cScores[i].w;
      int const wh = cScores[i].h;

      float const *gPtr = gScores[i].scores.get(); 
      float const *cPtr = cScores[i].scores.get(); 
      double rms = 0.f;
      for (int j=0; j<ww*wh; j++) {
        float const diff = (*gPtr++ - *cPtr++);
        rms += (diff*diff);
      }
      rms = sqrt(rms/(ww*wh));
      
      cout << "  at scale " << scale << ", rms = " << rms << endl;
    }
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  {
    TicToc tt(extraTimer, "detector evaluation");
    evalDetector("",
                 svmThreshold,svmThreshold,0.1f,  
                 nonmaxWidth,nonmaxWidth,1,  
                 nonmaxHeight,nonmaxHeight,1,
                 gt, mask, cScores, gScores);
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  if (optimizeDetector) {
    TicToc tt(extraTimer, "detector optimization");
    
    cout << "  detector optimization:" << endl;
    evalDetector("det_optimization.csv",
                 -1.5f, 0.2f, 0.1f,  2,16,2,  4,32,4,
                 gt, mask, cScores, gScores);
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  {
    TicToc tt(extraTimer,"classifier output");
    size_t i = 0;
    for (float scale=minScale; scale<=maxScale; scale*=scaleFactor, i++) {
      int const ww = cScores[i].w;
      int const wh = cScores[i].h;

      char scaleName[256];
      sprintf(scaleName, "%1.2f", scale);
      
      {
        blitz::Array<float,2> tmp(cScores[i].scores.get(), IndexType(wh,ww), 
                                  blitz::neverDeleteData);
        imwriteScaledTransposed(outdir + "/" + "cScore_" + scaleName + ".png",
                                &tmp, svmThreshold);
      }

      {
        blitz::Array<float,2> tmp(gScores[i].scores.get(), IndexType(wh,ww), 
                                  blitz::neverDeleteData);
        imwriteScaledTransposed(outdir + "/" + "gScore_" + scaleName + ".png",
                                &tmp, svmThreshold);
      }
    }
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  {
    TicToc tt(extraTimer, "detector output");
    pdcuda::RGBImage detOut(in);
    float alpha = 0.25f;

    // red highlight = CPU detections
    highlightWithDetections(&detOut, cDetections, 0, alpha);
    // green highlight = GPU detections
    highlightWithDetections(&detOut, gDetections, 1, alpha);

    lear::ImageIO::write(outdir + "/detections.png", detOut);
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  {
    TicToc tt(extraTimer,"free results");
    cDetections.clear();
    cScores.clear();
    gDetections.clear();
    gScores.clear();
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  printTimingProfile(argv[0], extraTimer, sharedTimer, cpuTimer, gpuTimer);
    
  exit(1); // Need to actually check for errors before allowing the test to pass
}

// See scores with unscaled image:
//   make -j16 && (rm -rf app/tcpout ; (cd app ; ./testcudapipeline) ; (montage -border 1 -bordercolor blue -geometry 233x120 -background blue app/tcpout/cScore_1.00.png app/tcpout/gScore_1.00.png t.png && display t.png))

// Visualize final detections:
//   make -j16 && (rm -rf app/tcpout ; (cd app ; ./testcudapipeline) ; display app/tcpout/detections.png)
