#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <vector>

#include <cuda_runtime.h>
#include <cutil/cutil_inline.h>
#include <cutil/error_checker.h>

#include <pdcuda/imtransfer.h>

#ifdef _WIN32
// We don't care about int->bool conversion performance loss.
#  pragma warning (disable: 4800) 
#endif

using namespace std;
using namespace pdcuda;

#define CHECK(cond) \
  ErrorChecker::condition(cond, __FILE__, __LINE__)

#define CheckSameCtx(a,b,ctx)                                             \
  if ((a) != (b)) {                                                       \
    cerr << "ERROR in " << __FILE__ << ":" << __LINE__ << endl;           \
    cerr << "  (" #a " = " << a << ") != (" #b " = " << b << ")" << endl; \
    cerr << "  CONTEXT: " << ctx << endl;                                 \
    cerr.flush();                                                         \
    exit(1);                                                              \
  }  

#define CheckSame(a,b)                                                    \
  if ((a) != (b)) {                                                       \
    cerr << "ERROR in " << __FILE__ << ":" << __LINE__ << endl;           \
    cerr << "  (" #a " = " << a << ") != (" #b " = " << b << ")" << endl; \
    cerr.flush();                                                         \
    exit(1);                                                              \
  }  


inline void 
printImage(char const *name, float const *img, int w, int h, int d, int pw) {
  cerr << endl << name << ":" << endl;
  int i=0;
  for (int y=0; y<h; y++) {
    for (int x=0; x<w; x++) {
      for (int c=0; c<3; c++) {
        cerr << setw(pw) << img[i++] << " ";
      }
      cerr << "  ";
    }
    cerr << endl;
  }
  cerr << endl;
}

///////////////////////////////////////////////////////////////////////////////

void testRgbToDeviceRgba(float const *hRgb, float const *hRgbT, int w, int h)
{
  //cerr << "testing " << w << "x" << h << endl;

  unsigned const S = w*h*3*sizeof(float);                                     

  float *dData = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dData, w*h*3*sizeof(float)));

  for (int dataColumnMajor=0; dataColumnMajor<=1; dataColumnMajor++) {
    for (int dataOnHost=0; dataOnHost<=1; dataOnHost++) {
      vector<float> ghRgba(w*h*4);
      {
        float const *input = NULL;
        if (dataColumnMajor) {
          input = hRgbT;
          if (!dataOnHost) {
            input = dData;
            cutilSafeCall(cudaMemcpy(dData, hRgbT, S, cudaMemcpyHostToDevice));
          }
        } else {
          input = hRgb;
          if (!dataOnHost) {
            input = dData;
            cutilSafeCall(cudaMemcpy(dData, hRgb, S, cudaMemcpyHostToDevice));
          }
        }

        float4 *dRgba = rgbToDeviceRgba(input, w, h, 
                                        dataOnHost, dataColumnMajor);
        cutilSafeCall(cudaMemcpy(&ghRgba[0], dRgba, w*h*4*sizeof(float), 
                                 cudaMemcpyDeviceToHost));
        cutilSafeCall(cudaFree(dRgba));
      }

      for (int y=0; y<h; y++) {
        for (int x=0; x<w; x++) {
          for (int c=0; c<3; c++) {
            unsigned inIdx = dataColumnMajor ? 
              (x*h*3 + y*3 + c) : (y*w*3 + x*3 + c);
            CheckSameCtx(ghRgba[y*w*4 + x*4 + c], hRgb[inIdx],
                         "w:"<<w<<" h:"<<h<<
                         " dataOnHost:"<<dataOnHost<<
                         " dataColumnMajor:"<<dataColumnMajor<<
                         " y:"<<y<<" x:"<<x<<" c:"<<c);
          }
        }
      }    
    }
  }

  cutilSafeCall(cudaFree(dData));
}

///////////////////////////////////////////////////////////////////////////////

texture<float4, 2, cudaReadModeElementType> t2;

// Simple kernel to extract data from a 2D CUDA texture to a row-major
// packed RGB image.  
__global__
static void Tex2ExtractRgb_Kernel(float *out, int w, int h)
{
  int const x = (threadIdx.x + blockDim.x*blockIdx.x);
  int const y = (threadIdx.y + blockDim.y*blockIdx.y);

  if (x >= w || y >= h) return;

  float4 const rgba = tex2D(t2, x+0.5, y+0.5);
  
  out[y*w*3 + x*3    ] = rgba.x;
  out[y*w*3 + x*3 + 1] = rgba.y;
  out[y*w*3 + x*3 + 2] = rgba.z;
}

void testRgbToCudaArray2d(float const *hRgb, float const *hRgbT, int w, int h)
{
  //cerr << "testing " << w << "x" << h << endl;

  unsigned const SZ = w*h*3*sizeof(float); // RGB bytes

  float *dData = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dData, SZ));

  for (int dataColumnMajor=0; dataColumnMajor<=1; dataColumnMajor++) {
    for (int dataOnHost=0; dataOnHost<=1; dataOnHost++) {
      vector<float> ghRgb(w*h*3);
      {
        // Pick input data
        float const *input = NULL;
        if (dataColumnMajor) {
          input = hRgbT;
          if (!dataOnHost) {
            input = dData;
            cutilSafeCall(cudaMemcpy(dData,hRgbT, SZ,cudaMemcpyHostToDevice));
          }
        } else {
          input = hRgb;
          if (!dataOnHost) {
            input = dData;
            cutilSafeCall(cudaMemcpy(dData,hRgb, SZ,cudaMemcpyHostToDevice));
          }
        }

        // Transfer to a 2D cudaArray (this is the call we're testing)
        cudaArray *dRgba = rgbToCudaArray2d(input, w, h, 
                                            dataOnHost, dataColumnMajor);

        // Lookup the data so we can validate it
        cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();
        t2.addressMode[0] = cudaAddressModeClamp; // Pixel-replicate @ borders
        t2.addressMode[1] = cudaAddressModeClamp; // Pixel-replicate @ borders
        t2.filterMode     = cudaFilterModePoint;  // No interpolation
        t2.normalized     = false;                // Uses ([0,w),[0,h)) coords

        cutilSafeCall(cudaBindTextureToArray(t2, dRgba, channelDesc));
        
        cudaGetLastError(); // reset error status
        dim3 const block(16,1);
        dim3 const grid((int)ceil(w/float(block.x)), 
                        (int)ceil(h/float(block.y)));
        Tex2ExtractRgb_Kernel<<<grid,block>>>(dData, w, h);
        cutilSafeCall(cudaGetLastError());

        cutilSafeCall(cudaUnbindTexture(t2));

        cutilSafeCall(cudaMemcpy(&ghRgb[0],dData, SZ,cudaMemcpyDeviceToHost));

        cutilSafeCall(cudaFreeArray(dRgba));
      }

      // Did we read back the correct values?
      for (int y=0; y<h; y++) {
        for (int x=0; x<w; x++) {
          for (int c=0; c<3; c++) {
            unsigned inIdx = dataColumnMajor ? 
              (x*h*3 + y*3 + c) : (y*w*3 + x*3 + c);
            CheckSameCtx(ghRgb[y*w*3 + x*3 + c], hRgb[inIdx],
                         "w:"<<w<<" h:"<<h<<
                         " dataOnHost:"<<dataOnHost<<
                         " dataColumnMajor:"<<dataColumnMajor<<
                         " y:"<<y<<" x:"<<x<<" c:"<<c);
          }
        }
      }    
    }
  }

  cutilSafeCall(cudaFree(dData));
}


///////////////////////////////////////////////////////////////////////////////

texture<float, 3, cudaReadModeElementType> t3;

// Simple kernel to extract data from a 3D CUDA texture to a row-major
// packed RGB image.  
__global__
static void Tex3ExtractRgb_Kernel(float *out, int w, int h)
{
  int const x = (threadIdx.x + blockDim.x*blockIdx.x);
  int const y = (threadIdx.y + blockDim.y*blockIdx.y);

  if (x >= w || y >= h) return;

  out[y*w*3 + x*3    ] = tex3D(t3, x+0.5f, y+0.5f, 0.5f);
  out[y*w*3 + x*3 + 1] = tex3D(t3, x+0.5f, y+0.5f, 1.5f);
  out[y*w*3 + x*3 + 2] = tex3D(t3, x+0.5f, y+0.5f, 2.5f);
}

void testRgbToCudaArray3d(float const *hRgb, float const *hRgbT, int w, int h)
{
  //cerr << "testing " << w << "x" << h << endl;

  unsigned const SZ = w*h*3*sizeof(float); // RGB bytes

  float *dData = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dData, SZ));

  for (int dataColumnMajor=0; dataColumnMajor<=1; dataColumnMajor++) {
    for (int dataOnHost=0; dataOnHost<=1; dataOnHost++) {
      vector<float> ghRgb(w*h*3);
      {
        // Pick input data
        float const *input = NULL;
        if (dataColumnMajor) {
          input = hRgbT;
          if (!dataOnHost) {
            input = dData;
            cutilSafeCall(cudaMemcpy(dData,hRgbT, SZ,cudaMemcpyHostToDevice));
          }
        } else {
          input = hRgb;
          if (!dataOnHost) {
            input = dData;
            cutilSafeCall(cudaMemcpy(dData,hRgb, SZ,cudaMemcpyHostToDevice));
          }
        }

        // Transfer to a 2D cudaArray (this is the call we're testing)
        cudaArray *dRgb = rgbToCudaArray3d(input, w, h, 
                                           dataOnHost, dataColumnMajor);

        // Lookup the data so we can validate it
        cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float>();
        t3.addressMode[0] = cudaAddressModeClamp; // Pixel-replicate @ borders
        t3.addressMode[1] = cudaAddressModeClamp; // Pixel-replicate @ borders
        t3.addressMode[2] = cudaAddressModeClamp; // Pixel-replicate @ borders
        t3.filterMode     = cudaFilterModePoint;  // No interpolation
        t3.normalized     = false;                // Uses ([0,w),[0,h)) coords

        cutilSafeCall(cudaBindTextureToArray(t3, dRgb, channelDesc));
        
        cudaGetLastError(); // reset error status
        dim3 const block(16,2,1);
        dim3 const grid((int)ceil(w/float(block.x)), 
                        (int)ceil(h/float(block.y)));
        Tex3ExtractRgb_Kernel<<<grid,block>>>(dData, w, h);
        cutilSafeCall(cudaGetLastError());

        cutilSafeCall(cudaUnbindTexture(t3));

        cutilSafeCall(cudaMemcpy(&ghRgb[0],dData, SZ,cudaMemcpyDeviceToHost));

        cutilSafeCall(cudaFreeArray(dRgb));
      }
      
      //printImage("ghRgb", (float const *)&ghRgb[0], w,h,3, 3);
      //printImage("hRgb",  (float const *)&hRgb[0],  w,h,3, 3);

      // Did we read back the correct values?
      for (int y=0; y<h; y++) {
        for (int x=0; x<w; x++) {
          for (int c=0; c<3; c++) {
            unsigned inIdx = dataColumnMajor ? 
              (x*h*3 + y*3 + c) : (y*w*3 + x*3 + c);
            CheckSameCtx(ghRgb[y*w*3 + x*3 + c], hRgb[inIdx],
                         "w:"<<w<<" h:"<<h<<
                         " dataOnHost:"<<dataOnHost<<
                         " dataColumnMajor:"<<dataColumnMajor<<
                         " y:"<<y<<" x:"<<x<<" c:"<<c);
          }
        }
      }    
    }
  }

  cutilSafeCall(cudaFree(dData));
}

///////////////////////////////////////////////////////////////////////////////

// Simple kernel to extract data from a 2D CUDA texture to a row-major
// packed RGBA image.  
__global__
static void Tex2ExtractRgba_Kernel(float4 *out, int w, int h)
{
  int const x = (threadIdx.x + blockDim.x*blockIdx.x);
  int const y = (threadIdx.y + blockDim.y*blockIdx.y);

  if (x >= w || y >= h) return;

  out[y*w + x] = tex2D(t2, x+0.5, y+0.5);
}

void testRgbaToCudaArray2d(float4 const *hRgba, float4 const *hRgbaT, 
                           int w, int h)
{
  unsigned const SZ = w*h*sizeof(float4); // RGBA bytes

  float4 *dData = NULL;
  cutilSafeCall(cudaMalloc((void**)(void*)&dData, SZ));

  for (int dataColumnMajor=0; dataColumnMajor<=1; dataColumnMajor++) {
    for (int dataOnHost=0; dataOnHost<=1; dataOnHost++) {
      vector<float> ghRgba(w*h*4);
      {
        // Pick input data
        float4 const *input = NULL;
        if (dataColumnMajor) {
          input = hRgbaT;
          if (!dataOnHost) {
            cutilSafeCall(cudaMemcpy(dData,hRgbaT, SZ,cudaMemcpyHostToDevice));
            input = dData;
          }
        } else {
          input = hRgba;
          if (!dataOnHost) {
            cutilSafeCall(cudaMemcpy(dData,hRgba, SZ,cudaMemcpyHostToDevice));
            input = dData;
          }
        }

        // Transfer to a 2D cudaArray (this is the call we're testing)
        cudaArray *dRgba = rgbaToCudaArray2d(input, w, h, 
                                             dataOnHost, dataColumnMajor);

        // Lookup the data so we can validate it
        cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();
        t2.addressMode[0] = cudaAddressModeClamp; // Pixel-replicate @ borders
        t2.addressMode[1] = cudaAddressModeClamp; // Pixel-replicate @ borders
        t2.filterMode     = cudaFilterModePoint;  // No interpolation
        t2.normalized     = false;                // Uses ([0,w),[0,h)) coords

        cutilSafeCall(cudaBindTextureToArray(t2, dRgba, channelDesc));
        
        cudaGetLastError(); // reset error status
        dim3 const block(16,1);
        dim3 const grid((int)ceil(w/float(block.x)), 
                        (int)ceil(h/float(block.y)));
        Tex2ExtractRgba_Kernel<<<grid,block>>>(dData, w, h);
        cutilSafeCall(cudaGetLastError());

        cutilSafeCall(cudaUnbindTexture(t2));

        cutilSafeCall(cudaMemcpy(&ghRgba[0],dData, SZ,cudaMemcpyDeviceToHost));

        cutilSafeCall(cudaFreeArray(dRgba));
      }

      // Did we read back the correct values?
      for (int y=0; y<h; y++) {
        for (int x=0; x<w; x++) {
          for (int c=0; c<4; c++) {
            unsigned inIdx = dataColumnMajor ? 
              (x*h*4 + y*4 + c) : (y*w*4 + x*4 + c);
            CheckSameCtx(ghRgba[y*w*4 + x*4 + c], 
                         ((float const *)hRgba)[inIdx],
                         "w:"<<w<<" h:"<<h<<
                         " dataOnHost:"<<dataOnHost<<
                         " dataColumnMajor:"<<dataColumnMajor<<
                         " y:"<<y<<" x:"<<x<<" c:"<<c);
          }
        }
      }    
    }
  }

  cutilSafeCall(cudaFree(dData));
}

///////////////////////////////////////////////////////////////////////////////

int main(int, char const **)
{
  int const Widths[]  = {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,
                         17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,
                         33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,
                         639,640,641, 1280, 1920, 2560};
  int const Heights[] = {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,
                         17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,
                         33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,
                         479,480,481, 720, 1080, 1600};

  cudaDeviceProp deviceProp;
  cutilSafeCall(cudaGetDeviceProperties(&deviceProp, 0));

  for (int hi=0; hi<sizeof(Heights)/sizeof(int); hi++) {
    int const h = Heights[hi];
    for (int wi=0; wi<sizeof(Widths)/sizeof(int); wi++) {
      int const w = Widths[wi];

      if (((w*h*3*4)*2 + (w*h*4*4)*4) > deviceProp.totalGlobalMem * 0.48) {
        // At least on Vista x64, we seem to only be able to use about
        // 48% of the total physical memory.  Skip tests that are
        // likely to generate out-of-memory errors.
        continue;
      }

      {
        vector<float> hRgb(w*h*3);
        vector<float> hRgbT(w*h*3);
        {
          for (int y=0; y<h; y++) {
            for (int x=0; x<w; x++) {
              for (int c=0; c<3; c++) {
#ifdef _WIN32
#  pragma warning (push)
#  pragma warning (disable: 4244) // suppress int->float precision warning
#endif
                hRgb [y*w*3 + x*3 + c] = y*w*3 + x*3 + c;
                hRgbT[x*h*3 + y*3 + c] = x*h*3 + y*3 + c;
#ifdef _WIN32
#  pragma warning (pop)
#endif
              }
            }
          }
        }
        
        testRgbToDeviceRgba(&hRgb[0], &hRgbT[0], w, h);
        testRgbToCudaArray2d(&hRgb[0], &hRgbT[0], w, h);
        if (valid3dRgbArraySize(w,h)) {
          testRgbToCudaArray3d(&hRgb[0], &hRgbT[0], w, h);
        }
        //testRgbToFft2d(&hRgb[0], &hRgbT[0], w, h);
        //testRgbToFft3d(&hRgb[0], &hRgbT[0], w, h);
      }

      {
        vector<float> hRgba(w*h*4);
        vector<float> hRgbaT(w*h*4);
        {
          for (int y=0; y<h; y++) {
            for (int x=0; x<w; x++) {
              for (int c=0; c<4; c++) {
                hRgba [y*w*4 + x*4 + c] = y*w*4 + x*4 + c;
                hRgbaT[x*h*4 + y*4 + c] = x*h*4 + y*4 + c;
              }
            }
          }
        }

        testRgbaToCudaArray2d((float4*)&hRgba[0], (float4*)&hRgbaT[0], w, h);
      }
    }
  }

  cerr << "SUCCESS!" << endl;
}
