function varargout = ibeamAnnotationGui(varargin)
% IBEAMANNOTATIONGUI interface for interactive annotation of pedestrian
%   locations using non-axis-aligned I-Beams.
% 
% ibeamAnnotationGui(imgname)
%   Create new annotations for the image in the given filename.
%
% ibeamAnnotationGui(ibeamFile)
%   Edits a whole set of existing annotations, where we load an
%   IBEAMHIER structure from a .MAT file.  The .MAT file's directory
%   is treated as the BASEDIR.
%
% ibeamAnnotationGui(basedir,ibeamHier)
%   starts with an existing ibeamHier relative to a given base directory.
%
% ibeamAnnotationGui(basedir,ibeamFilename)
%   Same as ibeamAnnotationGui(ibeamFile), but the basedir is explicitly
%   specified. If the file doesn't exist, we recursively import all images
%   under the basedir.
%
% ibeamAnnotationGui(imgname,locs)
%   Edit exiting annotations for the image in the given filename.  See
%   below for a definition of LOCS.
%
% [ibeamHier, basedir, ibeamFilename] = ibeamAnnotationGui(...)
%   Returns the final accepted annotations.  If available, the base
%   directory and filename of the annotations are given too.  If BASEDIR is
%   present, all 'imagename' fields of IBEAMHIER and the IBEAMFILENAME
%   are relative paths w.r.t. the BASEDIR.
%
% The saved annotations file will have a struct-array, ibeamHier, where
% each element ibeamHier(i) corresponds to one image and has the following
% fields:
%   IMAGENAME: name of the image for which the annotations are made.  The
%     full pathname is given by fullfile(basedir,imagename).
%
%   LOCS: a 4xN matrix where N is the number of annotated people for this
%     image.  Each column contains [headx;heady;footx;footy], where
%     (headx,heady) corresponds to the crown of a person's head if they are
%     standing erect and (footx,footy) corresponds to the point on the
%     ground below (headx,heady).
%
% UI CONTROLS:
%   * Clicking on an existing annotation selects it.  
%
%   * Dragging a solid-colored part of annotation moves it.
%
%   * SHIFT-click and/or CTRL-click adds/removes items from the selection
%     set.
%
%   * Clicking on the image creates a new annotation.
%
%   * Q /q      quit       exits the GUI, returning any committed edits. 
%
%   * <ESC>     unselect   deselects all annotations.
%
%   * <CTRL>-z  rollback   undoes all uncommitted edits to the current
%                          image.  
%
%   * <ENTER>   commit     commits current edits. 
%
%   * <DEL>     delete     deletes the currently-selected annotations. 
%
%   * -> / <-   navigate   arrow keys go to the next/previous image in the
%                          annotation set.  Auto-committing.
%
%   * <HOME>    first      go to the first image in the annotation set.
%                          Auto-committing.
%
%   * <END>     last       go to the first image in the annotation set.
%                          Auto-committing.
%
%   * <PGUP>    forward    Bring selected items forward
%
%   * <PGDN>    backward   Send selected items backward
%
%   * A / a     saveas     Prompts for a new annotation file for saving
%
%   * G / g     goto       select a particular image to go to.  
%
%   * H / h     hide       toggles between hiding and showing unselected
%                          annotations.
%
%   * S / s     save       saves to the current annotation file (reverts to
%                          saveas if there isn't a current one)
%
%   * <PLUS>    zoom in    2x zoom in, at the mouse' center
%
%   * <MINUS>   zoom out   2x zoom out, at the mouse' center
%
%   * F / f     flow       predict person locations using the previous
%                          frame's annotations and optical flow estimates
%
% EXAMPLES:
%   ibeamAnnotationGui(grandCentralDir('test/ngct01'), 'ibeams_20080221.mat')


% Declare non-UI data here so that they can be used in any functions in
% this GUI file.

% Master table of annotations
mBaseDir       = '';
mIbeamFilename = '';
mIbeamHier     = struct('imagename',{}, 'locs',[]); % master annotations structure

% Image being editing right now
mCurrImgNum      = -1;         % which entry in mIbeamHier are we using?
mLocs            = zeros(4,0); % IBeams for the current image
mImage           = [];         % Current image

% UI Theme
mNearTaggedPt2          = 1.^2;  % mouse deadzone for selection (squared screen pixels)
mNearVertex2            = 2.^2;  % mouse click must be within this radius of an vertex to consider it a selection (squared screen pixels)
mNearEdge2              = 3.^2;  % ...of an edge
mBorderThickenessMinScr = 1;     % min thickness for the bounding box border, in screen pixels
mInactiveColor          = [0 1 0];
mSelectedColor          = [1 0 0];
mTaggedColor            = [0 0 1];
mDefaultHeight          = 25;
mDefaultAspect          = 1/2; %20/128; % w/h aspect from exptExploreProportions

% Interaction states (see annotationGuiStateDiagram.vsd)
WAITING     = 1;
SELECTING   = 2;
UNSELECTING = 4;
MOVING      = 8;
STATES = {WAITING,SELECTING,UNSELECTING,MOVING}; %#ok<NASGU>

% Different parts of an annotation that can be tagged by the mouse
CROWN    = 1;
FOOTBASE = 2;
HEAD     = 4;
FEET     = 8;
BODY     = 16;
VERTICES = CROWN + FOOTBASE;
LINES    = HEAD + FEET + BODY;
PARTS    = LINES + VERTICES;

% Composite state machine state
mUiState = struct;
mUiState.state        = WAITING; % current state (member of STATES)
mUiState.hideInactive = false;   % when enabled, only show selected or tagged annotations (logical)
mUiState.shift        = false;   % is the <SHIFT> key pressed? (logical)
mUiState.ctrl         = false;   % is the <CTRL> key pressed? (logical)
mUiState.button       = false;   % is a mouse button down? (logical)
mUiState.prevCoords   = [-1;-1]; % scrX,scrY mouse coordinates from previous callback (x,y screen coordinates).  
mUiState.sel          = [];      % list of selected annotations (indices into mLocs)
mUiState.tagged       = [];      % annotation tagged as active (index into mLocs)
mUiState.taggedPart   = [];      % part of the tagged annotation that's tagged (member of PARTS)
mUiState.taggedCoords = [];      % scrX,scrY mouse coordinates when the tagging was done (x,y screen coordinates).  

% Create all the UI objects in this GUI here so that they can
% be used in any functions in this GUI
hMainFigure = figure(...
  'Units','characters',...
  'MenuBar','none',...
  'Toolbar','figure',...
  'WindowStyle', 'modal',...
  'visible','off',...
  'KeyPressFcn',           @hMainFigureKeyPressFcn,...
  'KeyReleaseFcn',         @hMainFigureKeyReleaseFcn,...
  'WindowButtonDownFcn',   @hMainFigureWindowButtonDownFcn,...
  'WindowButtonUpFcn',     @hMainFigureWindowButtonUpFcn,...
  'WindowButtonMotionFcn', @hMainFigureWindowButtonMotionFcn);
hImgAxes = axes(...
  'Parent',hMainFigure,...
  'vis','off',...
  'Units','normalized',...
  'Position',[0 0 1 1]);
hImg = []; 

% Make changes needed for proper look and feel and running on different
% platforms
prepareLayout(hMainFigure);

% Process the command line input arguments supplied when the GUI is
% invoked
processUserInputs(varargin{:});

% Put user-input variables into canonical form, where applicable
fixInvertedBboxes();

% Initialize the iconEditor using the defaults or custom data given through
% property/value pairs
localUpdateGui();

% Make the GUI on screen
set(hMainFigure,'visible', 'on');
maximize(hMainFigure);

% Make the GUI blocking
uiwait(hMainFigure);

% Return the final annotations if requested
outputArgs    = {};         
outputArgs{1} = mIbeamHier;
outputArgs{2} = mBaseDir;
outputArgs{3} = mIbeamFilename;
if nargout>0
  [varargout{1:nargout}] = outputArgs{:};
end

try close(hMainFigure); catch end

%------------------------------------------------------------------
  function localUpdateGui
    % helper function that updates the gui when the data changes
    %set(hImgAxes,'CData', mImage,'Visible','on');
    if isempty(hImg)
      hImg = imshow(mImage);
    else
      delete(setdiff(get(hImgAxes,'Children'), hImg));
    end
   
    [toShow,nShown] = getVisible();
    for i=1:size(mLocs,2)
      if (toShow(i))
        % TODO: make a third color when near an inactive?
        parts = [];
        if i == mUiState.tagged
          color = mSelectedColor;
          parts = mUiState.taggedPart;
        elseif ismember(i,mUiState.sel)
          color = mSelectedColor;
        else
          color = mInactiveColor;
        end
        drawAnnotation(floor(mLocs(:,i)), ...
          borderThicknessImg(nShown(i)), color, parts);
      end
    end
  end

%------------------------------------------------------------------
  function drawAnnotation(locs,thickness,color,taggedParts)
    hold on;
    % TODO: do I need to bias by +1 everywhere?

    % Find the beam offsets
    bodyvec     = (locs([3 4]) - locs([1 2]));
    height      = norm(bodyvec);
    bodyvecNorm = bodyvec / height;
    beamvecNorm = [bodyvecNorm(2) ; -bodyvecNorm(1)]; % 90deg rotation
    halfwidth   = mDefaultAspect * height / 2;
    beamvec     = beamvecNorm * halfwidth;
    beam        = [-beamvec ; beamvec ];

    line(locs([1 3]),         locs([2 4]),         'LineWidth',thickness, 'Color',colorPicker(color,taggedParts,BODY));
    
    line(locs(1)+beam([1 3]), locs(2)+beam([2 4]), 'LineWidth',thickness, 'Color',colorPicker(color,taggedParts,HEAD));
    line(locs(3)+beam([1 3]), locs(4)+beam([2 4]), 'LineWidth',thickness, 'Color',colorPicker(color,taggedParts,FEET));
        
    if bitand(CROWN,   taggedParts), plot(locs(1),locs(2), '.', 'MarkerSize',5, 'Color',mTaggedColor); end
    if bitand(FOOTBASE,taggedParts), plot(locs(3),locs(4), '.', 'MarkerSize',5, 'Color',mTaggedColor); end
    
    hold off;
    
    function c = colorPicker(untaggedColor,taggedParts,part)
      if bitand(part,taggedParts)
        c = mTaggedColor;
      else
        c = untaggedColor;
      end
    end
  end

%------------------------------------------------------------------
  function [toShow,nShown] = getVisible
    if mUiState.hideInactive
      toShow = ismember(1:size(mLocs,2), ...
                        unique([mUiState.sel mUiState.tagged]));
    else
      toShow = true(1,size(mLocs,2));
    end
    nShown = cumsum(toShow);
  end

%------------------------------------------------------------------
  function t = borderThicknessImg(annoIdx)
    t = borderThicknessScr(annoIdx)*getZoom;
  end

%------------------------------------------------------------------
  function t = borderThicknessScr(annoIdx)
    t = mBorderThickenessMinScr; %(mBorderThickenessMinScr+annoIdx);
  end

%------------------------------------------------------------------
  function z = getZoom
    % screen pixels * getZoom = img pixels
    oldUnits = get(hMainFigure, 'Units');
    set(hMainFigure, 'Units','pixels'); % TODO: can I just keep pixel units in the setup?
    pos = get(hMainFigure, 'Position');
    figWidth = pos(3);         figHeight = pos(4);
    imgWidth = size(mImage,2); imgHeight = size(mImage,1);
    figAspect = figWidth / figHeight;
    imgAspect = imgWidth / imgHeight;
    if figAspect > imgAspect % img touches top and bottom of figure
      z = imgHeight / figHeight;
    else % img touches left and right of figure
      z = imgWidth / figWidth;
    end
    set(hMainFigure, 'Units',oldUnits);
  end

%------------------------------------------------------------------
  function hMainFigureKeyPressFcn(hObject, eventdata) %#ok<INUSL>
    % Callback called when key is pressed
    mUiState.ctrl  = any(ismember(eventdata.Modifier, 'control'));
    mUiState.shift = any(ismember(eventdata.Modifier, 'shift'));
  end

%------------------------------------------------------------------
  function hMainFigureKeyReleaseFcn(hObject, eventdata) %#ok<INUSL>
    % Callback called when key is released
    mUiState.ctrl  = any(ismember(eventdata.Modifier, 'control'));
    mUiState.shift = any(ismember(eventdata.Modifier, 'shift'));    
    
    if strcmp(eventdata.Key, 'q') % quit GUI
      switch questdlg('Save before quitting?')
        case 'Cancel'
          % do nothing
        case 'Yes'
          if doSave
            uiresume;
          end
        case 'No'
          % just quit
          uiresume;
        otherwise
          error('unexpected return value');
      end
      
    elseif strcmp(eventdata.Key, 'escape') % deselect
      resetStateMachine;
      
    elseif strcmp(eventdata.Key, 'z') && mUiState.ctrl % rollback
      selectImage(mCurrImgNum);
      resetStateMachine;
      
    elseif strcmp(eventdata.Key, 'return') % commit
      commitChanges;
      
    elseif strcmp(eventdata.Key, 'delete') % delete selected
      toKeep       = setdiff(1:size(mLocs,2), mUiState.sel);
      mLocs        = mLocs(:,toKeep);
      resetStateMachine;
      
    elseif strcmp(eventdata.Key, 'rightarrow') % commit & next
      commitChanges;
      selectImage(mod(mCurrImgNum, length(mIbeamHier))+1);
      resetStateMachine;

    elseif strcmp(eventdata.Key, 'leftarrow') % commit & prev
      commitChanges;
      selectImage(mod(mCurrImgNum-2, length(mIbeamHier))+1);
      resetStateMachine;

    elseif strcmp(eventdata.Key, 'home') % commit & first
      commitChanges;
      selectImage(1);
      resetStateMachine;

    elseif strcmp(eventdata.Key, 'end') % commit & end
      commitChanges;
      selectImage(length(mIbeamHier));
      resetStateMachine;

    elseif strcmp(eventdata.Character, 'g') % goto
      newIdx = listdlg('ListString',{mIbeamHier.imagename}, ...
        'PromptString','Select an image', 'SelectionMode','single', ...
        'InitialValue',mCurrImgNum, 'ListSize',[200 400]);
      if ~isempty(newIdx)
        selectImage(newIdx);
        resetStateMachine;
      end
      
    elseif strcmp(eventdata.Character, 's') % save
      doSave;
      
    elseif strcmp(eventdata.Character, 'a') % saveas
      doSaveAs;
      
    elseif strcmp(eventdata.Character, 'h') % toggle inactive hiding
      mUiState.hideInactive = ~mUiState.hideInactive;
      localUpdateGui;
        
    elseif (strcmp(eventdata.Key, 'add') || ...
        strcmp(eventdata.Key, 'equal'))% zoom in
      gzoom(2);
      localUpdateGui;
        
    elseif strcmp(eventdata.Key, 'subtract')% zoom out
      gzoom(0.5);
      localUpdateGui;
      
    elseif strcmp(eventdata.Character, 'f') % flow prediction
      mLocs     = zeros(4,0);
      srcImgNum = mod(mCurrImgNum-2, length(mIbeamHier))+1;
      srcLocs   = mIbeamHier(srcImgNum).locs;
      srcLocs   = reshape(srcLocs, 2, numel(srcLocs)/2)';
      
      I1 = imread(fullfile(mBaseDir, mIbeamHier(srcImgNum).imagename));
      G1 = mat2gray(double(I1))*255; G1 = G1(:,:,1);
      I2 = imread(fullfile(mBaseDir, mIbeamHier(mCurrImgNum).imagename));
      G2 = mat2gray(double(I2))*255; G2 = G2(:,:,1);

      [locs,ids]   = pointtracker(G1, G2, srcLocs);
      crowns       = ids(mod(ids,2) == 0); % 0-indexed
      bases        = ids(mod(ids,2) == 1);
      wholes       = union(crowns/2, floor(bases/2));
      crownKeepers = intersect(crowns, wholes*2) + 1; % 1-indexed
      baseKeepers  = intersect(bases, wholes*2+1) + 1;
      mLocs = [locs(crownKeepers,:) locs(baseKeepers,:)]';

      resetStateMachine;
        
    elseif strcmp(eventdata.Key, 'pageup') % sort ordering: forward 
      if ~isempty(mUiState.sel)
        if mUiState.ctrl
          shiftSelected(size(mLocs,2));
        else
          shiftSelected(1.5); % extra 0.5 avoids collisions
        end
        localUpdateGui;
      end
      
    elseif strcmp(eventdata.Key, 'pagedown') % sort ordering: backward 
      if ~isempty(mUiState.sel)
        if mUiState.ctrl
          shiftSelected(-size(mLocs,2));
        else
          shiftSelected(-1.5); % extra 0.5 avoids collisions
        end
        localUpdateGui;
      end
      
    else
      % do nothing
    end
  end

%------------------------------------------------------------------
  function shiftSelected(shiftAmt)
    % Shifts the sort order of the selected annotations
    allIdcs   = 1:size(mLocs,2);
    selIdcs   = mUiState.sel;
    priorityList = allIdcs;
    priorityList(selIdcs) = selIdcs + shiftAmt;
    [yy,ii] = sort(priorityList);
    [dummy,mUiState.sel] = ismember(mUiState.sel, ii);
    if ~isempty(mUiState.tagged)
      mUiState.tagged = find(ii==mUiState.tagged);
    end
    mLocs = mLocs(:,ii);
  end
    
%------------------------------------------------------------------
  function done = doSave
    if ~isempty(mIbeamFilename)
      commitChanges;
      ibeamHier = mIbeamHier; %#ok<NASGU>
      save(fullfile(mBaseDir, mIbeamFilename), 'ibeamHier');
      clear ibeamHier;
      done = 1;
    else
      done = doSaveAs;
    end
  end

%------------------------------------------------------------------
  function done = doSaveAs
    [fname,pathname] = uiputfile(fullfile(mBaseDir, '*.mat'), ...
      'Save annotations as', mIbeamFilename);
    if fname==0
      done = 0;
    else
      commitChanges;
      ibeamHier = mIbeamHier; %#ok<NASGU>
      save(fullfile(pathname,fname), 'ibeamHier');
      clear ibeamHier;
      mBaseDir      = pathname;
      mIbeamFilename = fname;
      done = 1;
    end
  end

%------------------------------------------------------------------
  function resetStateMachine
    mUiState.state        = WAITING;
    % leave shift&ctrl alone...it's annoying to the user if we don't
    % leave button alone...to be consistent with shift&ctrl
    % leave prevCoords alone...it won't hurt anything
    mUiState.sel          = [];
    mUiState.tagged       = [];
    mUiState.taggedPart   = [];
    mUiState.taggedCoords = [];
    localUpdateGui;
  end

%------------------------------------------------------------------
  function commitChanges
    mIbeamHier(mCurrImgNum).locs = floor(mLocs);
  end

%------------------------------------------------------------------
  function hMainFigureWindowButtonDownFcn(hObject, eventdata) %#ok<INUSD>
    % Callback called when mouse is released
    mUiState.button = 1;
    if (ancestor(gco,'axes') == hImgAxes)
      eventDispatcher(0);
    end
  end

%------------------------------------------------------------------
  function hMainFigureWindowButtonMotionFcn(hObject, eventdata) %#ok<INUSD>
    % Callback called when mouse is moving
    eventDispatcher(1);
  end

%------------------------------------------------------------------
  function hMainFigureWindowButtonUpFcn(hObject, eventdata) %#ok<INUSD>
    % Callback called when mouse is released
    mUiState.button = 0;
    eventDispatcher(0);
  end

%------------------------------------------------------------------
  function eventDispatcher(mouseMoved)
    % Dispatcher for mouse-based events.  It was easier to collect the
    % state machine into a single dispatcher module.  See
    % annotationGuiStateDiagram.vsd (Visio diagram) for the design.
    [imgC,imgR,scrX,scrY] = getCurrentPoint;

    updateWindowTitle;
    
    switch mUiState.state
      case WAITING
        if mouseMoved || ~mUiState.button % WAITING(M***),WAITING(CU**): moving the mouse or spurious mouse up event
          if clearTag
            localUpdateGui;
          end
          
        elseif tagNearestObject(scrX,scrY) % WAITING(CD*S),WAITING(CD*U): near an object
          if mUiState.shift || mUiState.ctrl % WAITING(CDSS),WAITING(CDSU): near object to be toggled
            if ismember(mUiState.tagged, mUiState.sel) % WAITING(CDSS) --> UNSELECTING
              % Don't deselect yet; we might need to convert to MOVING
              % mode if the mouse moves before the button is released.
              mUiState.state = UNSELECTING;
              
            else % WAITING(CDSU) --> SELECTING
              mUiState.sel   = unique([mUiState.sel, mUiState.tagged]); 
              mUiState.state = SELECTING;
              
            end
          else % WAITING(CD-*): near object and no key modifiers
            if ismember(mUiState.tagged, mUiState.sel) % WAITING(CD-S) ---> SELECTING
              mUiState.state = SELECTING;
              
            else % WAITING(CD-U) ---setsel---> SELECTING
              mUiState.sel = mUiState.tagged;
              mUiState.state = SELECTING;
            end            
          end
          localUpdateGui;
          
        else % WAITING: far from everything
          if mUiState.shift || mUiState.ctrl % WAITING(CDSF) --> clear selection
            mUiState.sel = [];

          else % WAITING(CD-F) --> add new annotation
            % +0.5 is to move them to the center of the pixel for UI
            % purposes
            imgH = size(mImage,1);
            mLocs(:,end+1) = [...
              imgC;
              imgR;
              imgC;
              min(imgH-1, imgR+mDefaultHeight)] + 0.5;
            mUiState.sel = [ size(mLocs,2) ]; %#ok<NBRAK>
          end
          localUpdateGui;
        end
        
        
      case SELECTING
        assert(~isempty(mUiState.tagged));
        if mouseMoved % SELECTING(M***)
          if isNearTaggedCoords(scrX,scrY) % SELECTING(M**T)
            % do nothing
            
          else % SELECTING(M**t)
            moveSelected(scrX,scrY);
            mUiState.state = MOVING;
            localUpdateGui;
          end
          
        elseif mUiState.button % SELECTING(CD**)
          % do nothing
          
        else % SELECTING(CU**) ---untag---> WAITING
          clearTag;
          mUiState.state = WAITING;
          localUpdateGui;
          
        end
        
        
      case UNSELECTING
        assert(~isempty(mUiState.tagged));
        if mouseMoved % UNSELECTING(M***)
          if isNearTaggedCoords(scrX,scrY) % UNSELECTING(M**T)
            % do nothing
            
          else % SELECTING(M**t)
            moveSelected(scrX,scrY);
            mUiState.state = MOVING;
            localUpdateGui;
          end
          
        elseif mUiState.button % UNSELECTING(CD**)
          % do nothing
          
        else % UNSELECTING(CU**) ---untag,-sel---> WAITING
          mUiState.sel = setdiff(mUiState.sel, mUiState.tagged);
          clearTag;
          mUiState.state = WAITING;
          localUpdateGui;
          
        end
        
        
      case MOVING
        if mouseMoved || mUiState.button % MOVING(M***), MOVING(CD**)
          moveSelected(scrX,scrY);
          localUpdateGui;
                    
        else % MOVING(CU**)
          clearTag;
          fixInvertedBboxes;
          mUiState.state = WAITING;
          localUpdateGui;
          
        end
        
        
      otherwise, 
        error('Illegal state: %d', mUiState.state);
    end
    
    mUiState.prevCoords = [scrX;scrY];
  end

%------------------------------------------------------------------
  function cleared = clearTag
    cleared = ~isempty(mUiState.tagged);
    mUiState.tagged       = [];
    mUiState.taggedPart   = [];
    mUiState.taggedCoords = [];
  end
    
%------------------------------------------------------------------
  function near = isNearTaggedCoords(x,y)
    % Is the current mouse point near the place where the user clicked when
    % tagging the most recent object?  Reason for this function: if the
    % user hasn't moved too far from the click, suppress the motion.
    assert(numel(mUiState.tagged)==1);
    assert(numel(mUiState.taggedPart)==1);
    assert(numel(mUiState.taggedCoords)==2);
    near = (sum(([x;y]-mUiState.taggedCoords).^2) <= mNearTaggedPt2);
  end

%------------------------------------------------------------------
  function wasTagged = tagNearestObject(x,y)
    clearTag;
    wasTagged = 0;
    bestDist2 = inf;
    toShow = getVisible();
    for annoIdx=1:size(mLocs,2)
      if (toShow(annoIdx))
        [near,part,dist2] = findNearestAnnoPart(x,y, annoIdx, PARTS);
        if near && dist2 <= bestDist2
          mUiState.tagged       = annoIdx;
          mUiState.taggedPart   = part;
          mUiState.taggedCoords = [x;y];
          wasTagged             = 1;
          bestDist2             = dist2;
        end
      end
    end
  end

%------------------------------------------------------------------
  function moveSelected(x,y)
    assert(numel(mUiState.tagged)==1);
    assert(numel(mUiState.taggedPart)==1);
    assert(numel(mUiState.taggedCoords)==2);

    z = getZoom;
    dx = (x-mUiState.prevCoords(1)) * z;
    dy = (y-mUiState.prevCoords(2)) * z;

    bodyvecNorm = normalizeColumns(...
      mLocs([3 4],mUiState.sel) - mLocs([1 2],mUiState.sel));
    beamvecNorm = [bodyvecNorm(2,:) ; -bodyvecNorm(1,:)]; % 90deg rotation
    
    dxcrown=0; dycrown=0; dxfootbase=0; dyfootbase=0;
    switch mUiState.taggedPart
      case CROWN
        dxcrown=dx;    dycrown=dy;
      case FOOTBASE
        dxfootbase=dx; dyfootbase=dy;

      case HEAD
        proj = dot([dx dy], bodyvecNorm);
        dxcrown   =proj*bodyvecNorm(1,:);  dycrown   =proj*bodyvecNorm(2,:);
      case FEET
        proj = dot([dx dy], bodyvecNorm);
        dxfootbase=proj*bodyvecNorm(1,:);  dyfootbase=proj*bodyvecNorm(2,:);

      case BODY
        proj = dot([dx dy], beamvecNorm);
        dxcrown   =proj*beamvecNorm(1,:);  dycrown   =proj*beamvecNorm(2,:);
        dxfootbase=proj*beamvecNorm(1,:);  dyfootbase=proj*beamvecNorm(2,:);

      otherwise,
        error('Illegal tagged part');
    end
    
    mLocs(1,mUiState.sel) = mLocs(1,mUiState.sel) + dxcrown;
    mLocs(2,mUiState.sel) = mLocs(2,mUiState.sel) + dycrown;
    mLocs(3,mUiState.sel) = mLocs(3,mUiState.sel) + dxfootbase;
    mLocs(4,mUiState.sel) = mLocs(4,mUiState.sel) + dyfootbase;
  end    

%------------------------------------------------------------------
  function v = normalizeColumns(v)
    n = sqrt(sum(v.^2, 1));
    v = v ./ repmat(n, size(v,1),1);
  end

%------------------------------------------------------------------
  function fixInvertedBboxes
    yinv = (mLocs(2,:) > mLocs(4,:));
    
    tmp = mLocs(1:2,yinv);
    mLocs(1:2,yinv) = mLocs(3:4,yinv);
    mLocs(3:4,yinv) = tmp;
  end
 
%------------------------------------------------------------------
  function [near,part,dist2] = findNearestAnnoPart(x,y, annoIdx, allowedParts)
    locs = screenCenteredFromImg(mLocs(:,annoIdx))+0.5;  % -> screen coords

    % border thickness
    [toShow,nShown] = getVisible();
    t  = borderThicknessScr(nShown(annoIdx));
    
    % ibeam geometry
    bodyvec     = (locs([3 4]) - locs([1 2]));
    height      = norm(bodyvec);
    bodyvecNorm = bodyvec / height;
    beamvecNorm = [bodyvecNorm(2) ; -bodyvecNorm(1)]; % 90deg rotation
    halfwidth   = mDefaultAspect * height / 2;

    % convert to local ibeam coordinate system
    xyRelBase = [beamvecNorm bodyvecNorm]' * [x-locs(3) ; y-locs(4)];
    xrb = xyRelBase(1);
    yrb = xyRelBase(2);

    % distances to all parts
    dCrown    = sum(([x;y]-locs([1 2])).^2);
    dFootbase = sum(([x;y]-locs([3 4])).^2);
    dHead     = dist2toHorizSeg(xrb,yrb, t, -halfwidth,halfwidth, -height);
    dFeet     = dist2toHorizSeg(xrb,yrb, t, -halfwidth,halfwidth, 0);
    dBody     = dist2toVertSeg( xrb,yrb, t, 0, -height,0);
        
    % find what we're near, if anything, giving priority to hard-to-select
    % parts. 
    near = 1;
    if     bitand(allowedParts,CROWN)    && (dCrown    <= mNearVertex2), dist2=dCrown;    part = CROWN;
    elseif bitand(allowedParts,FOOTBASE) && (dFootbase <= mNearVertex2), dist2=dFootbase; part = FOOTBASE;
    elseif bitand(allowedParts,HEAD)     && (dHead     <= mNearEdge2  ), dist2=dHead;     part = HEAD;
    elseif bitand(allowedParts,FEET)     && (dFeet     <= mNearEdge2  ), dist2=dFeet;     part = FEET;
    elseif bitand(allowedParts,BODY)     && (dBody     <= mNearEdge2  ), dist2=dBody;     part = BODY;
    else
      near  = 0;
      part  = [];
      dist2 = 0;
    end
  end

%------------------------------------------------------------------
  function dist2 = dist2toHorizSeg(x,y, t, segxmin,segxmax,segy)
    assert(segxmin<=segxmax);
    if x<segxmin % left of line segment -- L2 distance to endpoint
      dist2 = sum(([x;y]-[segxmin;segy]).^2);
    elseif x>segxmax % right of line segment -- L2 distance to endpoint
      dist2 = sum(([x;y]-[segxmax;segy]).^2);
    else
      if y>segy-t/2 && y<segy+t/2 % inside band
        dist2 = 0;
      else
        dist2 = (abs(y-segy) - t/2).^2; % outside band
      end
    end
  end

%------------------------------------------------------------------
  function dist2 = dist2toVertSeg(x,y, t, segx, segymin,segymax)
    dist2 = dist2toHorizSeg(y,x, t, segymin,segymax, segx);
  end

%------------------------------------------------------------------
  function processUserInputs(varargin)
   
    if nargin == 0
      error('Invalid arguments.  Type HELP %s', mfilename);
      
    elseif nargin==1 && ischar(varargin{1})    
      fname = varargin{1};
      try img = imread(fname); catch img=[]; end
      if ~isempty(img)                              % ibeamAnnotationGui(imagename)
        [basedir,ff,ext] = fileparts(fname);
        mBaseDir                  = basedir;
        mIbeamFilename            = '';
        mIbeamHier(1).imagename   = [ff ext];
        mIbeamHier(1).locs        = zeros(4,0);
        cleanUpAndValidateImportedIbeams;
        selectImage(1);
        
      else                                          % ibeamAnnotationGui(ibeamFile)
        importIbeamFile(fname);
      end

    elseif nargin==2 && isstruct(varargin{2})       % ibeamAnnotationGui(basedir,ibeamHier)
      mBaseDir         = varargin{1};
      mIbeamFilename   = '';
      mIbeamHier       = varargin{2};
      cleanUpAndValidateImportedIbeams;
      selectImage(1);
      
    elseif nargin==2 && ischar(varargin{2})         % ibeamAnnotationGui(basedir, ibeamFilename)
      mBaseDir         = varargin{1};
      mIbeamFilename   = varargin{2};
      if exist(fullfile(mBaseDir, mIbeamFilename), 'file')
        importIbeamFile(fullfile(mBaseDir, mIbeamFilename));
      else
        mIbeamHier = buildNewHierFromDirSearch(mBaseDir);
        cleanUpAndValidateImportedIbeams;
        selectImage(1);
      end
      
    elseif nargin==2 && isnumeric(varargin{2});     % ibeamAnnotationGui(imgname,locs)
      [basedir,ff,ext] = fileparts(varargin{1});   
      mBaseDir                  = basedir;
      mIbeamFilename            = '';
      mIbeamHier(1).imagename   = [ff ext];
      mIbeamHier(1).locs        = varargin{2};
      cleanUpAndValidateImportedIbeams;
      selectImage(1);
      
    else
      error('Invalid arguments.  Type HELP %s', mfilename);
    end
  end

%------------------------------------------------------------------
  function importIbeamFile(fname)
    % Load the file (no changes if the load fails)
    data = load(fname, 'ibeamHier');
    
    mIbeamHier        = data.ibeamHier;
    [mBaseDir,ff,ext] = fileparts(fname);
    mIbeamFilename    = [ff ext];
    
    cleanUpAndValidateImportedIbeams;
    selectImage(1);
  end

%------------------------------------------------------------------
  function hier = buildNewHierFromDirSearch(basedir)
    hier = struct;
    [dummy,dummy,fnames] = dirr(basedir, '\.png\>', 'name');
    for i=1:length(fnames)
      hier(i).imagename = fnames{i}(length(basedir)+2:end);
      hier(i).locs      = zeros(4,0);
    end
  end

%------------------------------------------------------------------
  function cleanUpAndValidateImportedIbeams
    fixInvertedBboxes();

    for i=1:length(mIbeamHier)
      % datatype and dimensions checks
      anno = mIbeamHier(i);
      assert(ischar(anno.imagename));
      assert(isnumeric(anno.locs) && ndims(anno.locs)==2 && size(anno.locs,1)==4);
    end
    
    % remove unrecognized fields (since we'd probably handle them
    % incorrectly) 
    mIbeamHier = rmfield(mIbeamHier, ...
      setdiff(fieldnames(mIbeamHier), {'imagename','locs'}));
  end

%------------------------------------------------------------------
  function selectImage(num)
    assert(num>0 && num<=length(mIbeamHier));
    mCurrImgNum  = num;
    % +0.5 is to move them to the center of the pixel for UI purposes
    mLocs  = mIbeamHier(num).locs + 0.5;
    mImage = imread(fullfile(mBaseDir, mIbeamHier(num).imagename));
    hImg = [];

    updateWindowTitle;
  end

%------------------------------------------------------------------
  function updateWindowTitle
    [imgC,imgR,scrX,scrY,rawPt] = getCurrentPoint;
    
    coordsstr = sprintf('image raw (%f,%f),  image (%d,%d),  screen (%d,%d)',...
      rawPt(1),rawPt(2), imgC,imgR, scrX,scrY);

    if mIbeamHier(mCurrImgNum).imagename
      windowTitle = sprintf('%s (%d of %d) - %s - %s',...
        mIbeamHier(mCurrImgNum).imagename,...
        mCurrImgNum, length(mIbeamHier), mfilename, coordsstr);
    else
      windowTitle = [mfilename coordsstr];
    end
    
    set(hMainFigure, 'Name',windowTitle, 'NumberTitle','off');
  end

%------------------------------------------------------------------
  function [imgC,imgR,scrX,scrY,rawPt] = getCurrentPoint
    rawPt = get(hImgAxes,'currentpoint');
    rawPt = rawPt(1,1:2);
    
    imgC = floor(rawPt(1)+0.5);          
    imgR = floor(rawPt(2)+0.5);
    
    scrX = (rawPt(1)-0.5)/getZoom-0.5;   
    scrY = (rawPt(2)-0.5)/getZoom-0.5;
  end

%------------------------------------------------------------------
%   function [imgC,imgR] = imgFromScreen(scrX,scrY)
%   % image coordinates given screen coordinates.  Use floor on the results
%     to get image indices.
%     imgC = ((scrX-0.5)*getZoom + 1);
%     imgR = ((scrY-0.5)*getZoom + 1);
%   end

%------------------------------------------------------------------
  function [screenCoords] = screenCenteredFromImg(imgCoords)
    % screen coordinates of the center of the given image pixel.  
    screenCoords = zeros(size(imgCoords));
    screenCoords(1:2:end,:) = (floor(imgCoords(1:2:end,:)) - 1 + 0.5) / getZoom - 0.5;
    screenCoords(2:2:end,:) = (floor(imgCoords(2:2:end,:)) - 1 + 0.5) / getZoom - 0.5;
  end
end

%=========================================================================
function prepareLayout(topContainer)
% This is a utility function that takes care of issues related to
% look&feel and running across multiple platforms. You can reuse
% this function in other GUIs or modify it to fit your needs.
allObjects = findall(topContainer);
warning off  %#ok<WNOFF> %Temporary presentation fix
try
  titles=get(allObjects(isprop(allObjects,'TitleHandle')), 'TitleHandle');
  allObjects(ismember(allObjects,[titles{:}])) = [];
catch
end
warning on %#ok<WNON>

% Use the name of this GUI file as the title of the figure
defaultColor = get(0, 'defaultuicontrolbackgroundcolor');
if isa(handle(topContainer),'figure')
  set(topContainer,'Name', mfilename, 'NumberTitle','off');
  % Make figure color matches that of GUI objects
  set(topContainer, 'Color',defaultColor);
end

% Make GUI objects available to callbacks so that they cannot
% be changes accidentally by other MATLAB commands
% TODO: uncomment when done debugging
%%set(allObjects(isprop(allObjects,'HandleVisibility')), 'HandleVisibility', 'Callback');

% Make the GUI run properly across multiple platforms by using
% the proper units
if strcmpi(get(topContainer, 'Resize'),'on')
  set(allObjects(isprop(allObjects,'Units')),'Units','Normalized');
else
  set(allObjects(isprop(allObjects,'Units')),'Units','Characters');
end

% You may want to change the default color of editbox,
% popupmenu, and listbox to white on Windows
if ispc
  candidates = [findobj(allObjects, 'Style','Popupmenu'),...
    findobj(allObjects, 'Style','Edit'),...
    findobj(allObjects, 'Style','Listbox')];
  set(findobj(candidates,'BackgroundColor', defaultColor), 'BackgroundColor','white');
end

end
