function [experiments] = loadDetOptErrorData(basedir)
%[experiments] = loadDetOptErrorData(basedir)
%
%  If there are N experimental configurations, each with M(i) parameter
%  settings, then the return value has at least the following field set:
%
%    experiments.train.imageset                1xN cell array of strings
%    experiments.train.annoset                 1xN cell array of strings
%    cell array of strings                     1xN cell array of strings
%    experiments.train.hogId                   1xN cell array of strings
%    experiments.train.windowset               1xN cell array of strings
%    experiments.train.impl                    1xN cell array of strings
%    experiments.train.pass                    1xN cell array of strings
%
%    experiments.test.errorType                1xN cell array of strings
%
%    experiments.results.legend                1xN cell of strings
%
%    experiments.results.truePos               1xN cell of M(i)x1 ints
%    experiments.results.falsePos              1xN cell of M(i)x1 ints
%    experiments.results.falseNeg              1xN cell of M(i)x1 ints
%    experiments.results.trueNeg               1xN cell of M(i)x1 ints
%
%    experiments.results.pos                   1xN cell of ints
%    experiments.results.neg                   1xN cell of ints

% find files
[dummy,allfiles] = system(['ls ' basedir '/*/*/*/*/detopt.csv']);
allfiles         = cellfilt(@(f) length(f), split([10 13], allfiles));
%  For now, we only look for finished files.  There seem to be some really
%  odd file corruption problems when reading the .csv files on Windows
%  while they're being written to NFS from Linux.
% [dummy,alldonelogs] = system(['grep -l real ' basedir '/*/*/*/*/detopt.csv.log']);
% alldonelogs         = cellfilt(@(f) length(f), split([10 13], alldonelogs));
% allfiles            = map(@(f) f(1:end-4), alldonelogs);

% Filename parsing
extractParent   = @(n) n(length(basedir)+1+1:end-length('detopt.csv')-1);
parentDirNames  = cellfun(extractParent, allfiles, 'UniformOutput',0);
tokens          = regexp(parentDirNames, '([^_/]+)_([^_/]+)/([^_/]+)_([^_/]+)_([^_/]+)/([^_/]+)/([^_/]+)_error', 'tokens');
experimentsFlat = map(@(wrappedTokens) fullcollapse(wrappedTokens), tokens);
experimentsFlat = cat(1, experimentsFlat{:}); %     2D array

experiments = struct;
experiments.train = struct;
experiments.train.imageset  = {experimentsFlat{:,1}};
experiments.train.annoset   = {experimentsFlat{:,2}};
experiments.train.hogId     = {experimentsFlat{:,3}};
experiments.train.windowset = {experimentsFlat{:,4}};
experiments.train.impl      = {experimentsFlat{:,5}};
experiments.train.pass      = {experimentsFlat{:,6}};
experiments.test = struct;
experiments.test.errorType = {experimentsFlat{:,7}};

% Hard-coded parameters:
% Approximate number of windows tested.  We assume 1920x1080 frames sampled
% at every 8x8 pixel locations after scaling up by a factor of 2 in x and
% y, for 25 train and 12 test images.
numWin.train = (1920*1080/8/8*2*2 * 25);
numWin.test  = (1920*1080/8/8*2*2 * 12);

% Read PR data
experiments.results = struct;
experiments.results.legend = parentDirNames;
valid = true(length(allfiles),1);
for i=1:length(allfiles)
  fid = fopen(allfiles{i});
  line = fgetl(fid);
  try
    fieldNames = map(@strtrim, split(',', line));
    raw = textscan(fid,'%f', 'Delimiter',',');
    partialLineLen = mod(numel(raw{1}), length(fieldNames));
    raw = reshape(raw{1}(1:end-partialLineLen), ...
      [length(fieldNames), floor(numel(raw{1})/length(fieldNames))])';
    fclose(fid);

    arrs = struct;
    for j=1:length(fieldNames)
      arrs.(fieldNames{j}) = raw(:,j);
    end

    pos = (arrs.truePos + arrs.falseNeg);
    assert(all(pos == pos(1)));
    arrs.pos = pos(1);

    neg = numWin.(experiments.test.errorType{i}) - arrs.pos;
    assert(all(neg == neg(1)));
    arrs.neg = neg(1);

    arrs.trueNeg = arrs.neg - arrs.falsePos;

    for fnc = fieldnames(arrs)'
      fn = fnc{1};
      experiments.results.(fn){i} = arrs.(fn);
    end
  catch %#ok<CTCH>
    s = lasterror; %#ok<LERR>
    warning('io:emptyfile', 'Bad file: %s\n%s', allfiles{i}, s.message);
    valid(i) = false;
  end
end


for i=fieldnames(experiments)'
  i = i{1}; %#ok<FXSET>
  for j=fieldnames(experiments.(i))'
    j = j{1}; %#ok<FXSET>
    experiments.(i).(j) = {experiments.(i).(j){valid}};
  end
end
